/* FILE: cmpIFCompiler.c
 *
 * PROGRAM: User Friendly Metadata (UFM)
 *
 * DESCRIPTION: These functions manipulate the instructions that will be used to
 *              transform the input ODL into the output HTML. The instructions
 *              file compiler does two passes: the first pass (cmpPass1) operates
 *              upon an ODL tree of instructions, creating a series of lookup
 *              tables: 
 *					pMacroLUT           macro, translation string
 *					pAggByNameLUT		aggregate name, pointer to instruction
 *					pAggByClassLUT 		aggregate class name, pointer to instruction
 *					pParByNameLUT		parameter name, pointer to instruction
 *					pParByClassLUT		parameter class name, pointer to instruction
 *					pIDTranslationLUT	identifier name, translation string
 *					pClassDefLUT		class name, class definition.
 *              As the instructions are being processed they are individually 
 *              checked for completeness accroding to the definition given in
 *              data dictionary (IFDD.odl).
 *
 *              The second pass (cmpPass2) interprets the instructions with
 *              respect to the ODL input tree and builds the report structure.
 *
 * RCS INFO:
 *
 *      $Id: cmpIFCompiler.c,v 1.4 2000/04/07 00:44:02 ToolsDev Exp $
 *
 *      $Log: cmpIFCompiler.c,v $
 *      Revision 1.4  2000/04/07 00:44:02  ToolsDev
 *      Made changes to display the ODL file header informations:
 *      wrote code to check if header exist in the odl file and
 *      allocate the needed memory for it.
 *      (Amer Rahabi)
 *
 * Revision 1.3  1999/04/06  15:43:32  jarva
 * Promotion of branch 1.2.0 (cleared wlint warnings) for release.
 *
 * Revision 1.2.0.4  1999/03/10  14:37:04  stroup
 * Further correction of errors found by lint: changed some variables of
 * type 'int' to 'unsigned int'.
 *
 * Revision 1.2.0.3  1999/03/01  18:13:53  ToolsDev
 * Corrected comment with missing terminator.
 *
 * Revision 1.2.0.2  1999/02/17  15:40:18  ToolsDev
 * Minor corrections to eliminate Lint warnings.
 *
 * Revision 1.2.0.1  1999/02/11  19:33:49  jarva
 * Bug fix: "=" in conditionals rather than "==". (cmpHasClassInst).
 *
 * Revision 1.2  1999/02/03  18:34:26  jarva
 * Promotions of the 1.1.0 branch
 *
 * Revision 1.1.0.10  1999/02/02  21:23:43  jarva
 * Upgrade for predefined macros with arguments.
 *
 * Revision 1.1.0.9  1999/01/26  17:12:18  jarva
 * Corrected references to ToUpStrcmp & cmpComponentName.
 *
 * Revision 1.1.0.8  1999/01/21  22:34:46  jarva
 * Modification to cmpAssignmentProcessor; RightSide now takes sequences.
 *
 * Revision 1.1.0.6  1999/01/20  19:35:03  stroup
 * Replaced ODL_Node.h include with odlSupport.h include.
 * Removed odlDestructor function and its prototype: they are now in
 * odlSupport.c and odlSupport.h, respectively.
 *
 * Revision 1.1.0.5  1999/01/20  17:53:33  stroup
 * User can specify how a parameter's data is displayed in a table: (1) all
 * values of a parameter can be placed in their own cell of the table; or
 * all values of a parameter can be placed together in a single cell.
 *
 * Revision 1.1.0.4  1999/01/18  18:35:22  stroup
 * Put in capability to hide empty columns/rows of a table.  Incorporated use
 * oflstDestructor function in the Table Processor.  Incorporated use of
 * lstReplaceCurrent function in cmpElemMacExpand.  Added lstStopIteration
 * functions to correct bugs.
 *
 * Revision 1.1.0.3  1999/01/15  20:26:19  stroup
 * Changed data type of table_data--the table processor's internal
 * representation of a table's data--from an array of pointers to a lstLIST
 * structure.
 *
 * Revision 1.1.0.2  1999/01/14  19:32:28  jarva
 * Correction: removed 2 prototypes that had been inserted for
 * test purposes.
 *
 * Revision 1.1.0.1  1999/01/12  22:08:12  jarva
 * Intermediate fix to cmpAssignmentProcessor, which now accepts a
 * reference list string containing one or more ODL references. More
 * changes to this function are planned.
 *
 * Revision 1.1  1999/01/06  20:34:20  ToolsDev
 * Initial revision from prototype baseline.
 *
 *
*/
#define VERBOSE_DEBUG 0   /* Set to 1 to turn on verbose debug messages; to
                           * disable, set to 0. */
#define IDENTIFY_STUBS 0  /* Set to 1 to turn on messages that identify stubs; to
                           * disable, set to 0. */

/* INCLUDED FILES:
*/
#include "lstList.h"
#include "cmpIFCompiler.h"
#include "ufmMain.h"
#include "odldef.h"
#include "sltLUT.h"
#include "odlSupport.h"
#include "ListExpander.h"
#include "macMacro.h"
#include "synODL.h"
#include "strUtilities.h"


#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/* ---------------------------------------------------------------------*/

/* GLOBAL VARIABLES: the following variables may be used in other files.
 *    (global prototyes, if any, are given in the header file cmpIFCompiler.h)
*/
/*     The following declarations are of the various LUTs that are set up
 *     by the first pass of the compiler.  The LUTs are all initialized in
 *     their respecitve build function (e.g. BuildMacroLUT() ) the first
 *     time it's called.
 */
sltLUT pMacroLUT = (sltLUT) NULL;
sltLUT pAggByNameLUT= (sltLUT) NULL;
sltLUT pAggByClassLUT= (sltLUT) NULL;	
sltLUT pParByNameLUT= (sltLUT) NULL;
sltLUT pParByClassLUT= (sltLUT) NULL;
sltLUT pIDTranslationLUT= (sltLUT) NULL;
sltLUT pClassDefLUT = (sltLUT) NULL;

lstLIST pCCStack;  /* Points at the current-component stack */

synDD pIFDD; /* Points at the data dictionary for the IF */
/* ---------------------------------------------------------------------*/
/* EXTERNAL VARIABLES:
*/
extern ufmControls ufmCONTROLS; /* User options, etc. */

/* ---------------------------------------------------------------------*/
/* STATIC VARIABLES: the following variables are for the exclusive use
 *     of the functions defined within this file.
*/
FILE *pCompilerOut;
/* ---------------------------------------------------------------------*/

/* STATIC PROTOTYPES: the following prototypes are for functions defined
 *     and used exclusively within this file.
*/
/*   THE FIRST PASS COMPILER */
static INST_KIND cmpIKind(OBJECT pObj );
static void cmpBuildMacroLUT( OBJECT pObj );
static void cmpBuildAggLUTs( OBJECT pObj );
static void cmpBuildParLUTs( OBJECT pObj );
static void cmpBuildID_LUT( OBJECT pObj );
static void cmpBuildClassLUT( OBJECT pObj );
static int cmpCheckIDTransSyntax( OBJECT pObj );
static int cmpCheckMacroDefSyntax( OBJECT pObj );
static int cmpCheckAggTransSyntax( AGGREGATE pAgg );
static int cmpCheckParTransSyntax( AGGREGATE pAgg );
static int cmpCheckClassDefSyntax( OBJECT pObj );
static int cmpBuildIFDD( synDD pIFDD );

/*   THE SECOND PASS COMPILER    */

static lstLIST cmpListProcessor( char *listSpec, 
								 struct ODL_Node *pCurrentComp );
static lstLIST cmpTranslationProcessor(struct ODL_Node *pNode);
static OBJECT cmpFindInstruction( struct ODL_Node *pNode );
static OBJECT cmpGetClassInstruction( struct ODL_Node *pNode );
static char *cmpComponentID( struct ODL_Node *pNode, OBJECT pInst );
static char *cmpListSpec( AGGREGATE pAgg );
static REP_NODE cmpAssignmentProcessor( OBJECT pObj, 
									   struct ODL_Node *pNode );
static OBJECT cmpHasClassInst( struct ODL_Node *pNode );
static int cmpIsInClass( struct ODL_Node *pNode, OBJECT pClassDef);
static lstLIST cmpInstructionProcessor( OBJECT pInst, struct ODL_Node *pNode);
static REP_NODE cmpTableProcessor( OBJECT pObj, struct ODL_Node *pNode );


/*   THE TABLE PROCESSOR    */

static void cmpElemMacExpand (lstLIST pList);
static unsigned int cmpMajorDimSize (lstLIST major_hdrs, lstLIST table_data);
static unsigned int cmpMinorDimSize (lstLIST minor_hdrs, lstLIST table_data);
static lstLIST cmpTblBackfill (unsigned int num_major, unsigned int num_minor,
                               lstLIST major_hdrs, lstLIST minor_hdrs,
                               lstLIST table_data);
static struct reportTable *cmpInitRptTbl
    (unsigned int num_cols, unsigned int num_rows, char *caption,
     char *cols_title, char *rows_title, lstLIST row_headers,
     lstLIST col_headers, lstLIST table_data, const char *major);
static char *cmpGetExpInstCompVal (OBJECT pInst, const char *comp_name);
static lstLIST cmpGetExpInstCompList (OBJECT pInst, const char *comp_name);
static lstLIST cmpGetTableData (struct ODL_Node *pNode, lstLIST value_list,
                                char *cell_contents);
static void *cmpTblDataDestructor (void *col_row_list);
static unsigned int cmpHideEmpty (lstLIST table_data, lstLIST maj_hdrs);
static void cmpColRow2MajMin (unsigned int num_cols, unsigned int num_rows,
                              char *major,
                              unsigned int *num_maj, unsigned int *num_min);
static void cmpMajMin2ColRow (unsigned int num_maj, unsigned int num_min,
                              char *major,
                              unsigned int *num_cols, unsigned int *num_rows);

/*    GENERAL PURPOSE     */
static void cmpUserError( char *functionName, char *msg );
static void cmpUserWarning( char *functionName, char *msg );
static void cmpFatalError( char *functionName, char *msg );
static void cmpDebugMsg( char *functionName, char *msg );
static void cmpStubNotice( char *functionName, char *msg );
static void cmpOutputRepStruc( lstLIST pReportList, int level);

/* ---------------------------------------------------------------------*/
 
/*****************************************************************************
 * 
 * NAME: cmpPass1
 *
 * DESCRIPTION: The first pass of the compilation of the instructions file.
 *
 * PARAMETERS:  
 *     pIFTree -- (AGGREGATE) Points at the root node of the ODL tree that
 *                 has been parsed from the instructions file.
 *
 * RETURN VALUE:
 *     cmpPass1 -- int Reports the success or failure of the step. 1 ->
 *                     success; 0 -> failure.
 *     
 * GLOBAL VARIABLES:
 *     pMacroLUT		-- (sltLUT) Macro translation strings indexed by
 *                                  the name of the macro.
 *     pAggByNameLUT	-- (sltLUT) Pointer to an aggregate translation
 *                                  instruction indexed by the name of the
 *                                  aggregate to which the instruction 
 *                                  applies.
 *     pAggByClassLUT	-- (sltLUT) Pointer to an aggregate translation
 *                                  instruction indexed by the name of the
 *                                  class of aggrigate to which the 
 *                                  instruction applies.
 *     pParByNameLUT	-- (sltLUT) Pointer to a parameter translation
 *                                  instruction indexed by the name of the
 *                                  parameter to which the instruction 
 *                                  applies.
 *     pParByClassLUT	-- (sltLUT) Pointer to a parameter translation
 *                                  instruction indexed by the name of the
 *                                  class of parameter to which the 
 *                                  instruction applies.
 *     pIDTranslationLUT-- (sltLUT) ODL identifier translation strings indexed by
 *                                  the name of the identifier.
 *     pClassDefLUT		-- (sltLUT) Pointer to a class definition indexed
 *                                  by class name.
 *	
 * NOTES:
 *     1. Initializes all of the LUT tables used by the 2nd pass compiler.
 *     2. Steps through the instructions file and then process each 
 *        instruction found.
 *     3. This function is not a final version. In the prototype, an
 *        instructions file is required; later versions will have default
 *        instructions built into the code.
 *
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva  -- Original
 *   10/14/98  Ken Jarva  -- Updated doc header.
 *
 ****************************************************************************/

int cmpPass1( AGGREGATE pIFTree )
{
	AGGREGATE pObj;
	int success = 1, failure = 0, haveIF = 0;

	assert( pIFTree );
	if( !pIFTree->first_child ) {
		cmpUserError( "cmpPass1", "No instructions file. This version requires "
			"an instructions file." );
		cmpFatalError( "cmpPass1", "No instructions file. This version requires "
			"an instructions file." );
		return failure; /* only incase of return from cmpFatalError */
	} else haveIF = 1;


	/* Initialize all of the LUTs. */
	pMacroLUT = sltInitLUT();
	pAggByNameLUT = sltInitLUT();
	pAggByClassLUT = sltInitLUT();	
	pParByNameLUT = sltInitLUT();
	pParByClassLUT = sltInitLUT();
	pIDTranslationLUT = sltInitLUT();
	pClassDefLUT = sltInitLUT();

	/* Set up a data dictionary for the instructions file: */
	if( haveIF ) {
		pIFDD = synInitDataDict();
		if( !cmpBuildIFDD(pIFDD) ) {
		  cmpFatalError("cmpPass1", "Have no data dictionary for IF "
				"can't check syntax for instructions." );
		  }
		
	}

	/* Set the pointer to the output file: */
	pCompilerOut = ufmCONTROLS.pCompilerErrors;


	/* Get the first child of the root node */
	pObj = NextAggregate( pIFTree );

 	while( pObj != NULL ) { 
		switch( cmpIKind(pObj) )
		{
			case AGGREGATE_TRANSLATION:
				cmpBuildAggLUTs( pObj );
				break;
			case PARAMETER_TRANSLATION:
				cmpBuildParLUTs( pObj );
				break;
			case IDENTIFIER_TRANSLATIONS:
				cmpBuildID_LUT( pObj );
				break;
			case MACRO_DEFINITIONS:
				cmpBuildMacroLUT( pObj );
				break;
			case CLASS_DEFINITION:
				cmpBuildClassLUT( pObj );
				break;
			default:
				{
					char msg[256];
					sprintf( msg, "Unknown Instruction = %s", pObj->name );
					cmpUserWarning( "cmpPass1", msg );
				}
		}
		pObj = pObj->right_sibling;
	}

	/* Inform caller of status: */
	if( ufmCONTROLS.CompilerFailure == 0 )
		return success;
	else return failure;

} /* end cmpPass1 */

/*****************************************************************************
 * 
 * NAME: cmpIKind
 *
 * DESCRIPTION: Determines the kind of instruction that is represented by a
 *              particular ODL object.
 *
 * PARAMETERS:  
 *     pObj -- OBJECT ODL object that has been parsed from the instructions
 *             file.
 *
 * RETURN VALUE:
 *     cmpIKind -- INST_KIND Enumerated value representing the kind of 
 *                           instruction found.
 *     
 * GLOBAL VARIABLES: None
 *	
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva   -- Original
 *
 ****************************************************************************/

INST_KIND cmpIKind(OBJECT pObj )
{

	if( strcmp( pObj->name, "AGGREGATETRANSLATION") == 0 )
		return AGGREGATE_TRANSLATION;
	if( strcmp( pObj->name, "PARAMETERTRANSLATION") == 0 )
		return PARAMETER_TRANSLATION;
	if( strcmp( pObj->name, "IDENTIFIERTRANSLATIONS") == 0 )
		return IDENTIFIER_TRANSLATIONS;
	if( strcmp( pObj->name, "MACRODEFINITIONS") == 0 )
		return MACRO_DEFINITIONS;
	if( strcmp( pObj->name, "COMPONENTCLASSDEFINITION") == 0 )
		return CLASS_DEFINITION;
	return UNKNOWN_INSTRUCTION;

} /* end cmpIKind */
/*****************************************************************************
 * 
 * NAME: cmpFatalError
 *
 * DESCRIPTION: Reports fatal errors and shuts down the program.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

void cmpFatalError( char *functionName, char *msg )
{
	if( ufmCONTROLS.debug )
		fprintf( ufmCONTROLS.pDebugFile, 
		                              "%s: FATAL - %s\n", functionName, msg );
	cmpUserWarning( "cmpFatalError", 
		     "Fatal programming error is causing the program to shut down." );
	exit(-1);

} /* end cmpFatalError */
/*****************************************************************************
 * 
 * NAME: cmpDebugMsg
 *
 * DESCRIPTION: Reports debug messages.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

void cmpDebugMsg( char *functionName, char *msg )
{
	if( ufmCONTROLS.debug && VERBOSE_DEBUG ){
		fprintf( ufmCONTROLS.pDebugFile, 
		                              "%s: INFO - %s\n", functionName, msg );
	    fflush(ufmCONTROLS.pDebugFile);
	}
	return;

} /* end cmpDebugMsg */
/*****************************************************************************
 * 
 * NAME: cmpUserWarning
 *
 * DESCRIPTION: Reports warnings if the VERBOSE flag is set.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *	
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

void cmpUserWarning( char *functionName, char *msg )
{
	if( !ufmCONTROLS.verbose ) return;

    fprintf( ufmCONTROLS.pCompilerErrors, 
		                             "%s: WARNING - %s\n", functionName, msg );
	return;

} /* end cmpUserWarning */

/* --------------------------------------------------------------------------*/

/*****************************************************************************
 * 
 * NAME: cmpUserError
 *
 * DESCRIPTION: Reports fatal errors and sets CompilerFailure.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     pcmpro       -- (char *) pointer to the name (if known) of the cmpro
 *                     being expanded.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     pCompilerErrors -- FILE The output file to which error messages are to be 
 *                   written.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

void cmpUserError( char *functionName, char *msg )
{
    fprintf( ufmCONTROLS.pCompilerErrors, "%s: %s\n", functionName, msg );
	ufmCONTROLS.CompilerFailure = 1;

} /* end cmpUserError */
/*****************************************************************************
 * 
 * NAME: cmpStubNotice
 *
 * DESCRIPTION: Notifies pgmr of the existenc of stubs. 
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS -- 
 *	
 * REVISION HISTORY:
 *   09/25/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

void cmpStubNotice( char *functionName, char *msg )
{
	if( !IDENTIFY_STUBS ) return;

    fprintf( ufmCONTROLS.pDebugFile,"%s: STUB - %s\n", functionName, msg );
	return;

} /* end cmpStubNotice */

/* --------------------------------------------------------------------------*/
/*****************************************************************************
 * 
 * NAME: cmpBuildID_LUT
 *
 * DESCRIPTION: Builds the LUT containing identifiers and their substitution
 *		strings.
 *
 * PARAMETERS:  
 *     pObj -- OBJECT Pointer to an ODL object whose name is 
 *             IDENTIFIERTRANSLATIONS.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     pIDTranslationLUT -- sltLUT Lookup table holding ODL identifiers and
 *			                their translation stirngs.  The ODL ID is used as 
 *                          the 'index' into the LUT.
 *	
 * NOTES:
 *     1. It is assumed that only OBJECTS of the expected type will be 
 *        passed to this function.
 *
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva  -- Original
 *
 ****************************************************************************/	
void cmpBuildID_LUT( OBJECT pObj )
{
	PARAMETER pParm;
	VALUE pID, pTrans;

	/* Initialize the ID Translation LUT if it hasn't been already */
	if( pIDTranslationLUT == (sltLUT) NULL ) pIDTranslationLUT = sltInitLUT();

	/* Get the first parameter and make sure it's a TranslationPairs parm. */
	if( cmpCheckIDTransSyntax(pObj) ) {
		pParm = FirstParameter( pObj );
		while( pParm )
		{
			pID = FirstValue( pParm );
			pTrans = NextValue( pID );
			while( pID )
			{
				sltAdd( pIDTranslationLUT, pID->item.value.string,
					(void *) pTrans->item.value.string );
				pID = NextValue( pTrans );
				if( pID ) pTrans = NextValue( pID );
				else pTrans = NULL;
			}
			pParm = NextParameter( pParm );
		}
	} else {
		cmpUserError("cmpBuildID_LUT", "Could not process ID translations." );
	}


	return;
	
} /* end cmpBuildID_LUT */
/*****************************************************************************
 * 
 * NAME: cmpBuildAggLUTs
 *
 * DESCRIPTION: Builds the two aggregate translation LUTs, *pAggByClassLUT
 *		and *pAggByNameLUT.  Each time an AggregateTranslation object is
 *      found in the instructions file, this function is called. It checks
 *      the syntax of the object using cmpCheckAggTransSyntax and then,
 *      depending on whether Target or TargetClass is present in the 
 *      instruction, updates the appropriate LUT.
 *
 * PARAMETERS:  
 *     pObj -- OBJECT Pointer to AggregateTranslation object parsed from the
 *			   instructions file.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     pAggByClassLUT -- sltLUT LUT holding pointers to AggregateTranslation
 *						 instructions indexed by class (e.g., LABEL or 
 *						 OBJECT ).
 *     pAggByNameLUT  -- sltLUT LUT holding pointers to AggregateTranslation
 *						 instructions indexed by name (e.g. GPOLYGON)
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE: An instruction (*pObj) is passed that contains a TARGETCLASS
 *        specification. POST: The a new entry is placed in pAggByClassLUT
 *        that's indexed by name of the target class.
 *
 * NOTES:
 *     <<description>>
 *
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva   -- Original
 *
 ****************************************************************************/
void cmpBuildAggLUTs( OBJECT pObj )
{
	PARAMETER pParm;
	char *pID;

	assert( pObj);

	if( cmpCheckAggTransSyntax(pObj) ){
		if(  ( pParm = FindParameter(pObj, "TARGETCLASS") ) != NULL  ){
	        /* If necessary initialize the Aggregate translation LUT */
			if( pAggByClassLUT == NULL ) pAggByClassLUT = sltInitLUT();

			/* Update the LUT */
			pID = pParm->first_value->item.value.string;
			sltAdd( pAggByClassLUT, pID, (void *)pObj );
		}
		else if(  ( pParm = FindParameter(pObj, "TARGET") ) != NULL  ){
	        /* If necessary initialize the Aggregate translation LUT */
			if( pAggByNameLUT == NULL ) pAggByNameLUT = sltInitLUT();

			/* Update the LUT */
			pID = pParm->first_value->item.value.string;
			sltAdd( pAggByNameLUT, pID, (void *) pObj);

		}else {
			/* Bad instruction, but a programming error: syntax should have 
			 * been checked earlier!!!
			 */
			cmpFatalError( "cmpBuildAggLUTs", "Bad Agg translation instruction." );
		}

	} else {
		/* This instructions failed the syntax check...*/
		char msg[256]; /* temporary message buffer */

		/* Get identifier for the error message: */
		if(  ( pParm = FindParameter(pObj, "TARGETCLASS") ) != NULL  )
			pID = pParm->first_value->item.value.string;
		else if(  ( pParm = FindParameter(pObj, "TARGET") ) != NULL  )
			pID = pParm->first_value->item.value.string;
		else pID = strdup("No Target/TargetClass");
		
		sprintf( msg, "Bad AggregateTranslation instruction: %s", pID );
		cmpUserError("cmpBuildAggLUTs", msg );
	}

	return;

} /* end cmpBuildAggLUTs */

/*****************************************************************************
 * 
 * NAME: cmpBuildMacroLUT
 *
 * DESCRIPTION: Builds the LUT containing macros and their substitution
 *		strings.
 *
 * PARAMETERS:  
 *     pObj -- OBJECT Pointer to an ODL object whose name is 
 *             MACRODEFINITIONS.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     pIDTranslationLUT -- sltLUT Lookup table holding  user-defined macros
 *                          and their translation stirngs. The macro name
 *			                is used as the 'index' into the LUT.
 *	
 * PRE- & POST-CONDITIONS:
 *     <<description of conditions>>
 *
 * NOTES:
 *     1. It is assumed that only OBJECTS of the expected type will be 
 *        passed to this function.
 *
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/	
void cmpBuildMacroLUT( OBJECT pObj )
{
	PARAMETER pParm;
	VALUE pMacro, pTrans;

	/* Initialize the Macro Definitions LUT if it hasn't been already */
	if( pMacroLUT == (sltLUT) NULL ) pMacroLUT = sltInitLUT();

	/* If the syntax checks add the macro def to the LUT */

	if( cmpCheckMacroDefSyntax(pObj) ) {
		pParm = FirstParameter( pObj );
		while( pParm )
		{
			pMacro = FirstValue( pParm );
			pTrans = NextValue( pMacro );
			while( pMacro )
			{
				sltAdd( pMacroLUT, pMacro->item.value.string,
					(void *) pTrans->item.value.string );
				pMacro = NextValue( pTrans );
				if(pMacro ) pTrans = NextValue( pMacro );
				else pTrans = NULL;
			}
			pParm = NextParameter( pParm );
		}
	} else {
		cmpUserError("cmpBuildMacroLUT", "Could not process macro definitions." );
	}

	return;
	
} /* end cmpBuildMacroLUT */
/*****************************************************************************
 * 
 * NAME: cmpBuildParLUTs
 *
 * DESCRIPTION: Builds the two parameter translation LUTs, *pParByClassLUT
 *		and *pParByNameLUT.  Each time an ParameterTranslation object is
 *      found in the instructions file, this function is called. It checks
 *      the syntax of the object using cmpCheckParTransSyntax and then,
 *      depending on whether Target or TargetClass is present in the 
 *      instruction, updates the appropriate LUT.
 *
 * PARAMETERS:  
 *     pObj -- OBJECT Pointer to ParameterTranslation object parsed from the
 *			   instructions file.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     pParByClassLUT -- sltLUT LUT holding pointers to ParameterTranslation
 *						 instructions indexed by class (e.g., ATTRIBUTE or
 *                       POINTER).
 *     pParByNameLUT  -- sltLUT LUT holding pointers to ParameterTranslation
 *						 instructions indexed by name 
 *	
 * PRE- & POST-CONDITIONS:
 *     <<description of conditions>>
 *
 * NOTES:
 *     <<description>>
 *
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/
void cmpBuildParLUTs( OBJECT pObj )
{
	PARAMETER pParm;


	if( cmpCheckParTransSyntax(pObj) ){
		if(  ( pParm = FindParameter(pObj, "TARGETCLASS") ) != NULL  ){
	        /* If necessary initialize the Parameter translation LUT */
			if( pParByClassLUT == NULL ) pParByClassLUT = sltInitLUT();

			/* Update the LUT */
			sltAdd( pParByClassLUT, pParm->first_value->item.value.string,
				(void *)pObj );
		}
		else if(  ( pParm = FindParameter(pObj, "TARGET") ) != NULL  ){
	        /* If necessary initialize the Parameter translation LUT */
			if( pParByNameLUT == NULL ) pParByNameLUT = sltInitLUT();

			/* Update the LUT */
			sltAdd( pParByNameLUT, pParm->first_value->item.value.string, 
			(void *) pObj);
		} else cmpFatalError( "cmpBuildParLUTs", "Failed to process a"
			" parameter by name instruction." );
	}
	return;

} /* end cmpBuildParLUTs */

/*****************************************************************************
 * 
 * NAME: cmpBuildClassLUT
 *
 * DESCRIPTION: Builds the class definition LUT, *pClassDefLUT.
 *		Each time a ComponentClassDefinition object is
 *      found in the instructions file, this function is called. It checks
 *      the syntax of the object using cmpCheckClassDefSyntax and then,
 *      updates pClassDefLUT.
 *
 * PARAMETERS:  
 *     pObj -- OBJECT Pointer to ComponentClassDefinition object parsed from
 *			   the instructions file.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     pClassDefLUT   -- sltLUT LUT holding pointers to ComponentClass-
 *						 Definitions indexed by class name.
 *	
 * PRE- & POST-CONDITIONS:
 *     <<description of conditions>>
 *
 * NOTES:
 *     <<description>>
 *
 * REVISION HISTORY:
 *   09/25/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/
void cmpBuildClassLUT( OBJECT pObj )
{
	PARAMETER pParm;
	char *pID;

	if( cmpCheckClassDefSyntax(pObj) ){
		pParm = FindParameter(pObj, "CLASSNAME");
		assert( pParm );

	    /* If necessary initialize the Parameter translation LUT */
	    if( pClassDefLUT == NULL ) pClassDefLUT = sltInitLUT();
		assert( pClassDefLUT );

		/* Update the LUT */
		sltAdd( pClassDefLUT, pParm->first_value->item.value.string,
				(void *)pObj );
	} else {
		/* This instruction failed the syntax check...*/
		char msg[256]; /* temporary message buffer */

		/* Get identifier for the error message: */
		if(  ( pParm = FindParameter(pObj, "CLASSNAME") ) != NULL  )
			pID = pParm->first_value->item.value.string;
		else pID = strdup("No ClassName");
		
		sprintf( msg, "Bad class definition: %s", pID );
		cmpUserError("cmpBuildAggLUTs", msg );
	}

	return;

} /* end cmpBuildClassLUT */

/*****************************************************************************
 * 
 * NAME: cmpAssignmentProcessor
 *
 * DESCRIPTION: This function processes an AssignmentSpec from the 
 *		Instructions File (IF). It returns a report node with information
 *      as per the given instruction.
 *
 * PARAMETERS:  
 *		pInst -- OBJECT The assignment specification OBJECT from the IF.
 *		pNode -- (struct ODL_Node *) Pointer to the ODL_Node structure.
 *
 * RETURN VALUE: (REP_NODE) a pointer to a report node of type cmpDATA.
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. Asserts that pNode and pInst are nonNULL and that a new report 
 *        node struct was successfully allocated on the heap.
 *     2. There are also asserts regarding the two required components of
 *        the AssignmentSpec instruction. These asserts should always pass
 *        because the instruction's syntax should have been checked earlier.
 *     3. RIGHTSIDE's value is a sequence (possibly one) of component
 *        references. Each component refereence may be to one or more
 *        ODL parameters. The returned report node will have a list of strings
 *        representing all of the values of all of those parameters.
 *
 * REVISION HISTORY:
 *   09/29/98  Ken Jarva   -- Original.
 *   10/05/98  Ken Jarva   -- Removed code that altered the current component
 *                            stack.
 *   01/12/99  Ken Jarva   -- Removed the restriction that allowed only one
 *                            parameter in the right-hand list spec.
 *   01/21/99  Ken Jarva   -- RIGHTSIDE now can take a scalar string or a
 *                            sequence of strings.
 *
 ****************************************************************************/

REP_NODE cmpAssignmentProcessor( OBJECT pInst, 
									   struct ODL_Node *pNode )
{
	REP_NODE pRN; /* Pointer to the report node that'll be returned. */
	lstLIST pValList = NULL; /* the lstList that'll be attached to the
							  * returned report node. */
	lstLIST pCompRefsList = NULL; /* A list of one or more component-list
								   * strings from RIGHTSIDE. */
	char *pRighty = NULL, /* The RigthSide specification from the instruction */
		 *pLefty = NULL,  /* The LeftSide specification from the instruction */
		 *pTrans = NULL;  /* The pLefty string after macro expansion. This will 
						   * be the name for the report node. */

	PARAMETER pParm = NULL;
	struct ODL_Node *pParmNode = NULL;

	/* pNode is the Current Component: */
	assert( pNode );

	/* Create a new report node and mark it as DATA */
	pRN = malloc( sizeof( *pRN) );
	assert( pRN );
    pRN->itemKind = cmpDATA;

	/* Get the LeftSide specification from the instruction: */
	assert( pInst );
	pParm = FindParameter( pInst, "LEFTSIDE" );
	assert( pParm );
	/* len = strlen( pLefty ); */
	pLefty = odl_stringValue(  pParm );

	/* The LeftSide spec may contain macros, so expand it: */
	pTrans = macMacroExpander( pLefty );
    assert( pTrans );
	pRN->pName = pTrans;

	/*Get the RightSide specification from the instruction: */
	pParm = FindParameter( pInst, "RIGHTSIDE" );
	assert( pParm );
	pParmNode = odlMakeNode( PARAMETER_NODE, pParm ); /* converts to
				* a PARAMETER to a ODL_Node pointer. */

	/* Get all the values associated with the RIGHTSIDE parameter; notice
	 * that each of those values should be a component list of ODL parameter
	 * references: */
	pCompRefsList = odlGetValueList( pParmNode );
	assert( pCompRefsList );
	pValList = lstInitList(); /* combined list of the values of all
							   * the parameters from the right side spec. */
	pRighty = lstPluckFirst( pCompRefsList ); /* the first component list */
	while( pRighty )
	{
		/* Expand the refernece and get a value list: */
		lstLIST pRightList; /* Temporary list of ODL references from ListExpander
		                     * there sholud be at least one. */

		pRightList = ListExpander( *pNode, pRighty );
		if( lstNumEl(pRightList) > 0 ){
			/* Make one long list of all the values from each of the listed
			 * parameters: */
			struct ODL_Node *pParmNode;
			pParmNode = lstPluckFirst( pRightList );
			while( pParmNode ) {
			        void *(*pDestructor)( void *) = NULL;
				lstLIST pPVList; /* list of values for one parameter */
				pPVList = odlGetValueList( pParmNode );
				lstAppendTo(pValList, pPVList);
				if( !pValList || lstNumEl(pValList) == 0 ){
					/* We've got an empty value list */
					char msg[256];
					sprintf( msg, "%s doesn't have a value list.", odlComponentName(pNode) );
					cmpUserError( "cmpAssignmentProcessor", msg );
				}
			    	pPVList = (lstLIST) lstDestructor( pPVList, pDestructor );
				pParmNode = lstPluckFirst( pRightList );
			}
		}
		lstFreeList( pRightList );
		pRighty = lstPluckFirst( pCompRefsList );
	}
	lstFreeList( pCompRefsList );

	if( !pValList ) pValList = lstInitList(); /* returning an empty list */
	pRN->pItem = pValList;

	return pRN;

}/* end cmpAssignmentProcessor */

/*****************************************************************************
 * 
 * NAME: cmpPass2
 *
 * DESCRIPTION: Builds the report structure based upon the instructions and
 *		the input ODL file.
 *
 * PARAMETERS:  
 *     pReport -- REP* Pointer to the top node of a report structure.
 *     pRoot   -- OBJECT pointer to the root node of the ODL tree upon which
 *                the output report is to be based.
 *
 * RETURN VALUE:
 *     cmpPass2 -- int The return value indicates the status of the 2nd pass
 *			compilation. 1 --> Success; 0 --> failure.
 *     
 * GLOBAL VARIABLES: 
 *	    ufmCONTROLS -- the structure containing user options and other controls.
 *      pCCStack    -- the current component stack. The stack only exists for
 *                     the duration of this function.
 *	
 * NOTES:
 *     1. The current component stack get initialized in this function.
 *     2. The rerport structure gets initialized in this function.
 *     3. Asserts that sufficient space exists on the heap.
 *
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva -- Original
 *   10/05/98  Ken Jarva -- Fixed bug in logic that frees the current compo
 *                          stack.
 *
 ****************************************************************************/
       
int cmpPass2( struct REP *pReport, OBJECT pRoot )
{

	ODLNodePtr *pODLNode;
	int success = 1, failure = 0;
	int /* Debuggery controls... */
		printReportStruct = 0; /* Set to 1 to print the report structure */
	AGGREGATE pHnode = NULL; 

    /* Initialize the report structure: */
	pReport->pTitle = ufmCONTROLS.pTitle;
	pReport->pComment = ufmCONTROLS.pComment;
	pReport->pLevelOne = (lstLIST) NULL;

	/* Initialize the current component stack: */
	pCCStack = lstInitList();

	/* Create a buffer to store the header information */

        pHnode = pRoot->first_child;
	 
        if((ufmCONTROLS.DHeader == 0)||(pHnode->comment == NULL))
      
	   {
	 
	    pReport->pHeader=malloc(10);
	    assert(pReport->pHeader);
	    strcpy(pReport->pHeader," ");
	  }
	 else
	  {

	    pReport->pHeader=malloc(strlen(pHnode->comment)+1);
	    assert(pReport->pHeader); 
	    strcpy(pReport->pHeader,pHnode->comment);
	  }

	/* Create a component pointer structure & point it at ROOT: */
	pODLNode = malloc( sizeof(ODLNodePtr) );
	assert( pODLNode );
	pODLNode->node_kind = AGGREGATE_NODE;
	pODLNode->node_type.aggregate = pRoot;

	pReport->pLevelOne = cmpTranslationProcessor( pODLNode );

	/* Debuggery... */
	if( printReportStruct )
		cmpOutputRepStruc( pReport->pLevelOne, 0 );

	/* Free the current component stack: */
	{
		struct ODLNodePtr *pODLNode;
		lstStopIteration( pCCStack );
		pODLNode = lstIterator( pCCStack );
		if( lstNumEl( pCCStack ) != 0 ) cmpFatalError( "cmpPass2", 
			"The current component stack has not been flushed." );
		while( pODLNode ) {
			if( pODLNode ) free( pODLNode );
			pODLNode = lstIterator( pCCStack );
		}
		lstStopIteration( pCCStack );
		pCCStack = lstFreeList( pCCStack );
	}

	/* Inform caller of status: */
	if( ufmCONTROLS.CompilerFailure == 0 )
		return success;
	else return failure;

} /* end cmpPass2 */
/*****************************************************************************
 * 
 * NAME: cmpTranslationProcessor
 *
 * DESCRIPTION: Find and apply the approriate instruction to the given ODL
 *     component.
 *
 * PARAMETERS:  
 *     pNode -- (struct ODL_Node *) The ODL component descriptor for the node that
 *              is to be processed.
 *
 * RETURN VALUE:
 *     cmpTranslationProcessor -- (lstLIST)  A list of report nodes
 *			created in accordance with the instruction associated with
 *          *pNode.
 *     
 * GLOBAL VARIABLES:
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE: pNode has an associated instruction that says to ignore this
 *        component. POST: A list having a single element is returned. That
 *        single element has an associated report node that has been marked
 *        cmpIGNORE. Its item pointer (pItem) has been set to NULL. It's
 *        pName points at a string containing the name of the component or
 *        its translation string.
 *     2. PRE: The associated instruction says to replace the component with its
 *        substructure. POST: the node for the current component is freed; now
 *        instead of a list having one component being returned, a list having an
 *        arbitrary number of components (that is, one representing the sub-
 *        structure of the current component) will be returned.
 *     3. PRE: The given component is self-referential; that is, it already
 *        exists on the current component stack, implying that its instruction
 *        directly or indirectly causes it to be translated again. POST: An
 *        error message is output and the component is ignored.
 *
 * NOTES:
 *     1. This is the only function that pushes and pops the current component
 *        stack.
 *     2. The translation processor wants to build a report node with a sublist
 *        for every component it processes. The node tells the name (possibly
 *        translated) of the component and gives a sublist of report nodes that
 *        reflect the results of the applicable instruction having been applied
 *        to the current component.  This is a true for ODL parameters as it is
 *        for aggregates.  For most parameters and for some aggregates (e.g.,
 *        ROOT) we don't want the name of the component to be listed. For those
 *        components the ReplaceWithSubstructure instruction is used. When such
 *        an instruction is encountered, this function replaces the report node
 *        for the current component with its sublist.
 *
 * REVISION HISTORY:
 *   09/30/98  Ken Jarva -- Original
 *   10/05/98  Ken Jarva -- Restructured: broke out code for cmpInstruction-
 *                          Processor.
 *   10/07/98  Ken Jarva -- Protected agains self-referential instuctions.
 *   12/21/98  Ken Jarva -- Proteected against possiblitiy of nonexistent
 *							paramters, IGNORECOMPONENT & 
 *							REPLACEWITHSUBSTRUCTURE.
 *
 ****************************************************************************/
static lstLIST cmpTranslationProcessor(struct ODL_Node *pNode)
{
	lstLIST pList, pTmpList;
	REP_NODE pRepNode;
	OBJECT pInst; /* points at the instruction in the IF tree for
				   * *pNode ); */
    PARAMETER pParm;
	struct ODL_Node *pWasCurrentCompo, *pTestNode;
	char *pValueStr = NULL;

	/* Protect against self-referential instructions: check each
	 * element of the current component stack to make sure that the given
	 * componenet is not on it: */
	lstStopIteration( pCCStack );
	pTestNode = lstIterator( pCCStack );
	while( pTestNode ){
		if( odlAreEqualNodes(pTestNode, pNode) ) {
			/* There's a self-referential instruction...*/
			cmpUserError( "cmpTranslationProcessor", 
				"Self-referential instruction." );
			return lstInitList();
		}
		pTestNode = lstIterator( pCCStack );
	}
	lstStopIteration( pCCStack );
	
	/* Push the given ODL_Node onto the current component stack: */
	lstAddFirst( pCCStack, (void *) pNode );

	/* Initialize the list that'll contain the current node (unless
	 * its marked ReplaceWithSubstructure). */
	pList = lstInitList();

	/* Create the report node for the current component and place it
	 * in the lstList. Usually this will be the only element in the
	 * list that's returned to the caller...*/
	pRepNode = malloc( sizeof(struct reportNode) );
	assert( pRepNode );
	lstAddLast( pList, (void *) pRepNode );

	/* Find the appropriate instruction for the current component; */
	pInst = cmpFindInstruction( pNode );
	assert( pInst );

	/* Determine the proper identification string: */
	pRepNode->pName = cmpComponentID( pNode, pInst ); 

	/* Determine whether or not to ignore this component and any
	 * substructure it might have */
	pParm = FindParameter( pInst, "IGNORECOMPONENT" );
	if( pParm ) pValueStr = odl_stringValue( pParm );
	if( pValueStr && strToUpStrcmp( pValueStr, "YES") == 0 ) {
		/* This component is to be ignored...*/
		pRepNode->itemKind = cmpIGNORE;
		pRepNode->pItem = (void *) NULL;

		/* Pop the current component stack: */
        pWasCurrentCompo = lstPluckFirst( pCCStack );
		if( pWasCurrentCompo ) free( pWasCurrentCompo );

		return pList;
	}

	/* Determine whether the current component is to get replaced with
	 * it's substructure. If so, do it...*/
	pValueStr = NULL;
	pParm = FindParameter( pInst, "REPLACEWITHSUBSTRUCTURE" );
	if( pParm ) pValueStr = odl_stringValue( pParm );
	if( pValueStr && strToUpStrcmp( pValueStr, "YES") == 0 ) {
		/* This component is to be replaced with its substructure...*/
		pRepNode = lstPluckFirst( pList ); /* Lose the node for the current comp. */
		if( pRepNode )free( pRepNode );

		/* Carry out the instruction on the given component: */
		pTmpList = cmpInstructionProcessor( pInst, pNode );
		lstAppendTo( pList, pTmpList );
		pTmpList = lstFreeList( pTmpList );
	 } else {
		pRepNode->itemKind = cmpSUBLIST;
		/* Carry out the instruction on the given component: */
		pRepNode->pItem = cmpInstructionProcessor( pInst, pNode );
	 }

	/* Pop the current component stack */
    pWasCurrentCompo = lstPluckFirst( pCCStack );
	if( pWasCurrentCompo ) free( pWasCurrentCompo );

	return pList;

} /* end cmpTranslationProcessor */

/*****************************************************************************
 * 
 * NAME: cmpInstructionProcessor
 *
 * DESCRIPTION: Applies the given instruction (*pInst) to the given component
 *              (*pNode).
 *
 * PARAMETERS:  
 *     pInst -- OBJECT A pointer to an instruction for translating the given
 *              component.
 *     pNode -- (struct ODL_Node *) A pointer to a descriptor of an ODL 
 *              component -- the one to which the translation is to be applied.
 *
 * RETURN VALUE:
 *     cmpInstructionProcessor -- lstLIST A pointer to a list of report nodes
 *                                that represent the translation of the given
 *                                component.
 *     
 * GLOBAL VARIABLES: None
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE: Either or both pInst & pNode are NULL. POST: The program 
 *        terminates via assert().
 *     2. PRE: An instruction is given having one or more of the 
 *        recognized specifications. POST: A lstList of report nodes is 
 *        constructed having one or more members for each of the specifications.
 *     3. PRE: An instruction is given having a component that is NOT one of
 *        the recognized specification. POST: A lstList of report nodes is
 *        constructed ignoring the unknown specification.
 *     4. PRE: An instruction is given having NO recognizable specification. POST:
 *        an empty lstList is returned.
 *
 * NOTES:
 *     1. It is assumed that the instruction is syntactically correct.
 *
 * REVISION HISTORY:
 *   10/07/98  Ken Jarva  -- Original
 *
 ****************************************************************************/
lstLIST cmpInstructionProcessor( OBJECT pInst, struct ODL_Node *pNode )
{
	lstLIST pRepList;
	OBJECT pObj;

	assert( pInst );
	assert( pNode );

	pRepList = lstInitList();

	 /* Step through each of the components of the instruction and perform
	  * the necessary processing: */
	pObj = NextObject( pInst ); /* points at the next component of the 
								 * translation instruction */
	while( pObj ){
		if( strcmp(pObj->name, "LISTSPEC") == 0 ) {
			lstLIST pTmpList;
			char *pListSpec;
			pListSpec = cmpListSpec( pObj );
			pTmpList = cmpListProcessor( pListSpec, pNode );
			lstAppendTo( pRepList, pTmpList );
			pTmpList = lstFreeList( pTmpList );
		}else if( strcmp(pObj->name, "TABLESPECBYCOLUMN") == 0 ) {
			REP_NODE pTmpNode;
			pTmpNode = cmpTableProcessor( pObj, pNode );
			lstAddLast( pRepList, (void *) pTmpNode);
		}else if( strcmp(pObj->name, "TABLESPECBYROW") == 0 ) {
			REP_NODE pTmpNode;
			pTmpNode = cmpTableProcessor( pObj, pNode );
			lstAddLast( pRepList, (void *) pTmpNode);
		}else if( strcmp(pObj->name, "ASSIGNMENTSPEC") == 0 ) {
			REP_NODE pTmpNode;
			pTmpNode = cmpAssignmentProcessor( pObj, pNode );
			lstAddLast( pRepList, (void *) pTmpNode );
		}
		pObj = pObj->right_sibling;
	}
	return pRepList;

} /* end cmpInstructionProcessor */
/*****************************************************************************
 * 
 * NAME: cmpListProcessor
 *
 * DESCRIPTION: Given a list specification, this function causes the spec
 *		to be expanded into a list of component references. Each of those
 *      components is then processed according to the appropriate instruction.
 *
 * PARAMETERS:  
 *     lstSpec      -- (char*) String containing references to ODL
 *                             components.
 *     pCurrentComp -- (struct ODL_Node *) Component descriptor for the 
 *                             current component.
 *
 * RETURN VALUE:
 *     cmpListProcessor  -- (lstLIST) Pointer to a lstList of report nodes
 *                                    built according to the instructions
 *                                    applied to each of the components
 *                                    inferred from the *listSpec.
 *     
 * NOTES:
 *     1. A list specification is a string giving comma-delimited references to
 *        ODL components. The list spec. may also contain user-defined and
 *        predefined macros, but once those macros are resolved the string
 *        should just be a comma-delimited list of component references.
 *     2. The list spec. is expanded by ListExpander. The result is a lstList
 *        of component descriptors (struct ODL_Nodes) that correspond to the 
 *        component references given in the list spec.
 *
 * REVISION HISTORY:
 *   10/07/98  Ken Jarva  --  Original
 *
 ****************************************************************************/
static lstLIST cmpListProcessor( char *listSpec, 
								 struct ODL_Node *pCurrentComp )
{
	lstLIST pCompRefList, pList;
	ODLNodePtr *pNode;

	pCompRefList = ListExpander( *pCurrentComp, listSpec );

	/* Process each reference in the list and build the list of report nodes
	 *     for each component of the report that is to appear at this level: */
	pList = lstInitList(); /* Lists all the report nodes at this level */
	pNode = lstIterator( pCompRefList ); /* next ODL reference */
	while( pNode ){
		lstLIST pNodeList; /* Temporary list of report nodes created by the
						    * instruction for pNode. */
		pNodeList = cmpTranslationProcessor(pNode);
		lstAppendTo( pList, pNodeList );
		lstFreeList( pNodeList );
		pNode = lstIterator( pCompRefList );
	}
	lstStopIteration( pCompRefList );

	return pList;

} /* end cmpListProcessor */
/*****************************************************************************
 * 
 * NAME: cmpListSpec
 *
 * DESCRIPTION: Extacts a list specification from a ListSpec ODL object such
 *              as may be found in an instructions file.
 *
 * PARAMETERS:
 *     pObj -- OBJECT A pointer to an ODL object representing a list spec. In
 *                    particular, it must have an ELEMENTS parameter.
 *
 * RETURN VALUE:
 *     cmpListSpec -- (char *) The string contained by the Elements parameter
 *                             of a ListSpec object.
 *     
 * GLOBAL VARIABLES: None
 *
 * NOTES:
 *     1. If pObj is NULL, the program terminates via assert.
 *     2. If *pObj doesn't have an Element parameter, the program invokes
 *        fatal error handling.  It's assumed that the syntax of *pObj has
 *        been checked previous to the call of this function.
 *
 * REVISION HISTORY:
 *   09/30/98  Ken Jarva  --  Original
 *   10/08/98  Ken Jarva  --  Doc header and error handling.
 *
 ****************************************************************************/
char *cmpListSpec( OBJECT pObj )
{
	/* Return the list specification string from a ListSpec object */
	PARAMETER pParm;

	assert( pObj );
	pParm = FindParameter( pObj, "ELEMENTS" );

	if( !pParm ) {
		cmpFatalError( "cmpListSpec", "Failed to find ELEMENTS in ListSpec." );
		return strdup(""); /* only in case of return from cmpFatalError */
	}

	return odl_stringValue( pParm);

} /* end cmpListSpec */

/*****************************************************************************
 * 
 * NAME: cmpHasClassInst
 *
 * DESCRIPTION: 	
 *		Determines whether the given ODL node is a member of a defined class
 *      and whether an insturction exists for that class. Returns a pointer
 *      to the class instruction if one's found, NULL otherwise.
 *
 * PARAMETERS:  
 *     pNode -- (struct ODL_Node *) Pointer to a component descriptor for
 *                                  which an instruction is sought.
 *
 * RETURN VALUE:
 *     cmpHasClassInst -- OBJECT A pointer to the translation instruction for
 *                        the class to which the component referenced by *pNode
 *                        belongs.
 *     
 * GLOBAL VARIABLES: None
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE: A component is given that belongs to one or more classes.
 *        POST: Class membership is recognized for the first class that
 *        includes the given component. A translation instruction
 *        specific to that class is sought from the appropriate LUT.
 *        If an instruction exists for the class, a pointer to the 
 *        instruction is returned; otherwise NULL is returned.
 *     2. PRE: A component is given that belongs to no known class.
 *        POST: NULL is returned.
 *     3. PRE: The given component belongs to a user-defined class for 
 *        which no instruction is given. POST: A warning is (optionally)
 *        issued to the user, and NULL is returned.
 *
 * NOTES:
 *     1. Asserts that pNode is nonNULL.
 *     2. If the given component is neither an aggregate nor a parameter,
 *        fatal error handling is invoked. 
 *
 * REVISION HISTORY:
 *	 09/30/98  Ken Jarva  --  Original
 *   10/07/98  Ken Jarva  --  Header & cleanup.
 *   11/13/98  Ken Jarva  --  Corrected error message.
 *   12/15/98  Ken Jarva  --  Improved the way *pClassDefLUT is searched
 *                            ( employed sltKeys() ).
 *   02/11/99  Ken Jarva  --  Fixed 2 bugs: both had '=' in conditional where
 *                            it should have been "==". 
 *
 ****************************************************************************/
static OBJECT cmpHasClassInst( struct ODL_Node *pNode )
{
	OBJECT 
		pDef,     /* Points at the ODL object that defines a particular class
		           * of component. */
		pInst = (OBJECT) NULL; /*
		           * Points at the instruction to be carried out 
	               * for *pNode. */
	char *pNextIndex; /* Points in turn to the name of each class */
	static lstLIST pIndexList = NULL; /* list of all the indexes of
							    * pClassDefLUT */

	assert( pNode );

	if( !pIndexList ) pIndexList = sltKeys( pClassDefLUT );
	/* Visit each node in the Class Definition LUT and check *pNode against
	 * the listed definitions until a match is found: */
	lstStopIteration( pIndexList );
	pNextIndex = lstIterator( pIndexList ); /* points at next index for
	                                         * the class def. LUT.*/
	while( pNextIndex ) {

		/* Extract pointer to class definition from the LUT: */
		pDef = (OBJECT) sltLookUp( pClassDefLUT, pNextIndex );

		/* Check whether *pNode is in the class described by *pDef: */
		if( cmpIsInClass(pNode, pDef) ){
			if( pNode->node_kind == AGGREGATE_NODE ){ 
				/* Have a match -- *pNode is a member of this class */
				pInst = sltLookUp( pAggByClassLUT, pNextIndex );
				if( !pInst ) {
					char msg[256];
					sprintf( msg, "Have class member but no instruction "
							      "for class %s", pNextIndex );
					cmpUserWarning( "cmpHasClassInst", msg);
				}
			} else if( pNode->node_kind == PARAMETER_NODE ){
				/* Have a match -- *pNode is a member of this class */
				pInst = sltLookUp( pParByClassLUT, pNextIndex );
				if( !pInst ) {
					char msg[256];
					sprintf( msg, "Have class member but no instruction "
							      "for class %s", pNextIndex );
					cmpUserWarning( "cmpHasClassInst", msg);
				}
			} else {
				cmpFatalError( "cmpHasClassInst", 
					"Passed component is of unknown kind." );
				return (OBJECT) NULL; /* only in case of return from fatal
									   * error handling. */
			}
			lstStopIteration( pIndexList );
			return pInst;
		}
		pNextIndex = lstIterator( pIndexList ); /* points at next index for
	                                             * the class def. LUT.*/
	}
	lstStopIteration( pIndexList );

	return (OBJECT) NULL; /* Component belongs to no known class */

} /* end cmpHasClassInst */

/*****************************************************************************
 * 
 * NAME: cmpIsInClass
 *
 * DESCRIPTION: Determines whether a given component meets the given 
 *              definition of class membership.
 *
 * PARAMETERS:  
 *     pNode -- (struct ODL_Node *) Pointer to a descriptor of the component
 *                           that's to be tested for class membership.
 *     pClassDef -- (OBJECT) Pointer to the ODL giving the requirements for
 *                           class membership.
 *
 * RETURN VALUE:
 *     cmpIsInClass -- (int) Indicator of class membership: 1 -> class member,
 *                           0 -> not a member.
 *     
 * GLOBAL VARIABLES: None
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE: 
 *
 * NOTES:
 *     1. Additional debug output can be achived by setting one or more of
 *        the "debuggery controls" below.
 *     2. This function assumes that the class definition is of the 
 *        proper syntax. If it discovers that this is not the case, fatal
 *        error processing is invoked.
 *     3. Class membership is determined as follows:
 *        
 *            o  If an EnumerationList is given, only aggregates having 
 *               names on the list can statisfy the definition.  
 *            o  If a RequiredComponents list is given, only aggregates 
 *               containing all the components named in the list can satisfy
 *               the definition.  
 *            o  If an OptionalComponents list is given only aggregates that
 *               contain only components named in the union of the 
 *               OptionalComponents and RequiredComponents lists can satisfy 
 *               the definition.  
 *            o  If neither a RequiredComponents list nor an OptionalComponents
 *               list is given, the definition will apply to all OBJECTs 
 *               or all GROUPS or all ATTRIBUTES or all POINTERS, depending on
 *               the value ComponentKind.
 *			  o  If an OptionalComponents list that has no possiblilty of
 *               providing a match is given, only aggragates having exactly the
 *               membership specified by the RequiredComponents list can be 
 *               members. If the RequiredComponents list is not given, only 
 *               aggregates having no substructure can qualify.
 *            >  In the future it may be desirable to allow a class to be
 *               defined for a collection of aggregates or of parameters.
 *
 * REVISION HISTORY:
 *   09/30/98  Ken Jarva  --  Original
 *   10/09/98  Ken Jarva  --  Documentation header & debuggery controls.
 *
 ****************************************************************************/
int cmpIsInClass( struct ODL_Node *pNode, OBJECT pClassDef)
{
	PARAMETER pParm;
	char *pClassKind = NULL, *pCompoName;
	char *pGivenCompoName; /* The name of the given component */
	struct ODL_Node *pON_Parm;
	lstLIST 
		pEnumList,         /* optional list of components to which the 
		                   * definition applies */
		pReqCompList,     /* optional list of required component the given
						   * component must have in order for the definition
						   * to apply */
		pOptCompList,     /* optional list of optional components that the
						   * given component may have an still have the 
						   * definition apply */
		pChildCompoList;  /* list of all the children of the given component */

	int /* debuggery controls */
		list_required_components = 0,
		list_optional_components = 0,
		list_actual_components   = 0,
		list_enumeration_set     = 0;

	/* Extract the name of the given component: */
	pGivenCompoName = odlComponentName( pNode );
	assert( pGivenCompoName );

	/* Determine the kinds of components that can be in the class. */
	  pParm = FindParameter( pClassDef, "COMPONENTKIND" );
	  if( !pParm ) 
		/* This should not occur; it should be detected in the syntax 
		                                                           checker...*/
			cmpFatalError( "cmpIsInClass", 
			                       "Class definition without ComponentKind." );

	  pClassKind = pParm->first_value->item.value.string;
	
	/* Check that the given component is of the kind required by the 
	                                                             class def.: */
	  if( strToUpStrcmp(pClassKind, "OBJECT") == 0 ) {
		/* This definition only applies to OBJECTs...*/
		if( !odlIsObject(pNode) ) return 0;
	  }
	  if( strToUpStrcmp(pClassKind, "GROUP") == 0 ) {
		/* This definition only applies to GROUPs...*/
		if( !odlIsGroup(pNode) ) return 0;
	  }
	  if( strToUpStrcmp(pClassKind, "ATTRIBUTE") == 0 ) {
		/* This definition only applies to PARAMETERs...*/
		if( !odlIsAttribute(pNode) ) return 0;
	  }
	  if( strToUpStrcmp(pClassKind, "POINTER") == 0 ) {
		/* This definition only applies to POINTERs...*/
		if( !odlIsPointer(pNode) ) return 0;
	  }

	/* Get the enumerated list, if it exists: */
	 pParm = FindParameter( pClassDef, "ENUMERATIONLIST" );
	if( pParm ) {
		/* Since an EnumerationList was given, only components with names
		 * on that list are candidates for inclusion within the class. */

		char *pListItem; /* used to hold strings extracted from a list. */
		int isInList = 0; /* indicator of whether a given name is in the
						   * enumerated list */

		/* Create a component descriptor for the parameter: */
		pON_Parm = odlMakeNode( PARAMETER_NODE, pParm );

		/* Get the list of values from the ENUMERATIONLIST parameter: */
		pEnumList = odlGetValueList( pON_Parm );
		assert( pEnumList );

		/* Debuggery...*/
		if( list_enumeration_set ) {
			char msg[256];
			/* Write the list to the debug output...*/
			pListItem = lstIterator( pEnumList );
			  while( pListItem ) {
				sprintf( msg, "\tpossible component: %s", pListItem );
				cmpDebugMsg( "cmpIsInClass", msg);
				pListItem = lstIterator( pEnumList );
			  }
			lstStopIteration( pEnumList );
		}

		/* Check whether the given component is listed within the
		 * enumeration...*/
		isInList = 0; /* Gets set to 1 if it's on the list */
		pListItem = lstIterator( pEnumList );
		  while( pListItem ) {
			  if( strToUpStrcmp(pGivenCompoName, pListItem) == 0 ) {
				  isInList = 1;
				  break;
			  }
			  pListItem = lstIterator( pEnumList );
		  }
		  if( !isInList ) return 0; /* it's not amoung those listed */
		lstStopIteration( pEnumList );
	}

	/* Get the list of required components, if it exists: */
	pParm = FindParameter( pClassDef, "REQUIREDCOMPONENTS" );
	if( pParm ) {
		/* Create a component descriptor for the parameter */
		pON_Parm = odlMakeNode( PARAMETER_NODE, pParm );

		/* Get the list of values from the REQUIREDCOMPONENTS parameter: */
		pReqCompList = odlGetValueList( pON_Parm );

		/* Debuggery...*/
		if( list_required_components ) {
		    char *pReqComp;
			char msg[256];
			pReqComp = lstIterator( pReqCompList );
			while( pReqComp ) {
				sprintf( msg, "\trequired compo: %s", pReqComp );
				cmpDebugMsg( "cmpIsInClass", msg);
				pReqComp = lstIterator( pReqCompList );
			}
			lstStopIteration( pReqCompList );
		}
	} else { /* Their are no required components */
		pReqCompList = lstInitList(); /* Empty list */
	}

	/* Get the list of optional components, if it exists: */
	pParm = FindParameter( pClassDef, "OPTIONALCOMPONENTS" );
	if( pParm ) {

		/* Make a component descriptor for the OPTIOALCOMPOENTS parm: */
		pON_Parm = odlMakeNode( PARAMETER_NODE, pParm );

		/* Get the values of the parm.: */
		pOptCompList = odlGetValueList( pON_Parm );

		/* Debuggery...*/
		if( list_optional_components ) {
			char *pOptComp;
			char msg[256];
			pOptComp = lstIterator( pOptCompList );
			while( pOptComp ) {
				sprintf( msg, "\toptional compo: %s", pOptComp );
				cmpDebugMsg( "cmpIsInClass", msg);
				pOptComp = lstIterator( pOptCompList );
			}
			lstStopIteration( pOptCompList );
		}
	} else { /* Their are no optional components */
		pOptCompList = (lstLIST) NULL; /* No list */
	}

	/* Get the list of component children of the ODL node */
	pChildCompoList = odlChildComponentsList( pNode );

	/* Debuggery...*/
	if( list_actual_components ) { 
		char msg[256];
		char *pChild;
		pChild = lstIterator( pChildCompoList );
		while( pChild ) {
			sprintf( msg, "\tchild component: %s", pChild );
			cmpDebugMsg( "cmpIsInClass", msg);
			pChild = lstIterator( pChildCompoList );
		}
		lstStopIteration( pChildCompoList );
	}

	/* Check that all the required components are in the list
	 * of children. */
	pCompoName = lstIterator( pReqCompList ); /* required component */
	  while( pCompoName ) {
		if( !lstIsStrInList( pChildCompoList, pCompoName ) )
			/* missing a required component */
			return 0;
		pCompoName = lstIterator( pReqCompList );
	  }
	lstStopIteration( pReqCompList );

	/* If we got this far, we've probably got a class member 
	 * Now if a list of optional components was given, every child's name
	 * must appear in the union of the optional and required lists. */
	if( pOptCompList ) {

		/* Create the union by expanding pReqCompList:*/
		lstAppendTo( pReqCompList, pOptCompList ); /* create the union */

		/* Check the actual components found against the contents of the
		 * expanded pReqCompList; */
		pCompoName = lstIterator( pChildCompoList );
		  while( pCompoName ) {
			if( !lstIsStrInList(pReqCompList, pCompoName) )
			  /* there's a component that's not in the union */
			  return 0;
		    pCompoName = lstIterator( pChildCompoList );
		  }
		lstStopIteration( pChildCompoList );
	}

	return 1;

} /* end cmpIsInClass */

/*****************************************************************************
 * 
 * NAME: cmpFindInstruction
 *
 * DESCRIPTION: Finds the appropriate instruction for a given ODL component.
 *
 * PARAMETERS: 
 *     pNode -- (struct ODL_Node *) pointer to a component descriptor. It is
 *                                  component for which a translation 
 *                                  instruction is sought.
 *
 * RETURN VALUE:
 *     cmpFindInstruction -- (OBJECT) A pointer to the instruction ODL.
 *     
 * GLOBAL VARIABLES: None
 *
 * NOTES:
 *     1. This function checks directly for instructions that may have been
 *        specified for the given component by name.
 *     2. If no such instruction exists, this function calls cmpGetClass-
 *        Instruction() in order to get an instruction for the most
 *        specific class to which the component belongs.
 *     3. Since all components fall into at least one class covered by one of
 *        the default instructions, cmpGetClassInstruction should always return
 *        a pointer to an instruction.
 *
 * REVISION HISTORY:
 *   10/07/98  Ken Jarva  --  Added header.
 *
 ****************************************************************************/
OBJECT cmpFindInstruction( struct ODL_Node *pNode )
{
	char *pCompName;
	OBJECT pInst;

	assert( pNode );

	if(  pNode->node_kind == AGGREGATE_NODE ){
		/* Get the name of the aggregate: */
		pCompName = odlComponentName( pNode );
		pInst = sltLookUp( pAggByNameLUT, pCompName );
		if( pInst ) return pInst;
	}else if( pNode->node_kind == PARAMETER_NODE ){
		/* Get the name of the parameter: */
		pCompName = odlComponentName( pNode );
		pInst = sltLookUp( pParByNameLUT, pCompName );
		if( pInst ) return pInst;
	}
	
	return cmpGetClassInstruction( pNode );

} /* end cmpFindInstruction */

/*****************************************************************************
 * 
 * NAME: cmpGetClassInstruction
 *
 * DESCRIPTION: Determines the applicable instruction based upon class
 *              membership of the given component.
 *
 * PARAMETERS:
 *     pNode -- (struct ODL_Node) Points at component descriptor for the
 *				component for which the class instruction is sought.
 *
 * RETURN VALUE:
 *     cmpGetClassInstruction -- (OBJECT) Pointer to a translation instruction
 *              that's to be applied to the given component.
 *     
 * GLOBAL VARIABLES: None
 *
 * NOTES:
 *     1. Additional debug info can be had via the Debuggery controls. See
 *        below.
 *
 * REVISION HISTORY:
 *   10/07/98  Ken Jarva  --  Original
 *   10/14/98  Ken Jarva  --  Updated documentation & extended the processing
 *                            for general instructions.
 *   10/26/98  J.D. Stroup -- Replaced the lookups for "GROUP", "OBJECT",
 *                            "ATTRIBUTE", and "POINTER" with lookups for
 *                            "AGGREGATE" and "PARAMETER" instead.  Also added
 *                            the lookups for the default instructions:
 *                            "ufmROOT", "ufmAGGREGATE", and "ufmPARAMETER".
 *   11/13/98  Ken Jarva  --  Fixed indentation problems, added inline doc.
 *
 ****************************************************************************/
OBJECT cmpGetClassInstruction( struct ODL_Node *pNode )
{
	OBJECT pInst = NULL; /* Points at the instruction to be returned */
	int /* Debuggery controls */
		reportFoundClass = 0; /* Set to 1 if you want reprot classes
                                       * with instructions */

	/* Determine whether the given component is a member of a user-defied
	 * class having a translation instruction. */
	pInst = cmpHasClassInst( pNode );
	if( pInst ) {
		/* This is the expected result. We just return a pointer to the
		 * found instruction...*/

		/* Debuggery */
		if( reportFoundClass ) {
			char msg[256];
			sprintf( msg, "Have class instruction for %s", odlComponentName(pNode) );
			cmpDebugMsg( "cmpGetClassInstruction", msg );
		}

		/* Have an instruction for a class to which *pNode belongs...*/
		return pInst;
	}else if(  pNode->node_kind == AGGREGATE_NODE ){
		/* The given component does not belong to any of the user-defined 
		 * classes, so get the appropriate general class of instruction...*/
		char *compname = odlComponentName (pNode);

		/* Is this the root node?  If so, get the default instruction for it.
         * Otherwise get the general instruction for an AGGREGATE, whether it's
         * user-defined or the default. */

		if (strcmp (compname, "ROOT") == 0)
			pInst = sltLookUp (pAggByNameLUT, "ufmROOT");
		else {
			pInst = sltLookUp( pAggByClassLUT, "AGGREGATE" );
			if (! pInst)
				/* No user-defined instruction for agg; need the default 
				 * instruction...*/
				pInst = sltLookUp( pAggByClassLUT, "ufmAGGREGATE" );
		}

		if( pInst ) return pInst;

	}else if( pNode->node_kind == PARAMETER_NODE ){
		/* Get the general instruction for a PARAMETER, whether it's user-
		 * defined or the default */

		/* Get user-defined translation instruction for PARAMETER if it exists: */
		pInst = sltLookUp( pParByClassLUT, "PARAMETER" );
		/* If is doesn't exist, get the default translation instruction: */
		if (! pInst)
			pInst = sltLookUp( pParByClassLUT, "ufmPARAMETER" );

			if( pInst ) return pInst;
	}
	
	{  /* Couldn't find an instruction. This should never happen */
		char msg[256];
		sprintf( msg, "No instruction for %s", odlComponentName(pNode) );
		cmpFatalError( "cmpGetClassInstruction", msg );
		return (OBJECT) NULL; /* only incase of return from fatal err processing */
	}

} /* end cmpGetClassInstruction */

/*****************************************************************************
 * 
 * NAME: cmpComponentID
 *
 * DESCRIPTION: This function determines the appropriate identifier for the
 *              given component. It's first preferance is the optional
 *              TranslationString given in the given translation instruction;
 *              next it will check the ID LUT for a user-specifed ID 
 *              translation. If neither of those two methods produce a 
 *              result, the name of the component is returned.
 *
 * PARAMETERS: 
 *     pNode -- (struct ODL_Node *) Points at a component descriptor for the
 *                                  component for which the ID is needed.
 *     pInst -- (OBJECT) Points at the translation instruction being applied
 *                       to *pNode.
 *
 * RETURN VALUE:
 *     cmpComponentID -- (char *) The ID to be reported for the given 
 *                                component.
 *     
 * GLOBAL VARIABLES: None
 *	
 *
 * NOTES:
 *     1. Any macros in the TranslationString from the given instruction
 *        are expanded. 
 *
 * REVISION HISTORY:
 *   09/30/98  Ken Jarva  --  Original
 *   10/14/98  Ken Jarva  --  Doc header & macro expansion of the translation
 *                            string.
 *
 ****************************************************************************/
char *cmpComponentID( struct ODL_Node *pNode, OBJECT pInst )
{

	char *pIdentifier, *pName;
	PARAMETER pParm;

	/* First choice: the TranslationString attribute from the associated
	 * instruction: */
	pParm = FindParameter( pInst, "TRANSLATIONSTRING" );
	if( pParm ) {
		char *pTransStr;
		pTransStr = pParm->first_value->item.value.string;

		/* Expand any macros...*/
		pTransStr = macMacroExpander( pTransStr );

		return pTransStr; /* TranslationString with expanded macros */
	}

	/* Next try for a user-specified translation string from the LUT: */
	pName = odlComponentName( pNode );
	pIdentifier = sltLookUp( pIDTranslationLUT, pName );

	if( pIdentifier ) {
		pIdentifier = macMacroExpander( pIdentifier );
		return pIdentifier; /* The user-specified trans. str */
	}
	/* Finally, just return the name of the component: */
	else {
		return pName; /* Just the name of the given component */
	}

} /* end cmpComponentID */
/*****************************************************************************
 * 
 * NAME: cmpCheckClassDefSyntax
 *
 * DESCRIPTION: Checks the syntax of a class definition instruction.
 *
 * PARAMETERS:  
 *     pAgg -- (AGGREGATE) Pointer to the instruction. 
 *
 * RETURN VALUE:
 *     cmpCheckIDTransSyntax -- (int) Success = 1, Failure = 0.
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. This implementation is pretty generic: there's little here that's
 *        different from the checkers for other instructions. It's implemented
 *        as a separate function in case some time in the future special 
 *        reporting is required for problems in this kind of instruction.
 *     2. This function assumes that there is an unambiguous definition for
 *        the associated instruction.
 *
 * REVISION HISTORY:
 *   10/07/98  Ken Jarva  --  Original
 *   11/16/98  Ken Jarva  --  Doc. header & code cleanup.
 *   12/21/98  Ken Jarva  --  Calls to improved versions of syn functions.
 *
 ****************************************************************************/
static int cmpCheckClassDefSyntax( AGGREGATE pAgg )
{	
	/* returns 1 for success, 0 for failure. 
	 * Global variable: pCompilerOut.
	 */
	int Success = 1, Failure = 0;
	synAGGRESULT pResult = NULL; /* The result of the syntax check */
	synAGGDEF pDef = NULL; /* the definition of *pAgg from the DD */
	char *pIndex; /* Index into the data dictionary */
	
	assert( pAgg );

	/* Get the definition for given aggregate */
	pIndex = pAgg->name;
	pDef = synGetAggDef( pIFDD, pIndex, (synAGGDEF) NULL );

	pResult = synInitAggResult();
	synDeepAggChecker( pResult,  pAgg, pDef ,pIFDD);

	if( pResult->status == synPASS ) return Success;
	else return Failure;

} /* end cmpCheckClassDefSyntax */

/*****************************************************************************
 * 
 * NAME: cmpCheckAggTransSyntax
 *
 * DESCRIPTION: Checks the syntax of a AggregateTranslation instruction.
 *
 * PARAMETERS:  
 *     pAgg -- (AGGREGATE) Pointer to the instruction. 
 *
 * RETURN VALUE:
 *     cmpCheckIDTransSyntax -- (int) Success = 1, Failure = 0.
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. This implementation is pretty generic: there's little here that's
 *        different from the checkers for other instructions. It's implemented
 *        as a separate function in case some time in the future special 
 *        reporting is required for problems in this kind of instruction.
 *
 *     2. This function assumes that there is an unambiguous definition for
 *        the associated instruction.
 *
 * REVISION HISTORY:
 *   10/07/98  Ken Jarva  --  Original
 *   11/16/98  Ken Jarva  --  Doc. header & code cleanup.
 *   12/21/98  Ken Jarva  --  Calls to improved versions of syn functions.
 *
 ****************************************************************************/
int cmpCheckAggTransSyntax( AGGREGATE pAgg)
{
	/* returns 1 for success, 0 for failure. */
	int Success = 1, Failure = 0;
	synAGGRESULT pResult = NULL; /* The result of the syntax check */
	synAGGDEF pDef = NULL; /* the definition of *pAgg from the DD */
	char *pIndex; /* Index into the data dictionary */
	FILE *pF;
	int /* Debuggery Controls...*/
		printDefinition = 0; /* Set to 1 if you want the definition printed; 0
							  * otherwise. */

	pF = ufmCONTROLS.pCompilerErrors; /* For syntax errors */
	
	assert( pAgg );

	/* Get the definition for given aggregate */
	pIndex = pAgg->name;
	pDef = synGetAggDef( pIFDD, pIndex,(synAGGDEF) NULL );
	if( printDefinition ) synPrintAggDef( pF, pDef );

	/* Check the aggregate against it's definition: */
	pResult = synInitAggResult();
	synDeepAggChecker( pResult, pAgg, pDef, pIFDD );

	if( (pResult->status == synPASS) && (pResult->badChildren == 0) ) 
		return Success;
	else {
		synReportAggResult( pF, pResult );
		return Failure;
	}

} /* end cmpCheckAggTransSyntax */

/*****************************************************************************
 * 
 * NAME: cmpCheckParTransSyntax
 *
 * DESCRIPTION: Checks the syntax of a ParameterTranslation instruction.
 *
 * PARAMETERS:  
 *     pAgg -- (AGGREGATE) Pointer to the instruction. 
 *
 * RETURN VALUE:
 *     cmpCheckIDTransSyntax -- (int) Success = 1, Failure = 0.
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. This implementation is pretty generic: there's little here that's
 *        different from the checkers for other instructions. It's implemented
 *        as a separate function in case some time in the future special 
 *        reporting is required for problems in this kind of instruction.
 *
 *     2. This function assumes that there is an unambiguous definition for
 *        the associated instruction.
 *
 * REVISION HISTORY:
 *   10/07/98  Ken Jarva  --  Original
 *   11/16/98  Ken Jarva  --  Doc. header & code cleanup.
 *   12/21/98  Ken Jarva  --  Calls to improved versions of syn functions.
 *
 ****************************************************************************/
int cmpCheckParTransSyntax( AGGREGATE pAgg )
{
	/* returns 1 for success, 0 for failure. */
	int Success = 1, Failure = 0;
	synAGGRESULT pResult = NULL; /* The result of the syntax check */
	synAGGDEF pDef = NULL; /* the definition of *pAgg from the DD */
	char *pIndex; /* Index into the data dictionary */
	FILE *pF;
	int /* Debuggery Controls...*/
		printDefinition = 0; /* Set to 1 if you want the definition printed; 0
							  * otherwise. */

	pF = ufmCONTROLS.pCompilerErrors; /* For syntax errors */
	
	assert( pAgg );

	/* Get the definition for given aggregate */
	pIndex = pAgg->name;
	pDef = synGetAggDef( pIFDD, pIndex, (synAGGDEF) NULL );
	if( printDefinition ) synPrintAggDef( pF, pDef );

	/* Check the aggregate against it's definition: */
	pResult = synInitAggResult();
	synDeepAggChecker( pResult, pAgg, pDef, pIFDD );

	if( (pResult->status == synPASS) && (pResult->badChildren == 0) ) 
		return Success;
	else return Failure;

} /* end cmpCheckParTransSyntax */

/*****************************************************************************
 * 
 * NAME: cmpCheckMacroDefSyntax
 *
 * DESCRIPTION: Checks the syntax of a MacroDefinition instruction.
 *
 * PARAMETERS:  
 *     pAgg -- (AGGREGATE) Pointer to the instruction. 
 *
 * RETURN VALUE:
 *     cmpCheckIDTransSyntax -- (int) Success = 1, Failure = 0.
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. This implementation is pretty generic: there's little here that's
 *        different from the checkers for other instructions. It's implemented
 *        as a separate function in case some time in the future special 
 *        reporting is required for problems in this kind of instruction.
 *
 *     2. This function assumes that there is an unambiguous definition for
 *        the associated instruction.
 *
 * REVISION HISTORY:
 *   10/07/98  Ken Jarva  --  Original
 *   11/16/98  Ken Jarva  --  Doc. header & code cleanup.
 *   12/21/98  Ken Jarva  --  Calls to improved versions of syn functions.
 *
 ****************************************************************************/
int cmpCheckMacroDefSyntax( AGGREGATE pAgg )
{
	/* returns 1 for success, 0 for failure. */
	int Success = 1, Failure = 0;
	synAGGRESULT pResult = NULL; /* The result of the syntax check */
	synAGGDEF pDef = NULL; /* the definition of *pAgg from the DD */
	char *pIndex; /* Index into the data dictionary */
	FILE *pF;
	int /* Debuggery Controls...*/
		printDefinition = 0; /* Set to 1 if you want the definition printed; 0
							  * otherwise. */

	pF = ufmCONTROLS.pCompilerErrors; /* For syntax errors */
	
	assert( pAgg );

	/* Get the definition for given aggregate */
	pIndex = pAgg->name;
	pDef = synGetAggDef( pIFDD, pIndex, (synAGGDEF) NULL );
	if( printDefinition ) synPrintAggDef( pF, pDef );

	/* Check the aggregate against it's definition: */
	pResult = synInitAggResult();
	synDeepAggChecker( pResult, pAgg, pDef, pIFDD );

	if( (pResult->status == synPASS) && (pResult->badChildren == 0) ) 
		return Success;
	else return Failure;

} /* end cmpCheckMacroDefSyntax */
/*****************************************************************************
 * 
 * NAME: cmpCheckIDTransSyntax
 *
 * DESCRIPTION: Checks the syntax of an IdentifierTranslation instruction
 *
 * PARAMETERS:  
 *     pAgg -- (AGGREGATE) Pointer to the instruction, in this case, an ODL
 *             object representing an IdentifierTranslation instruction.
 *
 * RETURN VALUE:
 *     cmpCheckIDTransSyntax -- (int) Success = 1, Failure = 0.
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. This implementation is pretty generic: there's little here that's
 *        different from the checkers for other instructions. It's implemented
 *        as a separate function in case some time in the future special 
 *        reporting is required for problems in this kind of instruction.
 *     2. This function assumes that there is an unambiguous definition for
 *        the associated instruction.
 *
 * REVISION HISTORY:
 *   10/07/98  Ken Jarva  --  Original
 *   11/13/98  Ken Jarva  --  Doc. header & code cleanup.
 *   12/21/98  Ken Jarva  --  Calls to improved versions of syn functions.
 *
 ****************************************************************************/
int cmpCheckIDTransSyntax( AGGREGATE pAgg )
{
	/* returns 1 for success, 0 for failure. */
	int Success = 1, Failure = 0;
	synAGGRESULT pResult = NULL; /* The result of the syntax check */
	synAGGDEF pDef = NULL; /* the definition of *pAgg from the DD */
	char *pIndex = NULL; /* Index into the data dictionary */
	FILE *pF;
	int /* Debuggery Controls...*/
		printDefinition = 0; /* Set to 1 if you want the definition printed; 0
							  * otherwise. */

	pF = ufmCONTROLS.pCompilerErrors; /* For syntax errors */
	
	assert( pAgg );

	/* Get the definition for given aggregate */
	pIndex = pAgg->name;
	if( strcmp(pIndex, "IDENTIFIERTRANSLATIONS") == 0 ) {
		pDef = synGetAggDef( pIFDD, pIndex, (synAGGDEF) NULL );
		if( printDefinition ) synPrintAggDef( pF, pDef );

		/* Check the aggregate against it's definition: */
		pResult = synInitAggResult();
		synDeepAggChecker( pResult, pAgg, pDef, pIFDD );

		if( (pResult->status == synPASS) && (pResult->badChildren == 0) ) 
			return Success;
	} else {
		/* Error: this isn't an IDTranslation instruction... */
		char msg[256];

		sprintf( msg, "Bad aggregate kind. Expected IdentifierTranslation "
			"but got %s", pIndex );
		cmpFatalError( "cmpCheckIDTransSyntax",msg );
		return Failure; /* In case Fatal error processor returns cntl here */
	}
	return Failure;

} /* end cmpCheckIDTransSyntax */

/*****************************************************************************
 * 
 * NAME: cmpTableProcessor
 *
 * DESCRIPTION: This function processes a TableSpecByColumn or TableSpecByRow
 *     instruction in an instructions file.  It returns a report node
 *     containing information as per the given instruction.
 *
 * PARAMETERS:
 *     pInst (OBJECT)             A pointer to an instruction for translating
 *                                the given component.
 *     pNode (struct ODL_Node *)  A pointer to a descriptor of an ODL
 *                                component--the one to which the translation
 *                                is to be applied.
 *
 * RETURN VALUE:
 *     cmpTableProcessor (REP_NODE)  A pointer to a report node of type
 *                                   cmpTABLE.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Insufficient memory on heap.
 *        POST: Function terminates the program with a fatal error.
 *     2. PRE:  Sufficient memory on heap.
 *        POST: Function creates and returns a table-type report node.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     10/22/1998  J.D. Stroup  Original
 *     12/07/1998  J.D. Stroup  Component name is now included in fatal error
 *                              messages.
 *     01/15/1999  J.D. Stroup  Updated based on change of table_data from an
 *                              array of pointers (lstLIST *) to just a
 *                              lstLIST structure.
 *     01/18/1999  J.D. Stroup  Added ability to hide empty columns/rows in a
 *                              table (including cmpHideEmpty).  Also as part
 *                              of this effort, added cmpMajMin2ColRow, and
 *                              re-ordered some code.  Began using the
 *                              lstDestructor function where appropriate.
 *     01/18/1999  J.D. Stroup  Put in capability to allow multi-valued
 *                              parameters to be handled in one of two
 *                              separate ways: "SingleValue"--the default--in
 *                              which the parameter's values are each placed
 *                              into separate cells; or "WholeParameter" in
 *                              which all of the parameter's values are placed
 *                              in the same cell.  This includes the addition
 *                              of a new variable, cell_contents, which is
 *                              first retrieved from the instructions tree,
 *                              then passed to cmpGetTableData.
 *     03/08/1999  J.D. Stroup  Changed "int's" to "unsigned int's".
 *
 ****************************************************************************/

REP_NODE cmpTableProcessor (OBJECT pInst, struct ODL_Node *pNode)
{
  char          *caption, *cell_contents, *cols_title, *hide_empty, *major,
                *rows_title;
  unsigned int  num_cols, num_major, num_minor, num_rows;
  REP_NODE      pRN;
  lstLIST       col_hdrs, maj_hdrs, min_hdrs, pValList, row_hdrs, table_data;

  struct reportTable *rpt_table;

  /* Create a new report node, mark it as a table-kind, and set its name. */

  pRN = (REP_NODE) malloc (sizeof (struct reportNode));
  if (pRN == NULL)
    cmpFatalError ("cmpTableProcessor",
                   "malloc failure: cannot allocate memory for report node");
  else
    pRN->itemKind = cmpTABLE;

  pRN->pName = strdup ("");

  /* Get the required column or row data component part of the instruction.
     If there is no such component, issue a fatal error message and terminate
     the program.  Note that this error should have been trapped as part of
     the instructions file syntax checking. */

  pValList = cmpGetExpInstCompList (pInst, "COLUMNS");

  if (lstNumEl (pValList) == 0)
  {
    pValList = lstFreeList (pValList);

    pValList = cmpGetExpInstCompList (pInst, "ROWS");

    if (lstNumEl (pValList) == 0)
    {
      const char *Format = "[%s] missing the required 'Columns'/'Rows' "
                         "component of 'TableSpecBy[Column|Row]' instruction";
      char *curr_comp = odlComponentName (pNode);
      char *errmsg = (char *) malloc ((strlen (Format) +
                                     strlen (curr_comp) + 1) * sizeof (char));
      if (errmsg == NULL)
        cmpFatalError ("cmpTableProcessor",
           "malloc failure: cannot allocate memory for error message string");

      sprintf (errmsg, Format, curr_comp);

      cmpFatalError ("cmpTableProcessor", errmsg);
    }
    else
    {
      major = strdup ("row");  /* table data is row major */

      /* Get the optional row (major) and column (minor) headers component
         part of the instruction. */

      maj_hdrs = cmpGetExpInstCompList (pInst, "ROWHEADERS");
      min_hdrs = cmpGetExpInstCompList (pInst, "COLUMNHEADERS");

      /* Get the optional component indicating whether empty rows should be
         hidden or not in the resulting HTML output. */

      hide_empty = cmpGetExpInstCompVal (pInst, "HIDEEMPTYROWS");
    }
  }
  else
  {
    major = strdup ("column");  /* table data is column major */

    /* Get the optional column (major) and row (minor) headers component part
       of the instruction. */

    maj_hdrs = cmpGetExpInstCompList (pInst, "COLUMNHEADERS");
    min_hdrs = cmpGetExpInstCompList (pInst, "ROWHEADERS");

    /* Get the optional component indicating whether empty columns should be
       hidden or not in the resulting HTML output. */

    hide_empty = cmpGetExpInstCompVal (pInst, "HIDEEMPTYCOLUMNS");
  }

  /* Get the optional caption component part of the instruction. */

  caption = cmpGetExpInstCompVal (pInst, "CAPTION");

  /* Get the optional columns title component part of the instruction. */

  cols_title = cmpGetExpInstCompVal (pInst, "COLUMNSTITLE");

  /* Get the optional rows title component part of the instruction. */

  rows_title = cmpGetExpInstCompVal (pInst, "ROWSTITLE");

  /* Get the optional cell contents component part of the instruction. */

  cell_contents = cmpGetExpInstCompVal (pInst, "CELLCONTENTS");

  /* Get the table data from the user-supplied component lists (in
     pValList). */

  table_data = cmpGetTableData (pNode, pValList, cell_contents);

  /* Compute the number of major and minor axes in the table. */

  num_major = cmpMajorDimSize (maj_hdrs, table_data);
  num_minor = cmpMinorDimSize (min_hdrs, table_data);

  /* Backfill the headers and table data, if necessary. */

  table_data = cmpTblBackfill (num_major, num_minor, maj_hdrs, min_hdrs,
                               table_data);

  /* If the user requested, hide (read "remove") empty columns/rows in the
     table.  Empty columns are removed from a table in column major order.
     Empty rows are removed from a table in row major order.  In the process,
     update the number of major axes (colums/rows) in the table. */

  if (strToUpStrcmp (hide_empty, "YES") == 0)
    num_major -= cmpHideEmpty (table_data, maj_hdrs);

  /* Based on the number of major and minor axes, and an indication of the
     table's order (column major or row major), compute the number of columns
     and rows in the table. */

  cmpMajMin2ColRow (num_major, num_minor, major, &num_cols, &num_rows);

  /* Determine which headers are the column headers and which are the row
     headers. */

  col_hdrs = (strcmp (major, "column") == 0) ? maj_hdrs : min_hdrs;
  row_hdrs = (strcmp (major, "row")    == 0) ? maj_hdrs : min_hdrs;

  /* Initialize the reportTable structure with the data gathered above. */

  rpt_table = (void *) cmpInitRptTbl (num_cols, num_rows, caption,
                                      cols_title, rows_title, row_hdrs,
                                      col_hdrs, table_data, major);

  /* Set the report node's item pointer to the address of the reportTable
     structure */

  pRN->pItem = (void *) rpt_table;

  /* Perform cleanup activities. */

  pValList = lstDestructor (pValList, strDestructor);

  row_hdrs = lstFreeList (row_hdrs);
  col_hdrs = lstFreeList (col_hdrs);
  table_data = lstDestructor (table_data, cmpTblDataDestructor);

  /* Return the pointer to the table-kind report node created by the
     function. */

  return (pRN);
} /* end cmpTableProcessor */

/*****************************************************************************
 * 
 * NAME: cmpGetTableData
 *
 * DESCRIPTION: This function uses the given components list (in value_list)
 *     to create lists representing the table's data.  Each list contains the
 *     data in one column of the table--if column major--or one row of the
 *     table--if row major.
 *
 * PARAMETERS:
 *     pNode (struct ODL_Node *)  A pointer to a descriptor of an ODL
 *                                component--the one to which the translation
 *                                is to be applied.
 *     value_list (lstLIST)       Pointer to a list.  Each element of the list
 *                                is a user-specified component list from the
 *                                instruction file.  Each component list
 *                                resolves to one or more parameters whose
 *                                values comprise the table's data.
 *     cell_contents (char *)     A general indication of the contents of each
 *                                cell in a table: either "SingleValue" or
 *                                "WholeParameter".  This indicator is used
 *                                when multi-valued parameters are
 *                                encountered.  "SingleValue", which is the
 *                                default, tells the system that each value of
 *                                a parameter is to be placed in its own cell.
 *                                "WholeParameter" tells the system that all
 *                                values of a parameter are to be placed in
 *                                the same cell.
 *
 * RETURN VALUE:
 *     cmpGetTableData (lstLIST)  A list of lists representing the data in a
 *                                table.  Each element of the primary list
 *                                points to a list which represents one column
 *                                in a column major table, or one row in a row
 *                                major table.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Insufficient memory on heap.
 *        POST: Function terminates the program after issuing a fatal error
 *              message.
 *     2. PRE:  Sufficient memory on heap.
 *        POST: Lists representing the columns or rows of the table are
 *              allocated, filled, and returned.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     10/23/1998  J.D. Stroup  Original
 *     12/07/1998  J.D. Stroup  1) A column/row references that does not
 *                              resolve to a parameter is now considered a
 *                              fatal error.
 *                              2) Checks are made to verify that the given
 *                              column/row references resolve to at least one
 *                              component.  If not, it's a fatal error.
 *                              3) Component name is now included in fatal
 *                              error messages.
 *     01/15/1999  J.D. Stroup  Updated based on data type of returned value
 *                              from an array of pointers (lstLIST *) to just
 *                              a lstLIST structure.
 *     01/18/1999  J.D. Stroup  Added call to lstDestructor.  Added several
 *                              calls to lstStopIteration for prevenative
 *                              measures.
 *     01/18/1999  J.D. Stroup  Put in capability to allow muli-valued
 *                              parameters to be handled in one of two
 *                              separate ways: "SingleValue"--the default--in
 *                              which the parameter's values are each placed
 *                              into separate cells; or "WholeParameter" in
 *                              which all of the parameter's values are placed
 *                              in the same cell.  Included in this change is
 *                              the addition of the new argument,
 *                              cell_contents.
 *
 ****************************************************************************/

static lstLIST cmpGetTableData (struct ODL_Node *pNode, lstLIST value_list,
                                char *cell_contents)
{
  char       *comp_list;
  int        i;
  int        tot_cells = 0;
  lstLIST    col_row_list, major_dim_list, pParmValList, table_data;
  ODLNodePtr *odl_node = (ODLNodePtr *) malloc (sizeof (ODLNodePtr));

  if (odl_node == NULL)
    cmpFatalError ("cmpGetTableData",
                   "malloc failure: cannot allocate memory for ODL node "
                   "pointer structure");

  if (lstNumEl (value_list) == 0)
  {
    const char *Format = "[%s] no column/row data given";
    char *curr_comp = odlComponentName (pNode);
    char *errmsg = (char *) malloc ((strlen (Format) +
                                     strlen (curr_comp) + 1) * sizeof (char));
    if (errmsg == NULL)
      cmpFatalError ("cmpGetTableData",
           "malloc failure: cannot allocate memory for error message string");

    sprintf (errmsg, Format, curr_comp);

    cmpFatalError ("cmpGetTableData", errmsg);
  }

  /* Allocate array of pointers: one for each column/row. */

  table_data = lstInitList ();

  if (table_data == NULL)
    cmpFatalError ("cmpGetTableData",
       "malloc failure: cannot allocate memory for table data pointer array");

  /* Process each column/row data value (component list). */

  lstStopIteration (value_list);
  for (i = 0; (comp_list = (char *) lstIterator (value_list)) != NULL; i++)
  {
  	/* Expand the component list into the internal component list
       representation. */

    major_dim_list = ListExpander (*pNode, comp_list);

    tot_cells += lstNumEl (major_dim_list);

  	/*** Create a list of the column's/row's data. ***/

	  /* Initialize the list for this column's/row's data.  The pointer to this
       list points to a list of elements where each element contains a string.
       Taken together, all the elements (strings) represent the data in one
       column/row of the table.  The subsequent loop creates the
       column's/row's data. */

    col_row_list = lstInitList ();

	  /* Process each element (component) of the expanded component list. */

    lstStopIteration (major_dim_list);
    while ((odl_node = (ODLNodePtr *) lstIterator (major_dim_list)) != NULL)
    {
  		/* If the component is NOT a parameter node then issue a fatal error
         message and terminate the program via cmpFatalError. */

      if (odl_node->node_kind != PARAMETER_NODE)
      {
        const char *Format = "[%s] reference in a column/row specification "
                             "does not resolve to a parameter node";
        char *curr_comp = odlComponentName (pNode);
        char *errmsg = (char *) malloc ((strlen (Format) +
                                     strlen (curr_comp) + 1) * sizeof (char));
        if (errmsg == NULL)
          cmpFatalError ("cmpGetTableData",
           "malloc failure: cannot allocate memory for error message string");

        sprintf (errmsg, Format, curr_comp);

        cmpFatalError ("cmpGetTableData", errmsg);
      }

      /* Get the values of the reference as a list of strings. */

      pParmValList = odlGetValueList (odl_node);

      /* If all values of the parameter are to be placed into the same cell,
        then join the values into a single string delimited by "<BR>" (the
        HTML line break tag).  Free the current list of values, and create a
        new list with only one element: the delimited string that was just
        created. */

      if (strToUpStrcmp (cell_contents, "WholeParameter") == 0)
      {
        const char *Delimiter = strdup ("<BR>");
        char *string = lstJoinList ((char *) Delimiter, pParmValList);

        pParmValList = lstFreeList (pParmValList);
        pParmValList = lstInitList ();
        lstAddLast (pParmValList, (void *) string);
      }

    	/* Append the list of values for this component to the end of the list
         of column/row data being created. */

      lstAppendTo (col_row_list, pParmValList);

      pParmValList = lstFreeList (pParmValList);
    }

    /* Add the major axis list, either a column or row of data, to the table's
       data being constructed. */

    lstAddLast (table_data, (void *) col_row_list);

    /* Free the list including all the ODL node pointer structures to which
       the elements of the list points. */

    major_dim_list = lstDestructor (major_dim_list, odlDestructor);
  }

  /* Make sure there is table data.  Although the user has given column/row
     specifications, they may not have resolved to any components. */

  if (tot_cells == 0)
  {
    const char *Format =
               "[%s] no column/row data given: bad column/row specifications";
    char *curr_comp = odlComponentName (pNode);
    char *errmsg = (char *) malloc ((strlen (Format) +
                                     strlen (curr_comp) + 1) * sizeof (char));
    if (errmsg == NULL)
      cmpFatalError ("cmpGetTableData",
           "malloc failure: cannot allocate memory for error message string");

    sprintf (errmsg, Format, curr_comp);

    cmpFatalError ("cmpGetTableData", errmsg);
  }

  return (table_data);
} /* end cmpGetTableData */

/*****************************************************************************
 * 
 * NAME: cmpMajorDimSize
 *
 * DESCRIPTION: This function calculates the size of, or the number of
 *              elements in, a table's major axis.  See NOTES section, below.
 *
 * PARAMETERS:
 *     major_hdrs (lstLIST)  A list of the headers for the table's major axis.
 *     table_data (lstLIST)  A list of lists representing the data in a table.
 *                           Each element of the primary list points to a list
 *                           which represents one column in a column major
 *                           table, or one row in a row major table.
 *
 * RETURN VALUE:
 *     cmpMajorDimSize (int)  The size of the table's major axis--e.g., the
 *                            number of columns in a column major table.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     Columns are the major axis in a column major table.  Row are the minor
 *     axis.  The opposite is true for a row major table.  A table is column
 *     major if it is specified in a TableSpecByColumn instruction: it is row
 *     major if it is specified in a TableSpecByRow instruction.
 *
 * REVISION HISTORY:
 *     10/22/1998  J.D. Stroup  Original
 *     01/15/1999  J.D. Stroup  Changed the second argument.  It used to be an
 *                              integer--the number of column or row
 *                              specification given by the user.  Replaced
 *                              with a pointer to the table data.  Updated
 *                              based on change of table_data from an array of
 *                              pointers (lstLIST *) to just a lstLIST
 *                              structure.
 *     03/08/1999  J.D. Stroup  Changed function's return type (and the types
 *                              of the local variables, tot_maj_axis_elems &
 *                              tot_major_hdrs) from "int" to "unsigned int".
 *                              Cast returned value from lstNumEl to "unsigned
 *                              int".
 *
 ****************************************************************************/

static unsigned int cmpMajorDimSize (lstLIST major_hdrs, lstLIST table_data)
{
  /* Calculate the number elements in the table's major axis.  Also, calculate
     the number of headers given for the table's major axis. */

  unsigned int tot_maj_axis_elems = (unsigned int) lstNumEl (table_data);
  unsigned int tot_major_hdrs = (unsigned int) lstNumEl (major_hdrs);

  /* If headers for the table's major axis were given, and their number is
     different than the size of the table's major axis, issue a warning
     message to the user.  The data provided by the user generates a ragged
     table which will need to be backfilled. */

  if ((tot_major_hdrs != tot_maj_axis_elems) && (tot_major_hdrs != 0))
    cmpUserWarning ("cmpMajorDimSize",
           "number of elements differs in one or more columns/rows of table");

  /* Return the size of the table's major axis.  It is the greater of two
     values: the number of headers, and the size of the table's major axis. */

  return ((tot_major_hdrs >= tot_maj_axis_elems) ? tot_major_hdrs :
                                                   tot_maj_axis_elems);
} /* end cmpMajorDimSize */

/*****************************************************************************
 * 
 * NAME: cmpMinorDimSize
 *
 * DESCRIPTION: This function calculates the size of, or the number of
 *              elements in, a table's minor axis.  See NOTES section, below.
 *
 * PARAMETERS:
 *     minor_hdrs (lstLIST)  A list of the headers for the table's minor axis.
 *     table_data (lstLIST)  A list of lists representing the data in a table.
 *                           Each element of the primary list points to a list
 *                           which represents one column in a column major
 *                           table, or one row in a row major table.
 *
 * RETURN VALUE:
 *     cmpMinorDimSize (int)  The size of the table's minor axis--e.g., the
 *                            number of rows in a column major table.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     Columns are the major axis in a column major table.  Row are the minor
 *     axis.  The opposite is true for a row major table.  A table is column
 *     major if it is specified in a TableSpecByColumn instruction: it is row
 *     major if it is specified in a TableSpecByRow instruction.
 *
 * REVISION HISTORY:
 *     10/23/1998  J.D. Stroup  Original
 *     01/15/1999  J.D. Stroup  Removed third argument (num_data_major).
 *                              Updated based on change of table_data from an
 *                              array of pointers (lstLIST *) to just a
 *                              lstLIST structure.
 *     03/08/1999  J.D. Stroup  Changed function's return type (and the types
 *                              of the local "int" variables) from "int" to
 *                              "unsigned int".  Cast returned value from
 *                              lstNumEl to "unsigned int".
 *
 ****************************************************************************/

static unsigned int cmpMinorDimSize (lstLIST minor_hdrs, lstLIST table_data)
{
  /* Based on the number of headers along the table's minor axis, initialize
     the size of the minor axis. */

  lstLIST       maj_list;
  unsigned int  max_minor_size = (unsigned int) lstNumEl (minor_hdrs);
  unsigned int  curr_minor_size;
  unsigned int  diff_minor_sizes = 0;

  /* Step through each list in the given table data and compute its size.  If
     it's greater than the table's minor axis size previously computed, update
     the table's minor axis size with this value. */

  lstStopIteration (table_data);
  while ((maj_list = (lstLIST) lstIterator (table_data)) != NULL)
  {
    curr_minor_size = (unsigned int) lstNumEl (maj_list);
    if ((curr_minor_size != max_minor_size) && (max_minor_size != 0))
                                                           diff_minor_sizes = 1;
    if (curr_minor_size > max_minor_size) max_minor_size = curr_minor_size;
  }

  /* If any data along a minor axis in the table is different in size than any
     other data along the same axis, issue a warning message to the user.  The
     data provided by the user generates a ragged table which will need to be
     backfilled. */

  if (diff_minor_sizes)
    cmpUserWarning ("cmpMinorDimSize",
           "number of elements differs in one or more columns/rows of table");

  /* Return the size of the table's minor axis. */

  return (max_minor_size);
} /* end cmpMinorDimSize */

/*****************************************************************************
 * 
 * NAME: cmpTblBackfill
 *
 * DESCRIPTION: Given table specifications, this function backfills the table,
 *     if necessary so that all of it columns are the same length and all of
 *     its rows are the same length.  The table specifications are data
 *     structures containing column headers, row headers, and table data.
 *     Also, see NOTES section, below.
 *
 * PARAMETERS:
 *     num_major (unsigned int)
 *                           The size of, or the number of elements along, the
 *                           table's major axis.
 *     num_minor (unsigned int)
 *                           The size of, or the number of elements along, the
 *                           table's minor axis.
 *     major_hdrs (lstLIST)  A list of the headers for the table's major axis.
 *                           If this function determines that there is more
 *                           data along the major axis than headers, then new
 *                           "headers" are added to this list.  The added
 *                           headers are simply emtpy strings.
 *     minor_hdrs (lstLIST)  A list of the headers for the table's minor axis.
 *                           If this function determines that there is more
 *                           data along the minor axis than headers, then new
 *                           "headers" are added to this list.  The added
 *                           headers are simply emtpy strings.
 *     table_data (lstLIST)  A list of lists representing the data in a table.
 *                           Each element of the primary list points to a list
 *                           which represents one column in a column major
 *                           table, or one row in a row major table.  See
 *                           NOTES section, below.  If this function
 *                           determines that there are more column headers or
 *                           row headers than columns or rows of data,
 *                           then--depending on what's required--either lists
 *                           are added to the array, or new data elements are
 *                           added to the existing lists in this structure.
 *                           The added elements are simply emtpy strings.
 *                           Similarly if the lists are of differing length
 *                           new elements are added to the shorter lists,
 *                           making them equal in length to the longest list.
 *
 * RETURN VALUE:
 *     cmpTblBackfill (lstLIST)  A list of lists.  It is a backfilled version
 *                               of the input array table_data.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     Columns are the major axis in a column major table.  Row are the minor
 *     axis.  The opposite is true for a row major table.  A table is column
 *     major if it is specified in a TableSpecByColumn instruction: it is row
 *     major if it is specified in a TableSpecByRow instruction.
 *
 *     Taken together, the given column headers, row headers, and table data,
 *     represent a possibly ragged table.  The table is ragged if all of its
 *     columns are not of equal length or if all of its rows are not of equal
 *     length.  Remember that the row headers comprise a column of the table
 *     and the column headers comprise a row of the table.
 *
 *     The output array represents a table with all of its column of equal
 *     length and all of its rows of equal length.
 *
 * REVISION HISTORY:
 *     10/23/1998  J.D. Stroup  Original
 *     01/15/1999  J.D. Stroup  Removed last argument (num_data_major).
 *                              Updated based on change of table_data from an
 *                              array of pointers (lstLIST *) to just a
 *                              lstLIST structure.
 *     03/09/1999  J.D. Stroup  Changed type of "int" arguments to "unsigned
 *                              int".  Changed type of local variables--"i",
 *                              "j", and "num_data_major"--from "int" to
 *                              "unsigned int".
 *
 ****************************************************************************/

static lstLIST cmpTblBackfill (unsigned int num_major, unsigned int num_minor,
                               lstLIST major_hdrs, lstLIST minor_hdrs,
                               lstLIST table_data)
{
  char *empty_string = strdup ("");
  void *vp = empty_string;
  unsigned int i, j;
  lstLIST new_tbl_data, maj_list;
  unsigned int num_data_major = (unsigned int) lstNumEl (table_data);

  /* Backfill the headers for the major axis (e.g., columns in a column major
     table) in the table, if necessary. */

  if (lstNumEl (major_hdrs) != 0)
  {
    for (i = (unsigned int) lstNumEl (major_hdrs); i < num_major; i++)
      lstAddLast (major_hdrs, vp);
  }

  /* Backfill the headers for the minor axis (e.g., rows in a column major
     table) in the table, if necessary. */

  if (lstNumEl (minor_hdrs) != 0)
  {
    for (i = (unsigned int) lstNumEl (minor_hdrs); i < num_minor; i++)
      lstAddLast (minor_hdrs, vp);
  }

  /* Backfill the table data along the minor axis, if necessary. */

  lstStopIteration (table_data);
  while ((maj_list = (lstLIST) lstIterator (table_data)) != NULL)
  {
    for (j = (unsigned int) lstNumEl (maj_list); j < num_minor; j++)
      lstAddLast (maj_list, vp);
  }

  /* Backfill the table data along the major axis, if necessary.  This
     requires creating new columns or rows, whichever is the major order of
     the table. */

  if (num_major > num_data_major)
  {
    /* First create space for a new table. */

    new_tbl_data = lstInitList ();

    if (new_tbl_data == NULL)
      cmpFatalError ("cmpTblBackfill",
                     "malloc failure: cannot allocate memory for new table "
                     "data pointer array");

    /* Fill the new table with all the data from the old table. */

    lstStopIteration (table_data);
    while ((maj_list = (lstLIST) lstIterator (table_data)) != NULL)
      lstAddLast (new_tbl_data, (void *) maj_list);

    /* Backfill the new table as necessary. */

    for (i = num_data_major; i < num_major; i++)
    {
      maj_list = lstInitList ();
      for (j = 0; j < num_minor; j++)
        lstAddLast (maj_list, vp);
      lstAddLast (new_tbl_data, (void *) maj_list);
    }

    /* We no longer need the original table data pointer array, so free its
       space. */

    lstFreeList (table_data);
  }

  /* No need to backfill the table data, so set a pointer to point to the
    unchanged table data. */

  else
    new_tbl_data = table_data;

  return (new_tbl_data);
} /* end cmpTblBackfill */

/*****************************************************************************
 * 
 * NAME: cmpHideEmpty
 *
 * DESCRIPTION: This function removes empty columns or rows from a table.  It
 *     returns the number of columns/rows removed.  If the table is column
 *     major, then empty columns are removed.  If the table is row major, then
 *     empty rows are removed.  The function is passed the table's data and
 *     the headers along the table's major axis (because if an empty
 *     column/row is removed so is its corresponding header).
 *
 * PARAMETERS:
 *     table_data (lstLIST)  A list of lists representing the data in a table.
 *                           Each element of the primary list points to a list
 *                           which represents one column in a column major
 *                           table, or one row in a row major table.
 *     maj_hdrs (lstLIST)    A pointer to a list containing the headers along
 *                           the table's major axis.
 *
 * RETURN VALUE:
 *     cmpHideEmpty (int)  The number of columns/rows removed from the table.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Table data is in column major order and contains no empty
 *              columns.
 *        POST: No columns removed from table.  A value of zero is returned.
 *     2. PRE:  Table data is in row major order and contains no empty rows.
 *        POST: No rows removed from table.  A value of zero is returned.
 *     3. PRE:  Table data is in column major order and contains "n" empty
 *              columns, where "n" > 0 and "n" < the total number of columns
 *              in table.
 *        POST: "n" columns removed from table.  If there are column headers,
 *              then the corresponding "n" headers are removed, as well.  A
 *              value of "n" is returned.
 *     4. PRE:  Table data is in row major order and contains "n" empty rows,
 *              where "n" > 0 and "n" < the total number of rows in table.
 *        POST: "n" rows removed from table.  If there are row headers,
 *              then the corresponding "n" headers are removed, as well.  A
 *              value of "n" is returned.
 *
 * NOTES:
 *     It is assumed that the "table" passed to this function contains at
 *     least one column/row of data.  This is because a check for an empty
 *     table was previously made in cmpGetTableData.  If a table was found to
 *     be empty, a fatal error message was generated and the program
 *     terminated.  Thus we're guaranteed that at the end of "hiding" all
 *     empty columns/rows in this function, we'll still have at least one
 *     non-empty column/row.
 *
 *     It's also assumed--if major axis headers provided--that their number is
 *     equal to the number of items along the table's major axis.
 *
 *
 * REVISION HISTORY:
 *     01/13/1999  J.D. Stroup  Original
 *     03/08/1999  J.D. Stroup  Changed function's return type (and the type of
 *                              the local variable, num_removed) from "int" to
 *                              "unsigned int".
 *
 ****************************************************************************/

static unsigned int cmpHideEmpty (lstLIST table_data, lstLIST maj_hdrs)
{
  lstLIST       maj_axis_list; /* a list containing the items in one of the
                                * table's major axes (e.g., row elements of a
				* column, if table is column major) */
  int           hdrs_avail = (lstNumEl (maj_hdrs) > 0) ? 1 : 0; /* indicates
                              * whether major axis header have been provided */
  unsigned int  num_removed = 0; /* number of columns/rows removed from
                                  * table */

  lstStopIteration (table_data);
  lstStopIteration (maj_hdrs);

  /* Step through each columns/row--whichever is the major axis--of the given
     table's data. */

  while ((maj_axis_list = (lstLIST) lstIterator (table_data)) != NULL)
  {
    /* If there are any headers along the major axis, step to the next one. */

    if (hdrs_avail) lstIterator (maj_hdrs);

    /* If the current column/row of the table is empty, i.e., contains no
       elements, then remove it.  Also, if there are headers, then remove the
       current one.  Finally, bump a counter of the number of columns/rows
       removed. */

    if (lstIsListOfEmptyStrings (maj_axis_list))
    {
      lstPluckCurrSetPrev (table_data);
      if (hdrs_avail) lstPluckCurrSetPrev (maj_hdrs);
      num_removed++;
    }
  }

  /* Return the number of columns/rows removed. */

  return (num_removed);
}

/*****************************************************************************
 * 
 * NAME: cmpInitRptTbl
 *
 * DESCRIPTION: This function creates, fills, and returns the reportTable
 *     structure for a table.  All of the necessary table information is
 *     provided to the function via the calling arguments.
 *
 * PARAMETERS:
 *     num_cols (unsigned int)  Number of columns in the table.
 *     num_rows (unsigned int)  Number of rows in the table.
 *     caption (char *)         Pointer to the table's caption.
 *     cols_title (char *)      Pointer to the table's columns title.
 *     rows_title (char *)      Pointer to the table's rows title.
 *     row_headers (lstLIST)    A list of the row headers.  Each element in the
 *                              list contains a pointer to one row's header.
 *     col_headers (lstLIST)    A list of the column headers.  Each element in
 *                              the list contains a pointer to one column's
 *                              header.
 *     table_data (lstLIST)     A list of lists representing the data in a
 *                              table. Each element of the primary list points
 *                              to a list which represents one column in a
 *                              column major table, or one row in a row major
 *                              table.
 *     major (const char *)     Pointer to either the string "column"--if the
 *                              table is column major--or "row"--if the table
 *                              is row major.
 *
 * RETURN VALUE:
 *     cmpInitRptTbl (struct reportTable *)
 *                             Pointer to a reportTable structure containing
 *                             the necessary information to construct the
 *                             table.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     10/23/1998  J.D. Stroup  Original
 *     01/15/1999  J.D. Stroup  Updated based on change of table_data from an
 *                              array of pointers (lstLIST *) to just a
 *                              lstLIST structure.
 *     01/18/1999  J.D. Stroup  Added several calls to lstStopIteration to
 *                              correct bugs encountered in later processing.
 *     03/09/1999  J.D. Stroup  Changed type of "int" arguments to "unsigned
 *                              int".  Changed type of local variables--"col",
 *                              "row", and "cell"--from "int" to "unsigned
 *                              int".
 *
 ****************************************************************************/

static struct reportTable *cmpInitRptTbl
    (unsigned int num_cols, unsigned int num_rows, char *caption,
     char *cols_title, char *rows_title, lstLIST row_headers,
     lstLIST col_headers, lstLIST table_data, const char *major)
{
  unsigned int col, row;
  unsigned int cell = 0;

  /* Create the reportTable structure. */

  struct reportTable *rpt_table =
                  (struct reportTable *) malloc (sizeof (struct reportTable));

  if (rpt_table == NULL)
    cmpFatalError ("cmpInitRptTbl",
          "malloc failure: cannot allocate memory for reportTable structure");

  /* Set the caption field of the reportTable structure to the macro-expanded
     caption specification. */

  rpt_table->pCaption = caption;

  /* Set the columns title field of the reportTable structure to the
     macro-expanded columns title specification. */

  rpt_table->pColTitle = cols_title;

  /* Set the rows title field of the reportTable structure to the
     macro-expanded rows title specification. */

  rpt_table->pRowTitle = rows_title;

  /* If no column headers given then set the column headers field of the
     reportTable structure to NULL.  Otherwise, allocate memory for m
     character pointers, where m is the number of columns in the table.
     Assign each pointer to point to one column header string. */

  if (lstNumEl (col_headers) == 0)
    rpt_table->ppColHeaders = NULL;
  else
  {
    char *col_hdr;
    int i;

    rpt_table->ppColHeaders = (char **) malloc (num_cols * sizeof (char *));
    if (rpt_table->ppColHeaders == NULL)
      cmpFatalError ("cmpInitRptTbl",
                     "malloc failure: cannot allocate memory for the column "
                     "headers of reportTable structure");

    lstStopIteration (col_headers);
    for (i = 0; (col_hdr = (char *) lstIterator (col_headers)) != NULL; i++)
      rpt_table->ppColHeaders[i] = col_hdr;
  }

  /* If no row headers given then set the row headers field of the
     reportTable structure to NULL.  Otherwise, allocate memory for n
     character pointers, where n is the number of rows in the table.
     Assign each pointer to point to one row header string. */

  if (lstNumEl (row_headers) == 0)
    rpt_table->ppRowHeaders = NULL;
  else
  {
    char *row_hdr;
    int i;

    rpt_table->ppRowHeaders = (char **) malloc (num_rows * sizeof (char *));
    if (rpt_table->ppRowHeaders == NULL)
      cmpFatalError ("cmpInitRptTbl",
                     "malloc failure: cannot allocate memory for the row "
                     "headers of reportTable structure");

    lstStopIteration (row_headers);
    for (i = 0; (row_hdr = (char *) lstIterator (row_headers)) != NULL; i++)
      rpt_table->ppRowHeaders[i] = row_hdr;
  }

	/* Allocate memory for m x n character pointers, where m is the number of
     columns and n is number of rows in the table.  Based on whether the data
     is given by column or row, assign each pointer to point to one table
     cell's data, i.e., the data from one row and one column. */

  rpt_table->ppTable = (char **) malloc (num_cols * num_rows *
                                         sizeof (char *));
  if (rpt_table == NULL)
    cmpFatalError ("cmpInitRptTbl",
                   "malloc failure: cannot allocate memory for the table "
                   "data of reportTable structure");

  if (strcmp (major, "column") == 0)
  {
    /* Data given by columns, i.e., column major. */

    lstLIST *cols_list;
    lstLIST pList;
    int i = 0;
    
    cols_list = (lstLIST *) malloc ((unsigned int)lstNumEl (table_data) *
                                    sizeof (lstLIST));

    lstStopIteration (table_data);
    while ((pList = (lstLIST) lstIterator (table_data)) != NULL)
    {
      lstStopIteration (pList);
      cols_list[i++] = pList;
    }

    for (row = 0; row < num_rows; row++)
      for (col = 0; col < num_cols; col++)
        rpt_table->ppTable[cell++] = (char *) lstIterator (cols_list[col]);

    free (cols_list);
  }
  else
  {
    /* Data given by rows, i.e., row major. */

    char *string;
    lstLIST row_list;

    lstStopIteration (table_data);
    while ((row_list = (lstLIST) lstIterator (table_data)) != NULL)
    {
      lstStopIteration (row_list);
      while ((string = (char *) lstIterator (row_list)) != NULL)
        rpt_table->ppTable[cell++] = string;
    }
  }

  /* Assign the number of columns and number of rows fields of the repotTable
     structure based on the given input. */

  rpt_table->numCols = num_cols;
  rpt_table->numRows = num_rows;

  return (rpt_table);
} /* end cmpInitRptTbl */

/*****************************************************************************
 * 
 * NAME: cmpGetExpInstCompVal
 *
 * DESCRIPTION: This function returns the macro-expanded value of a given
 *     instruction's component.  It is given a pointer to an instruction and a
 *     pointer to the name of one of its components.  It returns the value of
 *     the component after all macros, if any, that it may contain have been
 *     expanded.  If the component cannot be found, then NULL is returned.
 *
 * PARAMETERS:
 *     pInst (OBJECT)            Pointer to an aggregate--specifically an
 *                               object--node in the ODL instruction tree.
 *     comp_name (const char *)  Pointer to the name of the instruction's
 *                               component whose value is to be returned.
 *
 * RETURN VALUE:
 *     cmpGetExpInstCompVal (char *)  Pointer to the macro-expanded value of
 *                                    the component, or NULL if the component
 *                                    does not exist in the given instruction.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Given component exists in the given instruction.
 *        POST: A pointer to its macro-expanded value is returned.
 *     2. PRE:  Given component does not exist in the given instruction.
 *        POST: NULL is returned.
 *
 * NOTES:
 *     Memory for the the returned string is allocated in macMacroExpander.
 *
 * REVISION HISTORY:
 *     10/23/1998  J.D. Stroup  Original
 *
 ****************************************************************************/

static char *cmpGetExpInstCompVal (OBJECT pInst, const char *comp_name)
{
  char *string, *expanded_string;

  /* Get the requested component part, comp_name, of the instruction. */

  PARAMETER pParm = FindParameter (pInst, (char *)comp_name);

  /* If the requested component part exists, get its specification, or value,
     in string format, and expand any macros in its specification. */

  if (pParm) /* component exists */
  {
    string = odl_stringValue (pParm);
    expanded_string = macMacroExpander (string);
  }
  else /* no such component: return NULL */
    expanded_string = NULL;

  return (expanded_string);
} /* end cmpGetExpInstCompVal */

/*****************************************************************************
 * 
 * NAME: cmpGetExpInstCompList
 *
 * DESCRIPTION: This function returns the macro-expanded values of a given
 *     instruction's component as a list.  It is given a pointer to an
 *     instruction and a pointer to the name of one of its components.  It
 *     returns the values of the component in a list after all macros, if any,
 *     that they may contain have been expanded.  If the component cannot be
 *     found, then an empty list is returned.
 *
 * PARAMETERS:
 *     pInst (OBJECT)            Pointer to an aggregate--specifically an
 *                               object--node in the ODL instruction tree.
 *     comp_name (const char *)  Pointer to the name of the instruction's
 *                               component whose values are to be returned.
 *
 * RETURN VALUE:
 *     cmpGetExpInstCompList (lstLIST)  Pointer to a list.  Each element of
 *                                      the list contains a pointer to one of
 *                                      the macro-expanded values of the
 *                                      component.  If the component does not
 *                                      exist, the returned list is empty.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Given component exists in the given instruction.
 *        POST: A pointer to a list containing the component's macro-expanded
 *              values is returned.
 *     2. PRE:  Given component does not exist in the given instruction.
 *        POST: An empty list is returned.
 *
 * NOTES:
 *     Memory for each of the macro-expanded strings is allocated in
 *     macMacroExpander.
 *
 * REVISION HISTORY:
 *     10/23/1998  J.D. Stroup  Original
 *     01/18/1999  J.D. Stroup  Updated calling sequence for cmpElemMacExpand
 *                              (see this function for details).  This
 *                              included expanding the scope of the variable
 *                              pValList to the entire function, and removing
 *                              the local variable out_list.
 *
 ****************************************************************************/

static lstLIST cmpGetExpInstCompList (OBJECT pInst, const char *comp_name)
{
  lstLIST pValList;

  /* Get the requested component part, comp_name, of the instruction. */

  PARAMETER pParm = FindParameter (pInst, (char *)comp_name);

  /* If the requested component part exists, get its specification, or values,
     as a list of strings, and expand any macros that they may contain. */

  if (pParm) /* component exists */
  {
    ODLNodePtr *odl_node = (ODLNodePtr *) malloc (sizeof (ODLNodePtr));

    if (odl_node == NULL)
      cmpFatalError ("cmpGetExpInstCompList",
                     "malloc failure: cannot allocate memory for ODL node "
                     "pointer structure");

    odl_node->node_kind = PARAMETER_NODE;
    odl_node->node_type.parameter = pParm;
    pValList = odlGetValueList (odl_node);
    cmpElemMacExpand (pValList);
  }
  else /* no such component: return an empty list */
    pValList = lstInitList ();

  return (pValList);
} /* end cmpGetExpInstCompList */

/*****************************************************************************
 * 
 * NAME: cmpMajMin2ColRow
 *
 * DESCRIPTION: This function converts the number of elements along a table's
 *     major and minor axes to the number of columns and rows in the table.
 *     It is passed the number of elements along the table's major and minor
 *     axes and an indication of whether the table is column major or row
 *     major.  Via output arguments, it returns the number of columns and rows
 *     in the table.  The function, itself, returns "void".
 *
 * PARAMETERS:
 *     num_maj  (unsigned int)    Number of elements along the table's major
 *                                axis.
 *     num_min  (unsigned int)    Number of elements along the table's minor
 *                                axis.
 *     major    (char *)          A string that indicates whether the table is
 *                                column major ("column") or row major ("row").
 *     num_cols (unsigned int *)  OUTPUT: The number of columns in the table.
 *     num_rows (unsigned int *)  OUTPUT: The number of rows in the table.
 *
 * RETURN VALUE:
 *     None.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Table data is in column major order.
 *        POST: The output arguments, num_cols & num_rows, are set to num_maj
 *              & num_min, respectively.
 *     2. PRE:  Table data is in row major order.
 *        POST: The output arguments, num_cols & num_rows, are set to num_min
 *              & num_maj, respectively.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     01/12/1999  J.D. Stroup  Original
 *     03/08/1999  J.D. Stroup  Changed type of "int" arguments to "unsigned
 *                              int".
 *
 ****************************************************************************/

static void cmpMajMin2ColRow (unsigned int num_maj, unsigned int num_min,
                              char *major,
                              unsigned int *num_cols, unsigned int *num_rows)
{
  if (strcmp (major, "column") == 0) /* table data in column major order */
  {
    *num_cols = num_maj;
    *num_rows = num_min;
  }
  else /* table data in row major order */
  {
    *num_rows = num_maj;
    *num_cols = num_min;
  }
  return;
}

/*****************************************************************************
 * 
 * NAME: cmpColRow2MajMin
 *
 * DESCRIPTION: This function converts the number of columns and rows in a
 *     table to the number of elements along a table's major and minor axes.
 *     It is passed the number of columns and rows in the table and an
 *     indication of whether the table is column major or row major.  Via
 *     output arguments, it returns the number of elements along the table's
 *     major and minor axes.  The function, itself, returns "void".
 *
 * PARAMETERS:
 *     num_cols (unsigned int)    The number of columns in the table.
 *     num_rows (unsigned int)    The number of rows in the table.
 *     major    (char *)          A string that indicates whether the table is
 *                                column major ("column") or row major ("row").
 *     num_maj  (unsigned int *)  OUTPUT: Number of elements along the table's
 *                                major axis.
 *     num_min  (unsigned int *)  OUTPUT: Number of elements along the table's
 *                                minor axis.
 *
 * RETURN VALUE:
 *     None.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Table data is in column major order.
 *        POST: The output arguments, num_maj & num_min, are set to num_cols
 *              & num_rows, respectively.
 *     2. PRE:  Table data is in row major order.
 *        POST: The output arguments, num_maj & num_min, are set to num_rows
 *              & num_cols, respectively.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     01/12/1999  J.D. Stroup  Original
 *     03/08/1999  J.D. Stroup  Changed type of "int" arguments to "unsigned
 *                              int".
 *
 ****************************************************************************/

static void cmpColRow2MajMin (unsigned int num_cols, unsigned int num_rows,
                              char *major,
                              unsigned int *num_maj, unsigned int *num_min)
{
  if (strcmp (major, "column") == 0) /* table data in column major order */
  {
    *num_maj = num_cols;
    *num_min = num_rows;
  }
  else /* table data in row major order */
  {
    *num_maj = num_rows;
    *num_min = num_cols;
  }
  return;
}

/*****************************************************************************
 * 
 * NAME: cmpElemMacExpand
 *
 * DESCRIPTION: Given a list of strings--one string in each element of the
 *     list, this function expands the macros in each string.  A new list of
 *     macro-expanded strings is returned.  See NOTES section below.
 *
 * PARAMETERS:
 *     pList (lstLIST)  On input it's a pointer to a list of strings.  On
 *                      output it's null.
 *
 * RETURN VALUE:
 *     cmpElemMacExpand (lstLIST)  A pointer to a list of macro-expanded
 *                                 strings.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     The given list is modified by this function.  Each string element is
 *     replaced with it macro-expanded equivalent.  The original string
 *     elements are freed.
 *
 * REVISION HISTORY:
 *     10/22/1998  J.D. Stroup  Original
 *     01/18/1999  J.D. Stroup  Instead of creating a new list--that is a
 *                              macro-expanded equivalent of the given
 *                              list--the given list is modified.  Each list
 *                              item (string) is replaced with its
 *                              macro-expanded string via lstReplaceCurrent.
 *                              Changed signature of function from
 *                                "static lstLIST cmpElemMacExpand (lstLIST
 *                                                                     pList)"
 *                              to
 *                                "static void cmpElemMacExpand (lstLIST
 *                                                                    pList)".
 *
 ****************************************************************************/

static void cmpElemMacExpand (lstLIST pList)
{
  char *string, *expand_string;

  /* Macro-expand each string of the given list.  Note that the original
     string from the given list is freed. */

  lstStopIteration (pList);
  while ((string = (char *) lstIterator (pList)) != NULL)
  {
    expand_string = macMacroExpander (string);
    lstReplaceCurrent (pList, (void *) expand_string);
    free (string);
  }

  return;
} /* end cmpElemMacExpand */

/*****************************************************************************
 * 
 * NAME: cmpTblDataDestructor
 *
 * DESCRIPTION: This is the destructor function for a table_data structure, as
 *     defined above in cmpTableProcessor.  It frees the items that comprise
 *     the table_data list.  These items are other list structures
 *     (lstLIST's), which point to either a column or row of the table.  Note
 *     that the items to which these "sublists" point are not freed.
 *
 * PARAMETERS:
 *     col_row_list (void *)  Pointer to a list (lstLIST).  This list
 *                            represents a major axis of a table (e.g., a
 *                            column in a column major table).
 *
 * RETURN VALUE:
 *     cmpTblDataDestructor (void *)  NULL
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     01/15/1999  J.D. Stroup  Original
 *
 ****************************************************************************/

static void *cmpTblDataDestructor (void *col_row_list)
{
  return (lstFreeList ((lstLIST) col_row_list));
}

/*****************************************************************************
 * 
 * NAME: cmpBuildIFDD
 *
 * DESCRIPTION: Builds a LUT table of ODL component definitions to be used
 *		as a data dictionary for the components of the Instructions file.
 *
 * PARAMETERS:  
 *     pIFDD -- (synDD) Pointer to the data dictionary structure.
 *
 * RETURN VALUE:
 *     cmpBuildIFDD -- (int) Status indicator: 1 -> success, 0 -> failure.
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. This function attempts to open and read IFDD.odl, the data 
 *        dictionary for the instructions file. If it fails to do that, it
 *        issues a debug message and returns Failure.
 *     2. If the DD is successfully read, synBuildDataDict is called to
 *        create the internal structure of the data dictionary. Failure
 *        to create the internal structure is treated as a fatal error.
 *
 * REVISION HISTORY:
 *   10/07/98  Ken Jarva  --  Original
 *   11/16/98  Ken Jarva  --  Cleanup.
 *
 ****************************************************************************/
static int cmpBuildIFDD( synDD pIFDD )
{
    FILE *pDDFile;
	AGGREGATE pDDTree;
	int Success = 1, Failure = 0;

	/* Open the file containing the data dictionary information: */
	pDDFile = fopen( "IFDD.odl", "r" );
	if( !pDDFile ) {
		cmpDebugMsg( "cmpBuildIFDD", "No DD for IF." );
		return Failure;
	}

  /* Create the root node of the data dictionary tree. */

	if ((pDDTree = NewAggregate (NULL, KA_GROUP, "DD_ROOT", NULL)) == NULL)
    cmpFatalError ("cmpBuildIFDD",
                   "Creation of the DD tree's root node failed");

  /* Read the data dictionary file, and create its tree under its root node. */

	if (ReadLabel (pDDFile, pDDTree) != Success)
		  cmpFatalError ("cmpBuildIFDD",
                  "ReadLabel failed to create the DD file's tree.");

  /* Build the data dictionary: */
	if( synBuildDataDict(pIFDD, pDDTree) ) return Success;
	else {
		cmpFatalError( "cmpBuildIFDD", "Failed to build the data dictionary for"
			" the IF." );
		return Failure; /* in case fatal error processing returns control here. */
	}
 
} /* end cmpBuildIFDD */


/* --------------------------------------------------------------------------*/
/* STUBS: temporary & develping functions used for testing purposes */

void cmpOutputRepStruc( lstLIST pReportList, int level)	
{ 
	/* Outputs the names of all the report nodes in the given list */
		char msg[256];
		REP_NODE pNode;
		char blanks[20] = "                   ";
		int index;

		level++;
		if( level > 19 ) level = 19;
		index = 19 - level;
		assert( pReportList);
		if( lstNumEl(pReportList) < 1 ) return;
		lstStopIteration( pReportList );
		pNode = lstIterator( pReportList );
		while( pNode ) {
			sprintf( msg, "\t%s%s", &blanks[index], pNode->pName );
			cmpDebugMsg(" ", msg );
			if( pNode->itemKind == cmpSUBLIST  ){
				                         
				assert(pNode->pItem);
				cmpOutputRepStruc( pNode->pItem, level );

			} else if( pNode->itemKind == cmpDATA ){
				char *pData;
				lstStopIteration( pNode->pItem );
				pData = lstIterator( pNode->pItem );
				while( pData ) {
					sprintf( msg, "\t%s%s", &blanks[index - 1], pData );
					cmpDebugMsg(" ", msg );
					pData = lstIterator( pNode->pItem );
				}
				lstStopIteration( pNode->pItem );
			}
			pNode = lstIterator( pReportList );
		}
		lstStopIteration( pReportList );
}

