/* FILE: lstList.c
 *
 * PROGRAM: Utility module
 *
 * DESCRIPTION
 *     These functions implement a general list processing capablitliy.
 *     The data structures involved are defined in lstList.h which should
 *     be included by any module for which this kind of list processing is
 *     desired.
 *
 *     In order to build a list, declare a pointer to a lstList struct 
 *     (The typedef lstLIST can be used to declare such a pointer.) The
 *     pointer should be initialized to lstLISTNULL.
 *
 *     lstInitList() will allocate a lstNode and set its attributes to 
 *         indicate the list is empty. This function SHOULD ALWAYS be used
 *         to initialize a new list.
 *
 *     lstFreeList can be used to free lists and the nodes that comprise 
 *         them.  This function DOES NOT free the items being listed since 
 *         they are data structures of unknown complexity. 
 *
 *     Items can be added before the first element or after the last in a list
 *         The first, current, or last element of the list can be 'plucked' so
 *         that the node is removed and a pointer to the item that had been
 *         stored there is returned. The following functions provide these
 *         functions:
 *
 *             lstAddFirst, lstAddLast, lstPluckFirst, lstPluckCurrSetPrev,
 *             lstPluckCurrSetNext, lstPluckLast
 *
 *     lstIterator steps to the 'next' node in the given list and return a 
 *         pointer to the data item associated with the node.
 *
 *         The 'next' node is determined by the list's current pointer.
 *         LstIterator modifies the current pointer in the lstList 
 *         struct; otherwise the list is left unchanged.
 *
 *     A similar function lstReverseIterator is provided so that the list
 *         can be traversed from the last to the first element.
 *
 *     A function lstAppendTo() is provided to allow one list to be appended 
 *         to another.
 *
 *     lstStopIteration can be used to terminate an ongoing iteration. It's
 *         important to do this if a list may be iterated more than once.
 *         It may be wise to invoke lstStopIteration before starting an on
 *         a particular list.
 *
 *     lstNumEl returns the number of elements currently in the list.    
 *      
 *
 * RCS INFO:
 *
 *      $Id: lstList.c,v 1.1 2002/02/01 20:04:19 bbane Exp $
 *
 *      $Log: lstList.c,v $
 *      Revision 1.1  2002/02/01 20:04:19  bbane
 *      Initial revision
 *
 *      Revision 1.3  1999/04/06 16:16:43  ToolsDev
 *      Promotion of branch 1.2.0 (cleared wlint warnings) for UFM release.
 *
 * Revision 1.2.0.1  1999/03/06  16:19:56  ToolsDev
 * Cleared lint warnings.
 *
 * Revision 1.2  1999/02/03  16:32:15  stroup
 * Promotion of 1.1.0 branch.
 *
 * Revision 1.1.0.5  1999/01/26  15:34:11  stroup
 * Replaced lstSplitList with SplitList from ListExpander.c, and renamed to
 * lstSplitList.  Changed "SplitList" & "FatalError" references to
 * "lstSplitList" & "lstFatalError", respectively.
 *
 * Revision 1.1.0.4  1999/01/25  21:35:40  jarva
 * Added call to lstStopIteration & improved documentation.
 *
 * Revision 1.1.0.3  1999/01/18  18:32:33  stroup
 * Add new function lstIsListOfEmptyStrings().
 *
 * Revision 1.1.0.2  1999/01/12  21:38:11  ToolsDev
 * New function lstDestructor().
 *
 * Revision 1.1.0.1  1999/01/12  20:34:11  stroup
 * Added lstJoinList, lstPluckCurrSetNext, & lstPluckCurrSetPrev.  Corrected bug in lstPluckFirst & lstPluckLast.  Some minor editing.
 *
 * Revision 1.1  1999/01/06  21:08:32  ToolsDev
 * Initial revision from prototype baseline.
 *
 *
 * NOTICE: this file was constructed from a template that supplied a
 *         control structure and some simple functions for reporting
 *         errors and debug information. See the static struct 
 *         lstCONTROLS below.
*/

/* INCLUDED FILES:
*/
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <malloc.h>
#include <string.h>

#include "lstList.h"
#include "strUtilities.h"

/* ---------------------------------------------------------------------*/
/* MACRO DEFINITIONS */

#define VERBOSE_DEBUG 1  /* Set to 1 for maximal debug output but also
                          * see lstCONTROLS.debug.

/* ---------------------------------------------------------------------*/
/* GLOBAL VARIABLES: the following variables may be used in other files.
 */
	/* NONE */

/* ---------------------------------------------------------------------*/
/* EXTERNAL VARIABLES: the following variables are defined in other files.
*/
	/* NONE */

/* ---------------------------------------------------------------------*/
/* STATIC VARIABLES: the following variables are for the exclusive use
 *     of the functions defined within this file.
*/
struct lstControls {
    /* Controls the message writing functions given in the source file
	 * template (e.g., lstDebugMsg().  This structure is
	 * imediately followed by its default settings. Typically the
	 * initailization function lstInitialization is invoked by
	 * the calling application to override these settings. */
	int debug;
	int verbose;
	int noExit;
	FILE *pDebugFile;
	FILE *pUserMessages;
} lstCONTROLS =
{
	1,
	1,
	0,
	NULL,
	NULL
};

/* ---------------------------------------------------------------------*/

/* STATIC PROTOTYPES: the following prototypes are for functions defined
 *     and used exclusively within this file.
*/
static void  lstFatalError( char *functionName, char *msg );
static void  lstUserWarning( char *functionName, char *msg );
static void  lstDebugMsg( char *functionName, char *msg );
static int   lstIsAList( lstLIST pList );

/* ---------------------------------------------------------------------*/
/*****************************************************************************
 * 
 * NAME: lstInitialization
 *
 * DESCRIPTION: Initializes the utilities control structure, overriding the
 *              established defaults. (See <<prefix>>Controls struct above.)
 *
 * PARAMETERS:  
 *     debugFlag     -- (int) Control to allow debug print to be output. Set
 *                      to 1 for ouput; 0 otherwise.
 *     verboseFlag   -- (int) Control to allow verbose messages for the user
 *                      to be output. Set to 1 for ouput; 0 otherwise.
 *	   noExit        -- (int) Control to prevent the functions within this
 *                      module from exiting the program.  Set to 1 and all
 *                      functions return to the caller, including 
 *                      lstFatalError(...).
 *     pDebugFile    -- (FILE *) Pointer to the debug file to which debug
 *                      messages (if allowed) will be written.  If passed
 *                      as NULL, stderr will be used if necessary.  Only
 *                      NULL or an active FILE pointer should be passed.
 *     pUserMessages -- (FILE *) Pointer to the debug file to which user
 *                      messages (if allowed) will be written.  If passed
 *                      as NULL, stderr will be used if necessary.  Only
 *                      NULL or an active FILE pointer should be passed.
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct lstCONTROLS:
 *     pDebugFile   -- (FILE*) The debug file to which error messages are to 
 *                     be written.
 *     debug        -- (int) control allowing debug messages to be output.
 *     noExit       -- (int) control preventing the function from exiting the
 *                      program.
 *     PUserMessage -- (FILE*) The file to which user messages are to be 
 *                     written.
 *     verbose      -- (int) control allowing verbose messages to be output.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void lstInitialization( int debugFlag, int verboseFlag, int noExit,
	FILE *pDebugFile, FILE *pUserMessages )
{
	lstCONTROLS.debug = debugFlag;
	lstCONTROLS.verbose = verboseFlag;
	lstCONTROLS.noExit = noExit;
	if( pDebugFile ) lstCONTROLS.pDebugFile = pDebugFile;
	else lstCONTROLS.pDebugFile = stderr;
	if( pUserMessages ) lstCONTROLS.pUserMessages = pUserMessages;
	else lstCONTROLS.pUserMessages = stderr;
	return;
} /* end lstInitialization */

/*****************************************************************************
 * 
 * NAME: lstFatalError
 *
 * DESCRIPTION: Reports fatal errors and shuts down the program. The debug
 *              control must be set to 1 in order for the message to be
 *              written, and preventExit must be set to 1 if the function
 *              is to be allowed to return to the caller. (See GLOBAL 
 *              VARIABLES below.) 
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct lstCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *     noExit     -- (int) control preventing the function from exiting the
 *                   program.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void lstFatalError( char *functionName, char *msg )
{
	if( lstCONTROLS.debug )
		fprintf( lstCONTROLS.pDebugFile, 
		                              "%s: FATAL - %s\n", functionName, msg );
	if( !lstCONTROLS.noExit ) {
		lstUserWarning( "lstFatalError", 
		     "Fatal programming error is causing the program to shut down." );
		exit(-1);
	}
	else return;

} /* end lstFatalError */

/*****************************************************************************
 * 
 * NAME: lstUserWarning
 *
 * DESCRIPTION: Reports warnings to the user if the verbose control is set.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct lstCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     verbose       -- (int) control allowing warning messages to be output.
 *	
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void lstUserWarning( char *functionName, char *msg )
{
	if( !lstCONTROLS.verbose ) return;

    fprintf( lstCONTROLS.pUserMessages, 
		                             "%s: WARNING - %s\n", functionName, msg );
	return;

} /* end lstUserWarning */
/*****************************************************************************
 * 
 * NAME: lstDebugMsg
 *
 * DESCRIPTION: Reports debug messages for the maintenance programmer if the
 *              debug flag is set and VERBOSE_DEBUG macro is 1.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct lstCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void lstDebugMsg( char *functionName, char *msg )
{
	if( lstCONTROLS.debug && VERBOSE_DEBUG ){
		fprintf( lstCONTROLS.pDebugFile, 
		                              "%s: INFO - %s\n", functionName, msg );
	    fflush(lstCONTROLS.pDebugFile);
	}
	return;

} /* end lstDebugMsg */
 
/* ---------------------------------------------------------------------*/
/* NAME: lstInitList
 *
 * RETURN: lstLIST
 *
 *     Returns a pointer to a lstList struct that's been allocated on
 *     the heap.
 *
 * DESCRIPTION:
 *
 *     Allocates a new lstList struct on the heap, sets the pointers that
 *     comprise the struct all to NULL, and returns the new struct's
 *     address.  Asserts that the allocation succeeded.
 *
 * PARAMETERS:
 *
 *     NONE
 *
 * GLOBAL VARIABLES:
 *
 *     lstLISTNULL
 *     lstNODENULL
 *
 * PRE- & POST-CONDITIONS:
 *
 *     1. PRE: room exists on the heap. POST: A new lstList struct has
 *        been allocated.
 *     2. PRE: sufficient room doesn't exist on the heap. POST: The
 *        program is terminated by assert().
 *
 * NOTES:
 *
 *     1. The lstList and its associated linked list should be freed
 *        using lstFreeList() when they are no longer needed.
 *
 * REVISION HISTORY:
 *
 *     1. Initial implementation: 9/09/98, Jarva.
 *
 * ---------------------------------------------------------------------
*/
lstLIST lstInitList( void ) 
{
    lstLIST pList = lstLISTNULL;  /* Points to new struct & is used
                                      to return the function's value. */
    pList = (lstLIST) malloc( sizeof(struct lstList) );

    if( pList == lstLISTNULL ) 
		lstFatalError( "lstInitList", 
		                  "malloc failure: cannot allocate root node" );

    pList->first = lstNODENULL;
    pList->last  = lstNODENULL;
    pList->current = lstNODENULL;
    pList->count = 0;

	strcpy( pList->isA, "LIST" );

    return pList;

} /* end lstInitList */

/* ---------------------------------------------------------------------
 *
 * NAME: lstAddFirst
 *
 * RETURN: void
 *
 * DESCRIPTION:
 *     Adds the given data item pointer to the beginning of the given
 *     list.
 *
 * PARAMETERS:
 *
 *     pItem: a void pointer to an arbitrary data item. The pointer will
 *         be added at the beginning of the list.
 *     pList: a pointer to a lstList struct.  This describes the linked
 *         list to which pItem is to be added.  pList must be
 *         non-NULL; otherwise, the program terminates.
 *
 * GLOBAL VARIABLES:
 *     None
 *
 * PRE- & POST-CONDITIONS:
 *
 *     1. PRE: there is sufficient space avaiable in the heap. POST:
 *        the structures are allocated, the pointers are set, and
 *        no value is returned.
 *     2. PRE: insufficient space on the heap exists. POST: the program
 *        terminates via assert().  This also applies to the space
 *        allocated by strdup().
 *
 * NOTES:
 *    1.  Assumes that the list was initialized by lstInitList().
 *    2.  Makes no change to pList->current. Adding to lists during
 *        iterations is not advised.
 *    3.  The list may be empty; so, in particular, pList->first may be
 *        NULL.
 *    4.  Also, if the list starts empty, pList->last must also change.
 *
 * REVISION HISTORY:
 *
 *     1. Initial implementation: 9/09/98, Jarva.
 *
 * ---------------------------------------------------------------------*/
void  lstAddFirst( lstLIST pList, void *pItem ) 
{
    lstNODE pNode = lstNODENULL; /* will point at the new node */

    /* Make sure that the list exists */
    if( !lstIsAList(pList) )
		lstFatalError( "lstAddFirst", "Cannot add to nonexistent list.");

    /* Allocate a new node for the list */
    pNode = (lstNODE) malloc( sizeof(struct lstNode) );
    assert( pNode != lstNODENULL );

    /* Associate node and data item */
    pNode->item = pItem;
    
    /* Place new node at the beginning of the list */
    pNode->next = pList->first; /* May be NULL */
    if( pNode->next )pNode->next->prev = pNode;
    pNode->prev = lstNODENULL;
    pList->first = pNode;
    (pList->count)++;
    if( pList->count == 1 ) pList->last = pNode;

    return;
}
/* ---------------------------------------------------------------------
 *
 * NAME: lstAddLast
 *
 * RETURN: void
 *
 * DESCRIPTION:
 *
 *     Adds the given data item pointer to the end of the given
 *     list.
 *
 * PARAMETERS:
 *
 *     pItem: a void pointer to an arbitrary data item. The pointer will
 *         be added at the end of the list.
 *     pList: a pointer to a lstList struct.  This describes the linked
 *         list to which string is to be added.  This pointer must be
 *         non-NULL; otherwise, the program terminates.
 *
 * GLOBAL VARIABLES:
 *     None
 *
 * PRE- & POST-CONDITIONS:
 *
 *     1. PRE: there is sufficient space avaiable in the heap. POST:
 *        the structures are allocated, the pointers are set, and
 *        no value is returned.
 *     2. PRE: insufficient space on the heap exists. POST: the program
 *        terminates via assert(). 
 *
 * NOTES:
 *    1.  Assumes that the list was initialized by lstInitList().
 *    2.  Makes no change to pList->current. Adding to lists during
 *        iterations is not advised.
 *    3.  The list may be empty; so, in particular, pList->last may be
 *        NULL.
 *    4.  Also, if the list starts empty, pList->first must also change.
 *
 * REVISION HISTORY:
 *
 *     1. Initial implementation: 9/09/98, Jarva.
 *
 * ---------------------------------------------------------------------*/
void  lstAddLast( lstLIST pList, void *pItem ) 
{
    lstNODE pNode = lstNODENULL; /* will point at the new node */

    /* That the list exists */
    if( !lstIsAList(pList) )
		lstFatalError( "lstAddLast", "Cannot add to nonexistent list.");
 
    /* Allocate a new node for the list */
    pNode = (lstNODE) malloc( sizeof(struct lstNode) );
    assert( pNode != lstNODENULL );

    /* Associate node and data item */
    pNode->item = pItem;
    
    /* Place new node at the end of the list */
    pNode->prev = pList->last;  /* May be NULL */
    if( pNode->prev) pNode->prev->next = pNode;
    pNode->next = lstNODENULL;
    pList->last = pNode;
    (pList->count)++;
    if( pList->count == 1 ) pList->first = pNode;

    return;
} /* end lstAddLast */
/* ---------------------------------------------------------------------
 *
 * NAME: lstIterator
 *
 * RETURN: void *
 * 
 *     Pointer to the data item in the list if there is one.
 *     If the list is empty or exhausted (due to one or more previous 
 *     calls to this function), NULL is returned.
 *
 * DESCRIPTION: 
 *
 *     Uses pList->current to determine the current position in the
 *     given list.  If the current pointer is NULL, lstIterator
 *     assumes that list processing is just starting, and it 
 *     makes the first node the current node, and it returns a
 *     pointer to the data item associated with that node. If current
 *     points to any node but the last, lstIterator makes the next
 *     node in the list 'current' and returns a pointer to its data 
 *     item. If the last node is the 'current' node, a NULL pointer is
 *     returned, and current does not change. Use lstStopIteration if
 *     you want to reset current for another iteration.
 *
 *     ***NOTICE*** You've got to be carful with this. For example if
 *     you're using this in a loop to find something, you can't just
 *     return when it's found...you've got to call lstStopIteration
 *     first or the next time you want to iterate this list you'll
 *     start where the search left off.
 *
 * PARAMETERS:
 *
 *     pList: pointer to the lstList struct that is being processed.
 *
 * GLOBAL VARIABLES:
 *
 *     NONE
 *      
 * PRE- & POST-CONDITIONS:
 *
 *     1. PRE: Processing is just beginning.
 *        POST: pList->current points at the first node; the data item 
 *        associated with that node has been returned. 
 *
 *     2. PRE: Processing is ongoing, list->current points at some
 *        node other than the last.
 *        POST: The next node in the list has been made
 *        the current node (i.e., it's now pointed to by
 *        list->current), and the void pointer associated with that node has
 *        been returned. 
 *
 *     3. PRE: The list has been exhausted; i.e., pList->current
 *        and pList->last are the same.
 *        POST: A NULL pointer is returned.  No change is made to 
 *        pList->current.
 *
 *     4. PRE: The list is non-existent; i.e., pList == NULL. 
 *        POST: The NULL pointer is returned.
 *
 * NOTES:
 *
 *     1. NONE
 *
 * REVISION HISTORY:
 *
 *     1. Initial implementation: 9/09/98, Jarva.
 *
 * ---------------------------------------------------------------------*/
void *lstIterator( lstLIST pList )
{
    lstNODE current;
    void    *pItem = (void *) NULL; /* pointer to the data item. */

    if( !lstIsAList(pList) ){ 
        /* The list is non-existent. */
        lstFatalError( "lstIterator", 
			"Cannont iterate a nonexistent list." );
    } else current = pList->current;

    if( current == lstNODENULL ){
        /* The iteration is just starting. */
        current = pList->first;
        if( current != lstNODENULL )pItem = current->item;
    } else {
        /* The iteration is ongoing. */
        if( current == pList->last ){
            /* The list has been exhausted or is empty. */
            pItem = (void *) NULL;
            /* No change to current. */
        } else {
            /* This is the normal condition. */
            current = current->next; /* Moved to next node. */
            assert( current ); /* Should never be NULL here. */
            pItem = current->item; /* Points to the string we want. */
        }
    }
    pList->current = current; /* Update list's current node pointer. */
    return pItem;
} /* end lstIterator */
/* ---------------------------------------------------------------------
 *
 * NAME: lstReverseIterator
 *
 * RETURN: void *
 * 
 *     Pointer to the data item in the list if there is one.
 *     If the list is empty or exhausted (due to one or more previous 
 *     calls to this function), NULL is returned.
 *
 * DESCRIPTION: 
 *
 *     Uses pList->current to determine the current position in the
 *     given list.  If the current pointer is NULL, lstReverseIterator
 *     assumes that list processing is just starting, and it 
 *     makes the last node the current node, and it returns a
 *     pointer to the data item associated with that node. If current
 *     points to any node but the first, lstIterator makes the previous
 *     node in the list 'current' and returns a pointer to its data 
 *     item. If the first node is the 'current' node, a NULL pointer is
 *     returned, and current does not change. Use lstStopIteration if
 *     you want to reset current for another iteration.
 *
 *     ***NOTICE*** You've got to be carful with this. For example if
 *     you're using this in a loop to find something, you can't just
 *     return when it's found...you've got to call lstStopIteration
 *     first or the next time you want to iterate this list you'll
 *     start where the search left off.
*
 * PARAMETERS:
 *
 *     pList: pointer to the lstList struct that is being processed.
 *
 * GLOBAL VARIABLES:
 *
 *     NONE
 *      
 * PRE- & POST-CONDITIONS:
 *
 *     1. PRE: Processing is just beginning.
 *        POST: pList->current points at the last node; the data item 
 *        associated with that node has been returned. 
 *
 *     2. PRE: Processing is ongoing, list->current points at some
 *        node other than the first.
 *        POST: The previous node in the list has been made
 *        the current node (i.e., it's now pointed to by
 *        list->current), and the void pointer associated with that node has
 *        been returned. 
 *
 *     3. PRE: The list has been exhausted; i.e., pList->current
 *        and pList->first are the same.
 *        POST: A NULL pointer is returned.  No change is made to 
 *        pList->current.
 *
 *     4. PRE: The list is non-existent; i.e., pList == NULL. 
 *        POST: The NULL pointer is returned.
 *
 * NOTES:
 *
 *     1. NONE
 *
 * REVISION HISTORY:
 *
 *     1. Initial implementation: 9/11/98, Jarva.
 *
 * ---------------------------------------------------------------------*/
void *lstReverseIterator( lstLIST pList )
{
    lstNODE current;
    void    *pItem = (void *) NULL; /* pointer to the data item. */

    if( !lstIsAList(pList) ){ 
        /* The list is non-existent. */
        lstFatalError( "lstReverseIterator", 
			"Cannot iterate an nonexistent list.");
    } else current = pList->current;

    if( current == lstNODENULL ){
        /* The iteration is just starting. */
        current = pList->last;
        if( current != lstNODENULL )pItem = current->item;
    } else {
        /* The iteration is ongoing. */
        if( current == pList->first ){
            /* The list has been exhausted or is empty. */
            pItem = (void *) NULL;
            /* No change to current. */
        } else {
            /* This is the normal condition. */
            current = current->prev; /* Moved to previous node. */
            assert( current ); /* Should never be NULL here. */
            pItem = current->item; /* Points to the string we want. */
        }
    }
    pList->current = current; /* Update list's current node pointer. */
    return pItem;
} /* end lstReverseIterator */
/*****************************************************************************
 * 
 * NAME: lstStopIteratrion
 *
 * DESCRIPTION: Adjusts the lstList structure so as to indicate that no
 *     iteration is on going.
 *
 * PARAMETERS:  
 *     pList -- lstLIST points at the list for which the iteration is to be
 *              terminated.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES: None
 *	
 * PRE- & POST-CONDITIONS:
 *     <<description of conditions>>
 *
 * NOTES:
 *     <<description>>
 *
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

void lstStopIteration( lstLIST pList )
{
    if( !lstIsAList( pList ) )
		lstFatalError( "lstStopIteration", 
			"Cannot operate on nonexistent list.");
	pList->current = lstNODENULL;
	return;

} /* end lstStopIteratrion */
/* ---------------------------------------------------------------------
 *
 * NAME: lstFreeList
 *
 * RETURN: lstLIST      
 * 
 *     Returned value will always be lstLISTNULL.
 *
 * DESCRIPTION: 
 *
 *     Steps through the lstNodes in the linked list, freeing the
 *     Nodes. Finally, it frees the lstList struct itself and returns 
 *     lstLISTNULL.
 *
 *     ***NOTICE*** the arbitrary data structures to which the list nodes
 *     point are not freed.  It's assumed that, if necessary, that has been
 *     taken care of previously.
 *
 * PARAMETERS:
 *
 *     pList: Pointer to the lstList struct whose components are to be 
 *	          freed.
 *
 * GLOBAL VARIABLES:
 *	   None
 *	
 * PRE- & POST-CONDITIONS:
 *
 *     1. PRE: normal lstLIST exists. POST: All the component pieces
 *	  of the list have beed deallocated using free, and a NULL
 *        pointer has been returned.
 *     2. PRE: empty or non-existent list has been passed. POST: A
 *	  NULL pointer is returned.
 *
 * NOTES:
 *
 *     1. USAGE: pList = lstFreeList( pList ) will free the various
 *	  components of the list and NULL the lstList pointer.
 *
 * REVISION HISTORY:
 *
 *     1. Initial implementation: 9/10/98, Jarva.
 *
 * ---------------------------------------------------------------------*/
lstLIST lstFreeList( lstLIST pList )
{
    lstNODE pNode; /* points at next node to be freed. */
    lstNODE pTmp;  /* temporary pointer. */

    if( pList == lstLISTNULL ) return pList; /* There was no purpose
                                                   to this call. */
	if( !lstIsAList(pList) )
		lstFatalError( "lstFreeList", "Bad list encountered." );
    pNode = pList->first; /* Now points at the first pNode. */

    while( pNode != lstNODENULL ) {
	    /* free each each lstNode in the list */
        pTmp = pNode->next;
	    free( pNode );
	    pNode = pTmp;
    }

    free( pList );
    return lstLISTNULL;
} /* end lstFreeList */
/*****************************************************************************
 * 
 * NAME: lstPluckFirst
 *
 * DESCRIPTION: Extracts the first node from the list and
 *    returns the pointer that it contains. The node is freed.
 *
 * PARAMETERS:  
 *     pList -- lstLIST Points at the list from which the data is to be 
 *         extracted
 *
 * RETURN VALUE:
 *     void pointer stored in the list's first element if it exists, NULL
 *         otherwise.
 *     
 * GLOBAL VARIABLES:
 *     None
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE: The list exists and has at least one element. POST: The first
 *        element is popped from the list and its item pointer is returned to
 *        the caller.  If there was only one element, the list is now empty.
 *     2. PRE: The list is empty. POST: A NULL pointer is
 *        returned to the caller.
 *     3. PRE: The list doesn't exist or has been improperly initialized.
 *        POST: An error message is written and the program terminates.
 *
 * NOTES:
 *     None
 *
 * REVISION HISTORY:
 *   09/10/98 Ken Jarva             Raytheon
 *   01/11/99 J.D. Stroup  Set the "previous" pointer of the first node in the
 *                         list to NULL, if the list is non-empty after the
 *                         pluck.
 *
 ****************************************************************************/

void *lstPluckFirst( lstLIST pList )
{
	void *pItem;
	lstNODE pNode;


	/* If the list doesn't exist, abort program */
	if( !lstIsAList(pList) )
		lstFatalError( "lstPluckFirst",
			"Cannot operate on nonexistent list.");

    /* If it's empty, return NULL */
	if( !(pList->first) ) return (void *) NULL;

  pNode = pList->first;
	pItem = pNode->item;
	pList->first = pNode->next;
	(pList->count)--;

  /* If the list is now empty, set the pointer to the last node to NULL.
     Otherwise, reset the "previous" pointer of the first node in the list to
     NULL. */

	if( pList->count == 0 )
    pList->last = (lstNODE) NULL;
  else
    pList->first->prev = (lstNODE) NULL;

	free( pNode );
  return pItem;

} /* end lstPluckFirst */

/*****************************************************************************
 * 
 * NAME: lstPluckLast
 *
 * DESCRIPTION: Extracts the last node from the list and
 *    returns the pointer that it contains. The node is freed.
 *
 * PARAMETERS:  
 *     pList -- lstLIST Points at the list from which the data is to be 
 *         extracted
 *
 * RETURN VALUE:
 *     void pointer stored in the list's first element if it exists, NULL
 *         otherwise.
 *     
 * GLOBAL VARIABLES:
 *     None
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE: The list exists and has at least one element. POST: The last
 *        element is popped from the list and its item pointer is returned to
 *        the caller.  If there was only one element, the list is now empty.
 *     2. PRE: The list is empty. POST: A NULL pointer is
 *        returned to the caller.
 *     3. PRE: The list doesn't exist or has been improperly initialized.
 *        POST: An error message is written and the program terminates.
 *
 * NOTES:
 *     None
 *
 * REVISION HISTORY:
 *   09/10/98 Ken Jarva             Raytheon
 *   01/11/99 J.D. Stroup  Set the "next" pointer of the last node in the list
 *                         to NULL, if the list is non-empty after the pluck.
 *
 ****************************************************************************/

void *lstPluckLast( lstLIST pList )
{
	void *pItem;
	lstNODE pNode;


	/* If the list doesn't exist, abort program */
	if( !lstIsAList(pList) )
		lstFatalError( "lstPluckLast",
			"Cannot operate on nonexistent list.");

    /* If it's empty, return NULL */
	if( !(pList->first) ) return (void *) NULL;

  pNode = pList->last;
	pItem = pNode->item;
	pList->last = pNode->prev;
	(pList->count)--;

  /* If the list is now empty, set the pointer to the first node to NULL.
     Otherwise, reset the "next" pointer of the last node in the list to
     NULL. */

	if( pList->count == 0 )
    pList->first = (lstNODE) NULL;
  else
    pList->last->next = (lstNODE) NULL;

	free( pNode );
  return pItem;

} /* end lstPluckLast */

/*****************************************************************************
 * 
 * NAME: lstPluckCurrSetNext
 *
 * DESCRIPTION: Extracts the current node from the list and returns the
 *     pointer that it contains.  The node is freed.  The pointer to the
 *     current node in the given list is reset to point to the node that
 *     followed the just plucked node, or NULL if no node followed it.
 *
 * PARAMETERS:  
 *     pList -- lstLIST Points at the list from which the data is to be 
 *         extracted
 *
 * RETURN VALUE:
 *     void pointer stored in the list's current element if it exists, NULL
 *         otherwise.
 *     
 * GLOBAL VARIABLES:
 *     None
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  The list doesn't exist or has been improperly initialized.
 *        POST: An error message is written and the program terminates.
 *     2. PRE:  The list is empty, or no iteration is being performed on the
 *              list (i.e., the pointer to the list's current node is NULL).
 *        POST: A NULL pointer is returned to the caller.
 *     3. PRE:  The list exists and has only one element.
 *        POST: The list's only element is popped from the list, and its item
 *              pointer is returned to the caller.  The list is now empty.
 *     4. PRE:  The list exists and has more than one element.
 *        POST: The element of the list pointed to by the current node pointer
 *              is popped from the list, and its item pointer is returned to
 *              the caller.  The two nodes that are adjacent to the current
 *              node are updated accordingly to now indicate that no node
 *              exists between them.  The number of element in the list is now
 *              one less than before the call.  The new current node points to
 *              the node immediately following the node just plucked.
 *
 * NOTES:
 *     In order for a valid--i.e., non-NULL--pointer to be returned by this
 *     function when working with a non-empty list, then the list must be in
 *     the process of being iterated on.  That is, a previous call to
 *     lstIterator or lstReverseIterator must have been made without an
 *     intervening call to lstStopIteration.
 *
 * REVISION HISTORY:
 *     01/11/1999  J.D. Stroup  Original.
 *
 ****************************************************************************/

void *lstPluckCurrSetNext (lstLIST pList)
{
  void *pItem;
  lstNODE curr_node;

	/* If the list doesn't exist, abort program */
	if (! lstIsAList (pList))
		lstFatalError ("lstPluckCurrSetNext",
			             "Cannot operate on nonexistent list.");

  curr_node = pList->current;

  /* If the current node is NULL, the returned item will be NULL. */

  if (curr_node == lstNODENULL)
    pItem = NULL;

  /* If the current node is pointing to the first or last item in the list,
     then call lstPluckFirst or lstPluckLast to pluck and return this item.
     Reset the current node to point to the node in the list following the
     node it was originally pointing to.  In the case where the current node
     was pointing to the first item in the list, it is reset to again point to
     the first item in the list--this item was the second item in the list
     before the pluck took place.  In the case where the current node was
     pointing to the last item in the list, it is reset to NULL, since no
     node follows the last node in the list. */

  else if (curr_node == pList->first)
  {
    pItem = lstPluckFirst (pList);
    pList->current = pList->first;
  }

  else if (curr_node == pList->last)
  {
    pItem = lstPluckLast (pList);
    pList->current = NULL;
  }

  /* Otherwise, the current node is pointing to a node in the middle of the
     list. */

  else
  {
    lstNODE next_node, prev_node;

    pItem = curr_node->item;

    /* Since the current node is going away, update the pointers of the
       current node's previous and next nodes.  The node in the list following
       the current node's previous node (prev_node) is the current node's next
       node (next_node).   The node in the list preceeding next_node is
       prev_node. */

    prev_node = curr_node->prev;
    next_node = curr_node->next;

    prev_node->next = next_node;
    next_node->prev = prev_node;

    /* Reset current to now point to next_node. */

    pList->current = next_node;

    /* Decrement the number of items in the list by one, and free the old
       current node. */

	  (pList->count)--;

    free (curr_node);
  }

  return (pItem);
} /* end lstPluckCurrSetNext */

/*****************************************************************************
 * 
 * NAME: lstPluckCurrSetPrev
 *
 * DESCRIPTION: Extracts the current node from the list and returns the
 *     pointer that it contains.  The node is freed.  The pointer to the
 *     current node in the given list is reset to point to the node that
 *     precedes the just plucked node, or NULL if no node followed it.
 *
 * PARAMETERS:  
 *     pList -- lstLIST Points at the list from which the data is to be 
 *         extracted
 *
 * RETURN VALUE:
 *     void pointer stored in the list's current element if it exists, NULL
 *         otherwise.
 *     
 * GLOBAL VARIABLES:
 *     None
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  The list doesn't exist or has been improperly initialized.
 *        POST: An error message is written and the program terminates.
 *     2. PRE:  The list is empty, or no iteration is being performed on the
 *              list (i.e., the pointer to the list's current node is NULL).
 *        POST: A NULL pointer is returned to the caller.
 *     3. PRE:  The list exists and has only one element.
 *        POST: The list's only element is popped from the list, and its item
 *              pointer is returned to the caller.  The list is now empty.
 *     4. PRE:  The list exists and has more than one element.
 *        POST: The element of the list pointed to by the current node pointer
 *              is popped from the list, and its item pointer is returned to
 *              the caller.  The two nodes that are adjacent to the current
 *              node are updated accordingly to now indicate that no node
 *              exists between them.  The number of element in the list is now
 *              one less than before the call.  The new current node points to
 *              the node immediately preceding the node just plucked.
 *
 * NOTES:
 *     In order for a valid--i.e., non-NULL--pointer to be returned by this
 *     function when working with a non-empty list, then the list must be in
 *     the process of being iterated on.  That is, a previous call to
 *     lstIterator or lstReverseIterator must have been made without an
 *     intervening call to lstStopIteration.
 *
 * REVISION HISTORY:
 *     01/12/1999  J.D. Stroup  Original.
 *
 ****************************************************************************/

void *lstPluckCurrSetPrev (lstLIST pList)
{
  void *pItem;
  lstNODE curr_node;

	/* If the list doesn't exist, abort program */
	if (! lstIsAList (pList))
		lstFatalError ("lstPluckCurrSetPrev",
			             "Cannot operate on nonexistent list.");

  curr_node = pList->current;

  /* If the current node is NULL, the returned item will be NULL. */

  if (curr_node == lstNODENULL)
    pItem = NULL;

  /* If the current node is pointing to the first or last item in the list,
     then call lstPluckFirst or lstPluckLast to pluck and return this item.
     Reset the current node to point to the node in the list preceding the
     node it was originally pointing to.  In the case where the current node
     was pointing to the first item in the list, it is reset to NULL, since no
     node precedes the first node in the list.  In the case where the current
     node was pointing to the last item in the list, it is reset to again
     point to the last item in the list--this item was the second-from-last
     item in the list before the pluck took place. */

  else if (curr_node == pList->first)
  {
    pItem = lstPluckFirst (pList);
    pList->current = NULL;
  }

  else if (curr_node == pList->last)
  {
    pItem = lstPluckLast (pList);
    pList->current = pList->last;
  }

  /* Otherwise, the current node is pointing to a node in the middle of the
     list. */

  else
  {
    lstNODE next_node, prev_node;

    pItem = curr_node->item;

    /* Since the current node is going away, update the pointers of the
       current node's previous and next nodes.  The node in the list following
       the current node's previous node (prev_node) is the current node's next
       node (next_node).   The node in the list preceeding next_node is
       prev_node. */

    prev_node = curr_node->prev;
    next_node = curr_node->next;

    prev_node->next = next_node;
    next_node->prev = prev_node;

    /* Reset current to now point to prev_node. */

    pList->current = prev_node;

    /* Decrement the number of items in the list by one, and free the old
       current node. */

	  (pList->count)--;

    free (curr_node);
  }

  return (pItem);
} /* end lstPluckCurrSetPrev */

/*****************************************************************************
 * 
 * NAME: lstAppendTo
 *
 * DESCRIPTION: Appends the contents of a list (pList2) to the end of another 
 *              list (pList1).  The list from which the inforamtion is
 *              retrieved is left unchanged. The void pointers from both
 *              lists point at the same data items.
 *
 * PARAMETERS:  
 *     pList1 -- lstLIST The list to which information it to be appended.
 *     pList2 -- lstLIST The list from which the information that is to be
 *               appended is retrieved.
 *
 * RETURN VALUE: 
 *     listAppendTo -- int If successfull, the number of elements in the
 *               augmented list. If either of the input lists is nonexistent,
 *               a -1 is returned.
 *     
 * GLOBAL VARIABLES: None
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE: One or both of the pointers passed to the function doesn't
 *        point at a list structure. POST: Fatal error processing is
 *        exercised.
 *     2. PRE: Pointers to two valid lists are passed to the function. POST:
 *        The NULL pointers from the nodes of the second list (pList2) are
 *        copied to new nodes which are then added to the end of the first
 *        list (pList1).  Notice that both lists now refer at least in part
 *        to exactly the same data items.  It would not therefore make sense
 *        to use lstDestructor with a special pupose destructor function
 *        until both lists were no longer of any use.  If pList2 is no longer
 *        of use, free the nodes with lstFreeList.
 *
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva  Original
 *   01/25/99  Ken Jarva  Added comments & inserted call to lstStopIteration.
 *
 ****************************************************************************/

int lstAppendTo( lstLIST pList1, lstLIST pList2 )
{
	int newNodeCounter = 0;
	void *pWhatever;

	if( pList1 == lstLISTNULL || pList2 == lstLISTNULL )
		return -1;

	if( !lstIsAList(pList1) )
		lstFatalError( "lstAppendTo",
			"First arguement represents nonexistent list.");

	if( !lstIsAList(pList2) )
		lstFatalError( "lstAppendTo",
			"Second arguement represents nonexistent list.");

	lstStopIteration( pList2 );
	pWhatever = lstIterator( pList2 );
	while( pWhatever != (void *) NULL )
	{
		newNodeCounter++;
		lstAddLast( pList1, pWhatever );

		pWhatever = lstIterator( pList2 );
	}
	lstStopIteration( pList2 );
	return pList1->count;

} /* end lstAppendTo */
/*****************************************************************************
 * 
 * NAME: lstNumEl
 *
 * DESCRIPTION: Returns the number of elements in the given list.
 *
 * PARAMETERS:  
 *     pList -- lstLIST Pointer to list of interest.
 *
 * RETURN VALUE: integer giving number of elements in the list
 *     
 * GLOBAL VARIABLES: None
 *	
 * PRE- & POST-CONDITIONS:
 *     <<description of conditions>>
 *
 * NOTES:
 *     <<description>>
 *
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

int lstNumEl( lstLIST pList )
{
	if( lstIsAList(pList) ) return pList->count;
	return 0;

} /* end lstNumEl */
/*****************************************************************************
 * 
 * NAME: lstIsStrInList
 *
 * DESCRIPTION: Determines whether the given sting exists within the given 
 *              list.
 *
 *		***NOTICE*** This function only applies to lists in which the
 *              'arbiratray' data item is in fact a string.
 *
 * PARAMETERS:  
 *     pList -- lstLIST Points to the list in question.
 *     pStr  -- (char *) Points to the string that is to be matched.
 *
 * RETURN VALUE: integer  0 -> the string is not in the list; 1-> it is.
 *     
 * GLOBAL VARIABLES: None
  *	
 * PRE- & POST-CONDITIONS:
 *     <<description of conditions>>
 *
 * NOTES:
 *     <<description>>
 *
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

int lstIsStrInList( lstLIST pList, char *pStr )
{
	char *pTestStr;

	assert( pList != lstLISTNULL && pStr != (char *) NULL );
	if( !lstIsAList(pList) ) lstFatalError( "lstIsStrInList",
		"Passed bad list pointer.");
	if( pStr == (char *) NULL ) lstFatalError( "lstIsStrInList",
		"Passed bad string pointer." );

	/* Get the first string in the list */
	pTestStr = (char *)lstIterator( pList );

	/* loop until the string is found or the list is exhausted...*/
	while( pTestStr != (char *) NULL ) {
		if( strToUpStrcmp(pStr, pTestStr) == 0 ){
			lstStopIteration( pList );
			return 1;
		}
		
		/* Get the next string in the list, if any...*/
		pTestStr = (char *)lstIterator( pList );
	}

	/* A match was not made...*/
	lstStopIteration( pList );
	return 0;

} /* end lstIsStrInList */

/* --------------------------------------------------------------------------*/

static int lstIsAList( lstLIST pList )
{
	/* Checks that the given list has be properly initialized. See 
	 * lstInitList.
	*/
	int i;
	char marker[5] = "LIST";

	if( pList == lstLISTNULL ) return 0;
	for( i = 0; i < 3; i++ ) if( pList->isA[i] != marker[i] ) return 0;

	return 1;
}

void *lstReplaceCurrent( lstLIST pList, void *pItem )
{
	/* 10/24/98:
	 * Replace the listed item of the current node; return a pointer to
	 * the item that was listed there (possibly for freeing). If the
	 * list is not being itereated or if the current pointer is NULL for
	 * any other reason, return NULL. */
	void *pOriginalItem;

	assert( pList );

	/* If the pointer to the current node is NULL, return NULL;
	 * otherwise:
	 *   save the item pointer of the current node;
	 *   set the item pointer of the current node pItem;
	 *   return the saved item pointer;
	 */
	 if( pList->current == NULL ){
		 pOriginalItem = NULL;
		 lstDebugMsg( "lstReplaceCurrent", 
			 "Unable to replace current (NULL) in list.");
	 } else {
		pOriginalItem = pList->current->item;
		pList->current->item = pItem;
	}
	return pOriginalItem;

}

/*****************************************************************************
 * 
 * NAME: lstMatchList
 *
 * DESCRIPTION: Searches the lstList for entries which include the given 
 *      indexString as the first strlen( indexString ) characeters. It returns
 *      a list of entries from the list for which this is the case.
 *
 * PARAMETERS:  
 *     pList        -- lstLIST The list to be searched.
 *     indexString -- (char *) Points at the string to be matched.
 *
 * RETURN VALUE: (void *) The data item assoicated with the given index string.
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. Note the use of the list interator. Be sure that lstStopIteration is
 *        called prior to exiting this function.
 *     2. Set printCount to 1 in order to get a debug message whenever more
 *        than one element of pList starts with the indexString.
 *
 * REVISION HISTORY:
 *   11/23/98  Ken Jarva -- Original ( Cloned from sltMatchList() ).
 *   12/02/98  Ken Jarva -- added debuggery controls.
 *
 ****************************************************************************/
lstLIST lstMatchList( lstLIST pList, char *indexString )
{
	lstLIST pNames;
	char *pEntry;
	int /* Debuggery controls...*/
		printCount = 0;  /* Set to one to print message when more than one
						  * match is found. */

	pNames = lstInitList(); /* The indexes that start with indexString */

	lstStopIteration( pList ); /* mere precaution */

	/* Search the LUT for an entry with an index that matches
	 * indexString. */
	pEntry = lstIterator( pList );
	while( pEntry != NULL ) {
		char *pFirstPart = strdup( pEntry );
		unsigned int n = strlen(indexString);
		if( strlen(pFirstPart) >= n ) pFirstPart[n] = '\0';
		if( strToUpStrcmp(pFirstPart, indexString) == 0 ){
			/* We've found a match. */
			lstAddLast( pNames, pEntry );
		}
		/* Continue checking all entires of pList: */
		pEntry = lstIterator( pList );
		free( pFirstPart );
	}

	/* Terminate the iteration before exiting function: */
	lstStopIteration( pList );

	if( printCount && lstNumEl( pNames ) > 1 ) {
		char msg[256];
		sprintf( msg, "Have %d entries starting with %s\n", 
			lstNumEl(pNames), indexString );
		lstDebugMsg("lstMatchList", msg );
	}

	return pNames;

} /* end lstMatchList */

/*****************************************************************************
 * 
 * NAME: lstSplitList
 *
 * DESCRIPTION: This function splits the given input string into zero or more
 *     tokens, or fields.  The string is split on a delimiter character that
 *     is also given as input.  The tokens are placed into a linked list data
 *     structure.  This function returns a pointer to the beginning of the
 *     linked list.
 *
 *     The following table shows the results of lstSplitList given a variety
 *     of input strings.  The examples in this table are representative of the
 *     possible input to lstSplitList.  In all examples the colon (":") is
 *     assumed to be the delimiter character.
 *
 *       Input String      Results                   Comments
 *     -----------------  ----------  ---------------------------------------
 *     "a:b"              {"a", "b"}  a list with two elements
 *     "a:"               {"a"}       a list with one element
 *     ":b"               {"", "b"}   a list with two elements (the first one
 *                                    is the empty string)
 *     ":"                {}          an empty, list
 *     "a"                {"a"}       a list with one element
 *     "" (empty string)  {}          an empty, list
 *
 * PARAMETERS:
 *     delimiter (char)           The delimiter character.
 *     delimited_string (char *)  The string containing 1 or more fields
 *                                (tokens) separated by the delimiter
 *                                character (delimiter).
 *
 * RETURN VALUE:
 *     lstSplitList (lstLIST)  Pointer to a linked list, such that each
 *                             element of the list contains one field (token)
 *                             from the input string.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Pointer to the input delimited string is NULL.
 *        POST: Function returns an empty list.
 *     2. PRE:  Input delimited string is empty ("").
 *        POST: Function returns an empty list.
 *     3. PRE:  Input delimited string contains delimiters, but--after leading
 *              and trailing white space removed--no tokens (e.g., assuming
 *              the delimiter is a comma, then ",, ,").
 *        POST: Function returns an empty list.
 *     4. PRE:  Input delimited string contains tokens and delimiters..
 *        POST: Function returns a linked list of type lstLIST, such that each
 *              element of the list contains a token from the input string.
 *     5. PRE:  Insufficient memory on heap.
 *        POST: Function issues an error message and terminates the program
 *              (via a call to FatalError).
 *
 * NOTES:
 *     Memory is allocated by this function during calls to lstInitList,
 *     strNextToken, strRightTrim, and lstAddLast.  If the allocation fails in
 *     lstInitList, lstAddLast, or strRightTrim, a message is generated and
 *     the program terminates.  The function strNextToken simply return a NULL
 *     pointer if it encounters a memory allocation failure.
 *
 * REVISION HISTORY:
 *     09/17/1998  J.D. Stroup  Original
 *     01/22/1999  J.D. Stroup  Replaced calls to FreeAllList with calls to
 *                              lstDestructor.  Replace the void pointer vp
 *                              with the character pointer trimmed_token, and
 *                              changed its scope from "global" for the
 *                              function to local for the "for" loop.
 *     01/26/1999  J.D. Stroup  Replaced version of lstSplitList that was
 *                              resident in this module with the SplitList
 *                              function from ListExpander.c.  Changed
 *                              "SplitList" and "FatalError" references to
 *                              "lstSplitList" and "lstFatalError",
 *                              respectively.
 *
 ****************************************************************************/

lstLIST lstSplitList (char delimiter, char *delimited_string)
{
  lstLIST pList;
  char    *token;
  char    delimiters[2];
  int     ListAllEmpty = 1;

  delimiters[0] = delimiter;
  delimiters[1] = '\0';

  /* Initialize the list to be returned. */

  pList = lstInitList ();

  /* Retrieve each field (token) from the input delimited string.  As each
     token is returned, add it to the list we're creating.  Remember that a
     void pointer must point to the item being listed. */

  for (token = strNextToken (delimited_string, delimiters);
       token != NULL;
       token = strNextToken (NULL, delimiters))
  {
    char *trimmed_token = strRightTrim (strLeftTrim (token));

    /* Make sure memory was successfully allocated for the "trimmed" token.
       If not, issue an error message and terminate the program. */

    if (trimmed_token == NULL)
      lstFatalError ("lstSplitList", "malloc failure in strRightTrim/strrev: "
                                     "cannot allocate memory for string");

    free (token);

    /* Add the new token to the list. */
    lstAddLast (pList, (void *) trimmed_token);

    /* Determine if we've encountered a non-empty string yet. */
    if (ListAllEmpty && (strlen (trimmed_token) > 0)) ListAllEmpty = 0;
  }

  /* If all the elements in the list contain "empty" tokens, then the
     given string contains no tokens.  Thus a list has been created such that
     each element contains an empty string, or token.  As a result we want to
     return an empty list, so we free the list just created, and create a new
     empty list to be returned. */

  if (ListAllEmpty)
  {
    pList = lstDestructor (pList, strDestructor);
    pList = lstInitList ();
  }

  return (pList);

} /* end lstSplitList */

/*****************************************************************************
 *
 * NAME: lstJoinList
 *
 * DESCRIPTION: This function joins each element of a given list into a single
 *     string.  Each element of the list is a string itself, and each
 *     element's string is separated from the next by the given delimiter
 *     string.  This function does the opposite of lstSplitList (see NOTES
 *     section, below).
 *
 * PARAMETERS:  
 *     delimiter (char *)  The delimiter string.  If NULL, then NULL is
 *                         returned by this function.
 *     plist (lstLIST)     Pointer to a linked list, such that each element of
 *                         the list contains a string.  If NULL, then NULL is
 *                         returned by this function.
 *
 * RETURN VALUE:
 *     lstJoinList (char *)  A string containing 1 or more fields (items from
 *                           from the given list) separated by the given
 *                           delimiter string.
 *     
 * GLOBAL VARIABLES:
 *	   None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Pointer to given list is NULL.
 *        POST: Function returns NULL.
 *     2. PRE:  Pointer to given delimiter is NULL, or the delimiter is the
 *              empty string.
 *        POST: Function returns a string made up solely of the elements in
 *              the given list concatenated together.
 *     3. PRE:  Given list contains no elements.
 *        POST: Function returns the empty string ("").
 *     4. PRE:  Given list contain 1 element.
 *        POST: Function returns that element.
 *     5. PRE:  Given list contains n elements, where n > 1.
 *        POST: Function returns a string made up of n fields separated by the
 *              given delimiter string.
 *     6. PRE:  Insufficient memory on heap.
 *        POST: Function and program terminates with a non-zero error code
 *              after message is generated--all via lstFatalError.
 *
 * NOTES:
 *     A string created by this function can not necessarily be correctly
 *     split--or split as expected--by lstSplitList.  The delimiter that this
 *     function puts between each list item can be a string, not necessarily a
 *     single character.  But lstSplitList is expecting to be passed a string
 *     whose delimiters are single character.
 *
 * REVISION HISTORY:
 *     09/19/98  J.D. Stroup  Original.
 *     01/08/99  J.D. Stroup  Moved from ListExpander.c & renamed from
 *                            JoinList.  Corrected space allocatation bug.
 *
 ****************************************************************************/

char *lstJoinList (char *delimiter, lstLIST plist)
{
  size_t new_size;
  char   *item;
  char   *delimited_string = strdup ("");
  char   *working_delimiter = (delimiter == NULL) ?
                                             strdup ("") : strdup (delimiter);
  int    num_elems;
  unsigned int    delimiter_len = strlen (working_delimiter);
  int    curr_list_pos = 0; /* current item position in plist, from 1 to
                               lstNumEl */

  /* If the pointer to the given linked list is NULL, return NULL. */

  if (plist == NULL) return (NULL);

  num_elems = lstNumEl (plist);
  lstStopIteration (plist);

  /* Step through each item in the given list. */

  while ((item = (char *) lstIterator (plist)) != NULL)
  {
    curr_list_pos++;

    /* Compute the amount of memory that must be allocated for the output
       string with the current list item appended to it. */

    new_size = strlen (delimited_string) + strlen (item) + 1;

    /* If this isn't the last item in the list, then we must account for the
       size of the delimiter. */

    if (curr_list_pos != num_elems) new_size += delimiter_len;

    /* Re-allocate memory for the output string based on its new size.  Check
       for allocation failure. */

    delimited_string = (char *) realloc ((void *) delimited_string, new_size);
    if (delimited_string == NULL)
      lstFatalError ("lstJoinList",
                        "realloc failure: cannot allocate memory for string");

    /* Append the current list item to the end of the output string. */

    strcat (delimited_string, item);

    /* Again, if this isn't the last item in the list, then append the
       delimiter to the end of the output string. */

    if (curr_list_pos != num_elems)
      strcat (delimited_string, working_delimiter);
  }

  free (working_delimiter);

  /* Return the delimited (output) string. */

  return (delimited_string);
} /* end lstJoinList */

/*****************************************************************************
 * 
 * NAME: lstNumTimesInList
 *
 * DESCRIPTION: Counts the number of times the given name appears in the given
 *              list.
 *
 * PARAMETERS:  
 *     pList -- (lstLIST) Pointer to a list of strings.
 *     pName -- (char *)  Pointer to a character string.
 *
 * RETURN VALUE:
 *     lstNumTimesInList -- (int) The number of times *pName was found in 
 *                        *pList. 
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     <<description>>
 *
 * REVISION HISTORY:
 *   12/02/98  Ken Jarva  --  Original
 *
 ****************************************************************************/
int   lstNumTimesInList( lstLIST pList, char *pName )
{
	int count = 0;
	char *pItem; /* points at a string from the given list */

	assert( pList );
	assert( pName );

	lstStopIteration( pList );
		pItem = lstIterator( pList );
		while( pItem ) {
			if( strToUpStrcmp( pItem, pName ) == 0 ) count++;
			pItem = lstIterator( pList );
		}
	lstStopIteration( pList );
	return count;

} /* end lstNumTimesInList */

/*****************************************************************************
 * 
 * NAME: lstPrintStringList
 *
 * DESCRIPTION: Prints the lstList to the given file, one entry per line,
 *     prefixed with the given indent string.
 *
 * PARAMETERS:  
 *     pF      -- (FILE *) The file to which the data is to be written.
 *     pList   -- (lstLIST) The list to be printed.
 *     pIndent -- (char *) the string that is to preface each line.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES: None
 *	
 * REVISION HISTORY:
 *   12/02/98  Ken Jarva  --  Added header
 *
 ****************************************************************************/
void lstPrintStringList( FILE *pF, lstLIST pList, char *pIndent )
{
	/* pList must be a list of strings. */

	char *pItem;

	assert( pF );
	assert( pList );

	lstStopIteration( pList );
	  pItem = (char *) lstIterator( pList );
	  while( pItem ) {
		fprintf( pF, "%s%s\n", pIndent, pItem );
		pItem = (char *) lstIterator( pList );
	  }
	lstStopIteration( pList );

	return;

} /* end lstPrintStringList */

/*****************************************************************************
 * 
 * NAME: lstIsListOfEmptyStrings
 *
 * DESCRIPTION: This function indicates whether the given list contains all
 *     empty strings or not.
 *
 * PARAMETERS:
 *     pList (lstLIST)  A list of strings.
 *
 * RETURN VALUE:
 *     lstIsListOfEmptyStrings (int)  0: Given list contains at least one
 *                                       non-empty string.
 *                                    1: Given list contains all empty
 *                                       strings.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Given list contains all empty strings.
 *        POST: Function returns 1.
 *     2. PRE:  Given list contains at least one non-empty string.
 *        POST: Function returns 0.
 *
 * NOTES:
 *     It is assumed that the items to which the elements in the given list
 *     point are strings.
 *
 *
 * REVISION HISTORY:
 *     01/17/1999  J.D. Stroup  Original
 *
 ****************************************************************************/

int lstIsListOfEmptyStrings (lstLIST pList)
{
  char *string;
  const char *empty_string = "";

  /* Step through each item in the given list checking for empty strings. */

  lstStopIteration (pList);
  while ((string = lstIterator (pList)) != NULL)
    if (strcmp (string, empty_string) != 0) return (0);

  /* If this point of the function is reached, then the given list contains
     only empty strings. */

  return (1);
}

/*****************************************************************************
 * 
 * NAME: lstDestructor
 *
 * DESCRIPTION: Applies the given destructor function (*pDestructor) to each
 *              of the elements of the given list and then frees the list.
 *              If no destructor function is given (i.e., if pDestructor == 
 *              NULL), the listed items are uneffected; the list and all its
 *              nodes are freed.  This function always returns a NULL pointer.
 *
 * PARAMETERS:  
 *     pList -- (lstLIST) Pointer to the list that is to be freed.
 *     pDestructor -- (function pointer) Points at a function capable of
 *                    destructing elements of this particular list. If
 *                    pDesturctor == NULL, the items of the list will
 *                    not be affected by this call; the list nodes will,
 *                    however, all be freed. 
 *
 * RETURN VALUE: (void *) NULL    //see note 3 below.
 *     
 * GLOBAL VARIABLES: None
 *      
 * PRE- & POST-CONDITIONS:
 *     1. PRE: A NULL pointer to a list is given. POST: The function returns
 *        a NULL pointer. No other action is taken.
 *     2. PRE: A list is given as is a function that is to act as a destructor
 *        for the items of the list. POST: the destructor function has been 
 *        applied to each of the list's items and the list (the root node and
 *        all the list nodes) have been freed.
 *     3. PRE: A list is given but the pointer to the destructor function is
 *        NULL. POST: The list is freed; no action is taken regarding the 
 *        items that were listed.
 *
 * NOTES:
 *     1. Lists and the items being listed are two distinct things.  A list
 *        in this context is a root node and zeor or more nodes that are
 *        each associated with a listed item.  These items may be arbitrarily
 *        complex. LstDestructor knows nothing about the structure of the
 *        items being listed and so must rely on the given destructor 
 *        function (*pDestructor) to free space allocated to a listed item.
 *     2. In many cases it is desirable to free a list without affecting the
 *        listed items.  Since the nodes merely point at the listed items,
 *        items might, for example, occur in more than one list. In such cases
 *        this function can be called giving pDestruction as the NULL pointer.
 *        This is effectively just like calling lstFreeList().
 *     3. This function always returns a (void *) NULL.  This is to accomodate
 *        the following usage:
 *                             pList = lstDestructor( pList, pDF );
 *        whereby the list to which pList points is freed and pList is set
 *        to the NULL pointer in a single statement.
 *
 * REVISION HISTORY:
 *   01/12/99  Ken Jarva  --  Original
 *
 ****************************************************************************/

void *lstDestructor(  lstLIST pList, void *(*pDestructor)( void *)  )
{
        /* Ignore nonexistent lists: */
        if( !pList ) return (void *)NULL;

        /* Perform a deep destruction, if possible, on each element of the 
         * list: */
        if( pDestructor ) 
        {
                /* A destructor function has been supplied...*/
                void *pItem;
                pItem = lstPluckFirst( pList ); /* plucks a node, freeing it as it
                                                 * does, and returns a pointer to the 
                                                 * listed item */
                while( pItem ) {
                        /* Perform deep destruction of the listed item: */
                        pItem = (*pDestructor)( pItem );
                        pItem = lstPluckFirst( pList );
                }
        }

        /* Now free the list regardless of whether a destructor for the
         * list's substructure (if any) was supplied. */
        return lstFreeList( pList ); /*lstFreeList always returns NULL */

} /* end lstDestructor */

