/* FILE: odlSupport.c
 *
 * PROGRAM: Utility module
 *
 * DESCRIPTION: Functions to manipulate ODL nodes.
 *
 * RCS INFO:
 *
 *      $Id: odlSupport.c,v 1.4 1999/06/14 15:37:23 ToolsDev Exp $
 *
 *      $Log: odlSupport.c,v $
 *      Revision 1.4  1999/06/14 15:37:23  ToolsDev
 *      promoted 1.2.0.3 for 2 format bugs.
 *
 * Revision 1.2.0.3  1999/06/14  15:22:52  jarva
 * Corrected format for integer seconds, and added 'Z' for UTC.
 *
 * Revision 1.2.0.2  1999/05/26  16:43:28  jarva
 * Added code to format the seconds part and added a documentation header.
 *
 * Revision 1.2.0.1  1999/04/23  15:24:48  jarva
 * Intermediate version sent to Siri Jodha Singh Khalsa to fix
 * the ODL date problem. Doesn't handle decimal seconds < 10 well yet.
 *
 * Revision 1.2  1999/02/03  16:29:28  stroup
 * Promotion of 1.1.0 branch.
 *
 * Revision 1.1.0.4  1999/02/02  21:14:11  jarva
 * Added odlListParmVals.
 *
 * Revision 1.1.0.3  1999/01/22  15:11:35  stroup
 * The function odlCompMatch calls odlComponentName.
 *
 * Revision 1.1.0.2  1999/01/22  13:19:44  stroup
 * Added odlRootNode, odlParentNode, odlCompMatch, odlIsRootNode,
 * odlOwnerAggregate, and odlValuesParameter.
 *
 * Revision 1.1.0.1  1999/01/20  19:33:03  stroup
 * Added odlDestructor.
 *
 * Revision 1.1  1999/01/06  21:24:20  ToolsDev
 * Initial revision from prototype baseline.
 *
 *
 * NOTICE: this file was constructed from a template that supplied a
 *         control structure and some simple functions for reporting
 *         errors and debug information. See the static struct 
 *         odlCONTROLS below.
 *
 *         This module was originally developed for User Friendly
 *         Metadata (UFM).
*/

/* INCLUDED FILES:
*/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "odlSupport.h"

#define VERBOSE_DEBUG 1
#define IDENTIFY_STUBS 0

/* ---------------------------------------------------------------------*/
/* MACRO DEFINITIONS */

#define VERBOSE_DEBUG 1  /* Set to 1 for maximal debug output but also
                          * see odlCONTROLS.debug.

/* ---------------------------------------------------------------------*/
/* GLOBAL VARIABLES: the following variables may be used in other files.
 */
	/* NONE */

/* ---------------------------------------------------------------------*/
/* EXTERNAL VARIABLES: the following variables are defined in other files.
*/
	/* NONE */

/* ---------------------------------------------------------------------*/
/* STATIC VARIABLES: the following variables are for the exclusive use
 *     of the functions defined within this file.
*/
struct odlControls {
    /* Controls the message writing functions given in the source file
	 * template (e.g., odlDebugMsg().  This structure is
	 * imediately followed by it default settings. Typically the
	 * initailization function odlInitialization is invoked by
	 * the calling application to override these settings. */
	int debug;
	int verbose;
	int noExit;
	FILE *pDebugFile;
	FILE *pUserMessages;
} odlCONTROLS =
{
	1,
	1,
	0,
	NULL,
	NULL
};

/* ---------------------------------------------------------------------*/

/* STATIC PROTOTYPES: the following prototypes are for functions defined
 *     and used exclusively within this file.
*/
static void odlFatalError( char *functionName, char *msg );
static void odlUserWarning( char *functionName, char *msg );
static void odlDebugMsg( char *functionName, char *msg );
static char *odlTimeValue( struct ODLDate *pODLDate );
static char *odlSprintValue( struct Value_Data *pData );
/* ---------------------------------------------------------------------*/
/*****************************************************************************
 * 
 * NAME: odlInitialization
 *
 * DESCRIPTION: Reports fatal errors and shuts down the program.
 *
 * PARAMETERS:  
 *     debugFlag     -- (int) Control to allow debug print to be output. Set
 *                      to 1 for ouput; 0 otherwise.
 *     verboseFlag   -- (int) Control to allow verbose messages for the user
 *                      to be output. Set to 1 for ouput; 0 otherwise.
 *	   noExit        -- (int) Control to prevent the functions within this
 *                      module from exiting the program.  Set to 1 and all
 *                      functions return to the caller, including 
 *                      odlFatalError(...).
 *     pDebugFile    -- (FILE *) Pointer to the debug file to which debug
 *                      messages (if allowed) will be written.  If passed
 *                      as NULL, stderr will be used if necessary.  Only
 *                      NULL or an active FILE pointer should be passed.
 *     pUserMessages -- (FIle *) Pointer to the debug file to which user
 *                      messages (if allowed) will be written.  If passed
 *                      as NULL, stderr will be used if necessary.  Only
 *                      NULL or an active FILE pointer should be passed.
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct odlCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *     noExit     -- (int) control preventing the function from exiting the
 *                   program.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void odlInitialization( int debugFlag, int verboseFlag, int noExit,
	FILE *pDebugFile, FILE *pUserMessages )
{
	odlCONTROLS.debug = debugFlag;
	odlCONTROLS.verbose = verboseFlag;
	odlCONTROLS.noExit = noExit;
	if( pDebugFile ) odlCONTROLS.pDebugFile = pDebugFile;
	if( pUserMessages ) odlCONTROLS.pUserMessages = pUserMessages;
	return;
} /* end odlInitialization */

/*****************************************************************************
 * 
 * NAME: odlFatalError
 *
 * DESCRIPTION: Reports fatal errors and shuts down the program.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct odlCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *     noExit     -- (int) control preventing the function from exiting the
 *                   program.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void odlFatalError( char *functionName, char *msg )
{
	if( odlCONTROLS.debug )
		fprintf( odlCONTROLS.pDebugFile, 
		                              "%s: FATAL - %s\n", functionName, msg );
	if( !odlCONTROLS.noExit ) {
		odlUserWarning( "odlFatalError", 
		     "Fatal programming error is causing the program to shut down." );
		exit(-1);
	}
	else return;

} /* end odlFatalError */

/*****************************************************************************
 * 
 * NAME: odlUserWarning
 *
 * DESCRIPTION: Reports warnings to the user if the VERBOSE flag is set.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct odlCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *	
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void odlUserWarning( char *functionName, char *msg )
{
	if( !odlCONTROLS.verbose ) return;

    fprintf( odlCONTROLS.pUserMessages, 
		                             "%s: WARNING - %s\n", functionName, msg );
	return;

} /* end odlUserWarning */
/*****************************************************************************
 * 
 * NAME: odlDebugMsg
 *
 * DESCRIPTION: Reports debug messages for the maintenance programmer if the
 *              debug flag is set and VERBOSE_DEBUG macro is 1.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct odlCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void odlDebugMsg( char *functionName, char *msg )
{
	if( odlCONTROLS.debug && VERBOSE_DEBUG ){
		fprintf( odlCONTROLS.pDebugFile, 
		                              "%s: INFO - %s\n", functionName, msg );
	    fflush(odlCONTROLS.pDebugFile);
	}
	return;

} /* end odlDebugMsg */
 
/*****************************************************************************
 * 
 * NAME: odlMakeNode
 *
 * DESCRIPTION: Build a component descriptor (ODL_Node struct) for a given ODL
 *		component.
 *
 * PARAMETERS:  
 *     node_kind -- NODE_KIND Enumerated value indicating whether the given
 *                  component (*pODLComp) is an ODL aggregate, parameter, or
 *                  value.
 *
 * RETURN VALUE: (struct ODL_Node *) Pointer to a new component descriptor
 *     that's been malloced on the heap.
 *     
 * GLOBAL VARIABLES: None
 *	
 *
 * NOTES:
 *     1. Asserts that pDODLComp is nonNULL and that it was able to malloc
 *        space on the heap.
 *     2. Fatal Errors: *pODLComp isn't of the kind expected; node_kind isn't
 *				really NODE_KIND.
 *
 * REVISION HISTORY:
 *   09/30/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/
struct ODL_Node *odlMakeNode( NODE_KIND node_kind, void *pODLComp )
{
	struct ODL_Node *pNode;

	assert( pODLComp );

	pNode = malloc( sizeof(*pNode) );

	assert( pNode );
	
	pNode->node_kind = node_kind;

	switch( node_kind )
	{
		case UNDEFINED_NODE:
			pNode->node_type.undefined = (void *) NULL;
			break;
		case AGGREGATE_NODE:
			pNode->node_type.aggregate = pODLComp;
			if( ((AGGREGATE)pODLComp)->kind != KA_OBJECT &&
				((AGGREGATE)pODLComp)->kind != KA_GROUP )
				 odlFatalError( "odlMakeNode", "Component isn't AGGREGATE" );
			break;
		case PARAMETER_NODE:
			pNode->node_type.parameter = pODLComp;
			if( ((PARAMETER)pODLComp)->node_kind != KP_ATTRIBUTE &&
				((PARAMETER)pODLComp)->node_kind != KP_POINTER )
				 odlFatalError( "odlMakeNode", "Component isn't PARAMETER" );
			break;
		case VALUE_NODE:
			pNode->node_type.value = pODLComp;
			if( ((VALUE)pODLComp)->item.type != TV_NULL &&
				((VALUE)pODLComp)->item.type != TV_INTEGER &&
				((VALUE)pODLComp)->item.type != TV_REAL &&
				((VALUE)pODLComp)->item.type != TV_SYMBOL &&
				((VALUE)pODLComp)->item.type != TV_STRING &&
				((VALUE)pODLComp)->item.type != TV_DATE &&
				((VALUE)pODLComp)->item.type != TV_TIME &&
				((VALUE)pODLComp)->item.type != TV_DATE_TIME
			  )
				 odlFatalError( "odlMakeNode", "Component isn't VALUE" );
			break;
		default:
			odlFatalError( "odlMakeNode", "Not NODE_KIND" );
	}

	return pNode;

} /* end odlMakeNode */

/*****************************************************************************
 * 
 * NAME: odlChildComponentsList
 *
 * DESCRIPTION: Returns a pointer to a list of the names of all
 *		the child components (aggregates & parameters ) of a given
 *      ODL node.
 *
 * PARAMETERS:  
 *     pNode -- (struct ODL_Node *) Componenet descriptor of the ODL
 *              component for which the list of children -- that is, all
 *              groups, objects, & parameters -- is sought.
 *
 * RETURN VALUE: lstLIST -- pointer to a lstList of strings each of which
 *              is a child component of *pNode's referenced ODL component.
 *     
 * GLOBAL VARIABLES: None
 *	
 *
 * NOTES:
 *     1. The lstList will be returned empty if pNode is undefined, or if
 *        its referenced ODL component is childless.
 *     2. Asserts that pNode is nonNULL.
 *
 * REVISION HISTORY:
 *   09/27/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

lstLIST odlChildComponentsList( struct ODL_Node *pNode )
{
	ODLNodePtr childNode; /* descriptor used in turn for each of the components
						   * of *pNode's reference ODL component. */
	lstLIST pNameList; /* the list of names that will be returned. */

	assert( pNode );

	/* Set up the child node as an UNDEFINED_NODE so odlNextChildNode will find
	 * the first child of *pNode's referent. */
	childNode.node_kind = UNDEFINED_NODE;
	childNode.node_type.undefined = (void *) NULL;

	pNameList = lstInitList();

	childNode = odlNextChildNode( *pNode, childNode );
	while( childNode.node_kind != UNDEFINED_NODE ) {
		char *pName;
		pName = odlComponentName( &childNode );
		lstAddLast( pNameList, (void *) pName );
		childNode = odlNextChildNode( *pNode, childNode );
	}
	return pNameList;

} /* end odlChildComponentsList */

/*****************************************************************************
 * 
 * NAME: odl_stringValue
 *
 * DESCRIPTION: For parameters with scalar values that are strings: it
 *      returns the value string. If it's not a string, it issues a debug 
 *      message and returns an empty string. 
 *
 * PARAMETERS:  
 *     pParm -- PARAMETER Pointer to the parameter node in the ODL tree.
 *
 * RETURN VALUE: (char *) pointer to the string value in the ODL tree.
 *     
 * GLOBAL VARIABLES: None
 *
 * PRE- & POST-CONDITIONS:
 *     1. PRE: The given PARAMETER pointer is NULL. POST: fatal error
 *        processing is invoked.
 *     2. PRE: *pNode refers to a parameter whose values are not strings.
 *        POST: the function invokes fatal error handling.
 *     3. PRE: *pNode refers to a string-valued ODL parameter. 
 *        POST: A pointer to he first value associated with that parameter
 *        has been returned.
 *
 * NOTES:
 *     1. ***CAREFUL*** The pointer that is returned points at the string
 *        stored for the given parameter. Altering that string could have
 *        unexpected side effects.
 *     2. A pointer to a single string is returned. FirstValue() is used to
 *        get a pointer to the value node. For sequences of strings use
 *        odlGetValueList.
 *
 * REVISION HISTORY:
 *   09/30/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/
char *odl_stringValue( PARAMETER pParm )
{
	VALUE pVal; /* The frist value of the particular parameter. */

	if( pParm ){
		pVal = FirstValue( pParm );
		if(  pVal && 
			( pVal->item.type == TV_STRING ||
			  pVal->item.type == TV_SYMBOL 
			)
		   ) return pVal->item.value.string;
		else{
			char msg[256];
			sprintf( msg, "%s is not string-valued.", pParm->name );
			odlFatalError("odl_stringValue", msg );
		}			
	} else odlFatalError("odl_stringValue", "NULL passed for PARAMETER!" );
	
	
	return strdup("");
} /* end odl_stringValue */

/*****************************************************************************
 * 
 * NAME: odlGetValueList
 *
 * DESCRIPTION: Goes to the given node and if it's a parameter, returns a list
 *		of the values of that node expressed as strings.
 *
 * PARAMETERS:  
 *     pNode -- (struct ODL_Node *) Pointer to an ODL node descriptor for a 
 *              particular parameter. If the node descriptor is not for a
 *              ODL parameter, the function invokes the fatal error processing.
 *              Asserts that the pNode isn't null.
 *
 * RETURN VALUE:
 *     odlGetValueList -- lstLIST Functions returned value: a list of the 
 *			values of the referenced ODL parameter all expressed as character
 *          strings.
 *     
 * GLOBAL VARIABLES: None
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE: The given pointer to pNode is NULL. POST: program terminates
 *        with assertion.
 *     2. PRE: *pNode refers to an ODL component that is not a PARAMETER.
 *        POST: the function invokes fatal error handling.
 *     3. PRE: *pNode refers to an ODL parameter. POST: All the values 
 *        associated with that parameter have been converted to strings (via
 *        sprintf) if necessary and each of the strings have been added to
 *		  a lstList that's been allocated on the heap.
 *
 * REVISION HISTORY:
 *   09/30/98  Ken Jarva             Original
 *
 ****************************************************************************/
lstLIST odlGetValueList( struct ODL_Node *pNode )
{
	lstLIST pValList; /* pointer to tthe list of values to be returned */
	PARAMETER pParm; /* the parameter to which pNode refers */
	VALUE pValue;
	struct Value_Data *pData;
	char *pDataAsString; /* string containg an sprinted data value */

	assert( pNode );

	/* Initialize the list to be returned: */
	pValList = lstInitList();

	if( pNode->node_kind == PARAMETER_NODE ){
		/* Get the pointer to the ODL parameter from the descriptor: */
		pParm = pNode->node_type.parameter;

		/* Loop through all the values of the given parameter: */
		pValue = FirstValue( pParm );
		while( pValue ) {
			pData = &(pValue->item);
			pDataAsString = odlSprintValue( pData );
			lstAddLast( pValList, (void *) pDataAsString );
			pValue = pValue->right_sibling;
		}
	} else {
		lstFreeList( pValList);
		odlFatalError( "odlGetValueList", "Node passed was not a parameter!" );
		return NULL; /* Only in case control returns for odlFatalError */
	}
	return pValList;

} /* end odlGetValueList */
/*****************************************************************************
 * 
 * NAME: odlSprintValue
 *
 * DESCRIPTION: Given a pointer to an ODL Value_Data struct, this function
 *		returns a pointer to a string buffer into which the value has been
 *      sprintf'd.
 *
 * PARAMETERS:  
 *     pData -- (struct Value_Data *) Pointer to the ODL structure for which
 *			    the formatted string is required.
 *
 * RETURN VALUE:
 *     odlSprintValue -- (char *) String containing the formated represention
 *                    of the referenced ODL data.
 *     
 * GLOBAL VARIABLES: None
  *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE: pData = NULL. POST: program terminates with assertion.
 *     2. PRE: pData->type is not one of the handled types (see note 1).
 *        POST: Fatal error processing.
 *
 * NOTES:
 *     1. Only handels values that are of ODL types INTEGER, STRING, SYMBOL, 
 *        REAL, DATE, TIME, & DATE_TIME.
 *
 * REVISION HISTORY:
 *   09/30/98  Ken Jarva -- Original.
 *   10/02/98  Ken Jarva -- Made unknown pData->type a fatal error.
 *   04/15/99  Ken Jarva -- Handling ODL date & time formats.
 *
 ****************************************************************************/

char *odlSprintValue( struct Value_Data *pData )
{
	char msg[1024] = "  ";
	struct ODLInteger *pODLInt;
	struct ODLReal *pODLReal;
	struct ODLDate *pODLDate;

	assert( pData );

	if( pData->type == TV_INTEGER ) {
		pODLInt = &(pData->value.integer);
		sprintf( msg, "%d", pODLInt->number );
	}else if( pData->type == TV_STRING ) {
		sprintf( msg, "%s", pData->value.string );
	}else if( pData->type == TV_SYMBOL ) {
		sprintf( msg, "%s", pData->value.string );
	}else if( pData->type == TV_REAL ) {
		pODLReal = &(pData->value.real);
		sprintf( msg, "%f",pODLReal->number );
	}else if( pData->type == TV_DATE ) {
		pODLDate = &(pData->value.date_time );
		sprintf( msg, "%02d-%02d-%02d", pODLDate->year, pODLDate->month,
			pODLDate->day );
	}else if( pData->type == TV_TIME ) {
		char *pTime; /* Will point to string representing time */
		pODLDate = &(pData->value.date_time );
		pTime = odlTimeValue( pODLDate );
		sprintf( msg, "%s", pTime );
		free( pTime );		
	}else if( pData->type == TV_DATE_TIME ) {
		char *pTime; /* Will point to string representing time */
                pODLDate = &(pData->value.date_time );
                pTime = odlTimeValue( pODLDate );
                sprintf( msg, "%02d-%02d-%02dT%s", pODLDate->year, pODLDate->month,
			pODLDate->day, pTime );
                free( pTime );
	}else {
		char errmsg[256];
		sprintf( errmsg, "Can't handle ODL data type: %d.",
			(int) pData->type );
		odlFatalError( "odlSprintValue", errmsg );
	}

	return strdup( msg );

} /* end odlSprintValue */

/*****************************************************************************
 * 
 * NAME: odlTimeValue
 *
 * DESCRIPTION: Converts an ODL date/time struct to a string representation.
 *
 * PARAMETERS:  
 *     pODLDate -- (struct ODLDate *) pointer to an ODL date/time struct.
 *
 * RETURN VALUE:
 *     odlTimeValue -- (char *) pointer to a string buffer containing a string
 *			representation of the given date and time.
 *     
 * NOTES:
 *     1. Had to put in extra code to format times having fractional seconds.
 *        I wasn't getting a leading 0 with '%02.9g' when there were less
 *        than 10 seconds.
 *
 * REVISION HISTORY:
 *   05 May 1999  Ken Jarva   original
 *   14 Jun 1999  Ken Jarva   corrections to formats.
 *
 ****************************************************************************/
char *odlTimeValue( struct ODLDate *pODLDate )
{
	char strTime[50];
	char strSeconds[14];
	double rSeconds = 0.000;
	assert( pODLDate );

	/* Format the seconds part of the expression: */
	if( pODLDate->nanoseconds == 0 ) {
		sprintf( strSeconds, "%02d", pODLDate->seconds );
	}else {
		rSeconds = pODLDate->seconds +
		          ( (double) pODLDate->nanoseconds / 1000000000 );
		if( rSeconds < 10.0 ) 
			sprintf( strSeconds, "0%1.9g", rSeconds );
		else
			sprintf( strSeconds, "%2.9g", rSeconds );
	}

	/* Times ***must*** be expressed either as UTC (Z) or with a zone
	 * offset...*/
	if( pODLDate->zone_hours == 0 && pODLDate->zone_minutes == 0 ) {
		/* We have UTC */
		sprintf( strTime, "%02d:%02d:%sZ", pODLDate->hours,
			pODLDate->minutes, strSeconds );
	} else
		sprintf( strTime, "%02d:%02d:%s+%02d:%02d",
			pODLDate->hours, pODLDate->minutes, strSeconds,
			pODLDate->zone_hours, pODLDate->zone_minutes );

	return strdup( strTime );
}

/*****************************************************************************
 * 
 * NAME: odlAreEqualNodes
 *
 * DESCRIPTION: Determines whether two components are equal
 *
 * PARAMETERS:  
 *     pNode1 -- (struct ODL_Node*) Pointer to a component descriptor.
 *     pNode2 -- (struct ODL_Node*) Pointer to a component descriptor.
 *
 * RETURN VALUE:
 *     odlAreEqualNodes -- int 1 -> the components are the same; 0 ->
 *                             they are not.
 *     
 * GLOBAL VARIABLES: None
 *	
 * REVISION HISTORY:
 *   10/07/98  Ken Jarva  --  Original
 *
 ****************************************************************************/
int odlAreEqualNodes( struct ODL_Node *pNode1, struct ODL_Node *pNode2 )
{
	if( pNode1->node_kind != pNode2->node_kind ) return 0;
	if( (void *) pNode1->node_type.aggregate != 
		(void *) pNode2->node_type.aggregate ) return 0;
	return 1;
} /* end odlAreEqualNodes */

/*****************************************************************************
 * 
 * NAME: odlIsObject, odlIsGroup, odlIsAttribute, odlIsPointer
 *
 * DESCRIPTION: Checks whether the given component is of the kind
 *              indicated by the function name.
 *
 * PARAMETERS: 
 *      pNode -- (struct ODL_Node *) points at component descriptor.
 *
 * RETURN VALUE:
 *     odlIs<component kind> -- (int) Indicator of whether the given component
 *                                    is of the kind named in the function's
 *                                    name. 1 --> yes, 0 --> no.
 *     
 * GLOBAL VARIABLES: None
 *
 * REVISION HISTORY:
 *   10/09/98  Ken Jarva  --  Original
 *
 ****************************************************************************/

int odlIsObject( struct ODL_Node * pNode )
{
	assert( pNode );

	if( pNode->node_kind == AGGREGATE_NODE ) {
		AGGREGATE pAgg;
		pAgg = pNode->node_type.aggregate;
		if( pAgg->kind == KA_OBJECT ) return 1;
	}
	return 0;

} /* end odlIsObject */

int odlIsGroup( struct ODL_Node * pNode )
{
	assert( pNode );

	if( pNode->node_kind == AGGREGATE_NODE ) {
		AGGREGATE pAgg;
		pAgg = pNode->node_type.aggregate;
		if( pAgg->kind == KA_GROUP ) return 1;
	}
	return 0;

} /* end odlIsGroup */

int odlIsParameter( struct ODL_Node * pNode )
{
	assert( pNode );

	if( pNode->node_kind == PARAMETER_NODE ) return 1;

	return 0;

} /* end odlIsParameter */

int odlIsAttribute( struct ODL_Node * pNode )
{
	assert( pNode );

	if( pNode->node_kind == PARAMETER_NODE ) {
		PARAMETER pAgg;
		pAgg = pNode->node_type.parameter;
		if( pAgg->node_kind == KP_ATTRIBUTE ) return 1;
	}
	return 0;

} /* end odlIsAttribute */

int odlIsPointer( struct ODL_Node * pNode )
{
	assert( pNode );

	if( pNode->node_kind == PARAMETER_NODE ) {
		PARAMETER pAgg;
		pAgg = pNode->node_type.parameter;
		if( pAgg->node_kind == KP_POINTER ) return 1;
	}
	return 0;

} /* end odlIsPointer */

/*****************************************************************************
 * 
 * NAME: odlRootNode
 *
 * DESCRIPTION: This function returns a structure that points to the root node
 *     of the ODL tree.
 *
 * PARAMETERS:
 *     base_node (ODLNodePtr)  A structure pointing to a node in the ODL tree.
 *
 * RETURN VALUE:
 *     odlRootNode (ODLNodePtr)  A structure pointing to the root node of the
 *                               ODL tree.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Base_node is pointing to a node in the ODL tree.
 *        POST: Function returns a structure pointing to the root node.
 *     2. PRE:  Base_node is an undefined/null structure.
 *        POST: Function issues an error message and terminates the program
 *              (via a call to odlFatalError).
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/18/1998  J.D. Stroup  Original
 *     01/21/1999  J.D. Stroup  Moved RootNode from module ListExpander.c and
 *                              renamed.  Updated the calls to some functions
 *                              (odl...) due to the functions being renamed.
 *
 ****************************************************************************/

ODLNodePtr odlRootNode (ODLNodePtr base_node)
{
  ODLNodePtr parent_node;  /* the structure to return */
  int        found_root = 0;

  parent_node = base_node;

  /* Start searching for the root node. */

  while (! found_root)
  {
    switch (parent_node.node_kind)
    {
      /* If we're at an aggregate node, then we need to determine whether
         it's the root node or not.  If it is, then terminate the search--we
         found what we were looking for. */

      case AGGREGATE_NODE:

        if (odlIsRootNode (parent_node.node_type.aggregate))
          found_root = 1;
        else
          parent_node = odlParentNode (parent_node);

        break;

      /* If we're at a parameter or value node, then we need to get the
         parent/owner node and continue the search. */

      case PARAMETER_NODE:
      case VALUE_NODE:
        parent_node = odlParentNode (parent_node);
        break;

      /* We should never reach this point, thus it's considered a fatal error
         if we do. */

      default:
      {
        char *format = "invalid node kind, '%d', encountered";
        char *errmsg = (char *) malloc ((strlen (format) + 20) *
                                        sizeof (char));
          /* 20 is a fudge factor that takes into account the number of digits
             in the number field - %d */

        if (errmsg == NULL)
          odlFatalError ("odlRootNode",
           "malloc failure: cannot allocate memory for error message string");

        sprintf (errmsg, format, parent_node.node_kind);
        odlFatalError ("odlRootNode", errmsg);
        break;
      }
    }
  }

  /* Return the structure pointing to the root node. */

  return (parent_node);

} /* end odlRootNode */

/*****************************************************************************
 * 
 * NAME: odlParentNode
 *
 * DESCRIPTION: This function returns a structure that points to the parent
 *     node of the given node in an ODL tree.  If the given node is an
 *     aggregate node, then a structure pointing to its parent (another
 *     aggregate node) is returned.  If the given node is a parameter node,
 *     then a structure pointing to its owner (an aggregate node) is returned.
 *     If the given node is a value node, then a structure pointing to its
 *     associated parameter node is returned.  If the given node is the root
 *     node, then it has no parent, and a undefined/null pointer structure is
 *     returned (i.e., the structure's "node_kind" field is set to UNDEFINED,
 *     and the "node_type" pointer--which, in this case, is the "undefined"
 *     pointer--is set to NULL).
 *
 * PARAMETERS:
 *     curr_node (ODLNodePtr)  A structure pointing to a node in the ODL tree.
 *
 * RETURN VALUE:
 *     odlParentNode (ODLNodePtr)  A structure pointing to the parent or owner
 *                                 of curr_node.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Curr_node is pointing to the root node.
 *        POST: Function returns an undefined/null pointer structure.
 *     2. PRE:  Curr_node is pointing to an aggregate node that is not the
 *              root.
 *        POST: Function returns a structure pointing to curr_node's parent
 *              (another aggregate node).
 *     3. PRE:  Curr_node is pointing to a parameter node.
 *        POST: Function returns a structure pointing to curr_node's owner (an
 *              aggregate node).
 *     4. PRE:  Curr_node is pointing to a value node.
 *        POST: Function returns a structure pointing to curr_node's parameter
 *              node.
 *     5. PRE:  Curr_node is none of the above.
 *        POST: Function issues an error message and terminates the program
 *              (via a call to odlFatalError).
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/17/1998  J.D. Stroup  Original
 *     01/21/1999  J.D. Stroup  Moved ParentNode from module ListExpander.c
 *                              and renamed.  Updated the calls to some
 *                              functions (odl...) due to the functions being
 *                              renamed.
 *
 ****************************************************************************/

ODLNodePtr odlParentNode (ODLNodePtr curr_node)
{
  ODLNodePtr parent_node;  /* the structure to return */

  /* Determine the "kind" of the ODL node pointed to by the given structure
     (curr_node). */

  switch (curr_node.node_kind)
  {
    case AGGREGATE_NODE:

      /* Determine whether we're sitting at root or not.  If so, return an
         undefined/null pointer structure.  Otherwise return a structure
         pointing to the parent of curr_node. */

      if (odlIsRootNode (curr_node.node_type.aggregate))
      {
        parent_node.node_kind = UNDEFINED_NODE;
        parent_node.node_type.undefined = NULL;
      }
      else
      {
        parent_node.node_kind = AGGREGATE_NODE;
        parent_node.node_type.aggregate =
                              ParentAggregate (curr_node.node_type.aggregate);
      }
      break;

    case PARAMETER_NODE:
      parent_node.node_kind = AGGREGATE_NODE;
      parent_node.node_type.aggregate =
                            odlOwnerAggregate (curr_node.node_type.parameter);
      break;

    case VALUE_NODE:
      parent_node.node_kind = PARAMETER_NODE;
      parent_node.node_type.parameter =
                               odlValuesParameter (curr_node.node_type.value);
      break;

    default:
    {
      /* We should never reach this point, thus it's considered a fatal error
         if we do. */
      
      char *format = "invalid node kind, '%d', passed to odlParentNode";
      char *errmsg = (char *) malloc ((strlen (format) + 20) * sizeof (char));
        /* 20 is a fudge factor that takes into account the number of digits
           in the number field - %d. */

      if (errmsg == NULL)
        odlFatalError ("odlParentNode",
           "malloc failure: cannot allocate memory for error message string");

      sprintf (errmsg, format, curr_node.node_kind);
      odlFatalError ("odlParentNode", errmsg);
      break;
    }
  }

  return (parent_node);

} /* end odlParentNode */

/*****************************************************************************
 * 
 * NAME: odlNextChildNode
 *
 * DESCRIPTION: This function returns a structure pointing to the next child
 *     node of the given parent node.  The next child node is determined by
 *     several factors: (1) the parent node's kind (aggregate, parameter, or
 *     value), (2) the child node's kind, and (3) the pre-order traversal of
 *     the ODL tree.  If the parent node is an aggregate node, then all of its
 *     aggregate child nodes are returned first followed by its parameter
 *     nodes.  If the parent node is a parameter node, then all of its value
 *     nodes are returned.  Finally, an "undefined" structure is returned if
 *     any of the following conditions are met:
 *
 *     1) the parent node is an aggregate node and all of its parameter nodes
 *        have been returned
 *     2) the parent node is a parameter node and all of its value nodes have
 *        been returned
 *     3) the parent node is a value node
 *
 *     An "undefined" structure is defined as an ODLNodePtr structure whose
 *     indicator of which kind of ODL node the structure is pointing to is set
 *     to UNDEFINED.  See NOTES, below, for an example.
 *
 * PARAMETERS:
 *     parent_node (ODLNodePtr)  A structure pointing to a node in the ODL
 *                               tree.  This function returns the next child
 *                               (aggregate, parameter, or value node) of
 *                               parent_node, if there is one.  If this node
 *                               has no children, an undefined structure is
 *                               returned.
 *     child_node (ODLNodePtr)   A structure pointing to a node in the ODL
 *                               tree.  This node is the child of the node
 *                               pointed to by parent_node.  If child_node is
 *                               an undefined structure, then this function
 *                               returns the first child of parent_node, if
 *                               there is one; or an undefined structure
 *                               otherwise.  If child_node is not undefined,
 *                               but is pointing to a child of parent_node,
 *                               then this function returns the next sibling
 *                               of child_node, if there is one; or an
 *                               undefined structure otherwise.
 *
 * RETURN VALUE:
 *     odlNextChildNode (ODLNodePtr)  A structure pointing to a node in the
 *                                    ODL tree or an undefined structure.  If
 *                                    it's not undefined, then it is pointing
 *                                    to a child of parent_node.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Parent_node is an aggregate; child_node is undefined.
 *        POST: Function returns one of the following, depending of their
 *              existence:
 *              1) the first aggregate child of parent_node; or
 *              2) the first parameter node of parent_node; or
 *              3) an undefined structure.
 *     2. PRE:  Parent_node is an aggregate; child_node is an aggregate.
 *        POST: Function returns one of the following, depending of their
 *              existence:
 *              1) the next aggregate child of parent_node (i.e., the next
 *                 sibling of child_node); or
 *              2) the first parameter node of parent_node; or
 *              3) an undefined structure.
 *     3. PRE:  Parent_node is an aggregate; child_node is a parameter.
 *        POST: Function returns one of the following, depending of their
 *              existence:
 *              1) the next parameter child of parent_node (i.e., the next
 *                 sibling of child_node); or
 *              2) an undefined structure.
 *     4. PRE:  Parent_node is an aggregate; child_node is none of above.
 *        POST: Function issues an error message and terminates the program
 *              (via a call to odlFatalError).
 *     5. PRE:  Parent_node is a parameter; child_node is undefined.
 *        POST: Function returns one of the following, depending of their
 *              existence:
 *              1) the first value child of parent_node; or
 *              2) an undefined structure.
 *     6. PRE:  Parent_node is a parameter; child_node is a value.
 *        POST: Function returns one of the following, depending of their
 *              existence:
 *              1) the next value child of parent_node (i.e., the next
 *                 sibling of child_node); or
 *              2) an undefined structure.
 *     7. PRE:  Parent_node is a parameter; child_node is neither undefined
 *              nor a value.
 *        POST: Function issues an error message and terminates the program
 *              (via a call to odlFatalError).
 *     8. PRE:  Parent_node is a value.
 *        POST: Function returns an undefined structure.
 *     9. PRE:  Parent_node is undefined (or some "kind" not listed above).
 *        POST: Function issues an error message and terminates the program
 *              (via a call to odlFatalError).
 *
 * NOTES:
 *     For example assume a program is using this function in a loop to visit
 *     each child of some aggregate node in an ODL tree.  Assume that this
 *     particular aggregate node has four children: two aggregate children and
 *     two parameter children.  Parent_node is a structure pointing to this
 *     aggregate node.  Child_node at this point is undefined.  Both
 *     parent_node and child_node are passed to odlNextChildNode, which
 *     returns a structure pointing to the first aggregate child node of
 *     parent_node.  This returned structure is then passed to the second call
 *     to odlNextChildNode as the child_node.  The second call returns a
 *     structure pointing to the second aggregate child node of parent_node,
 *     i.e., the aggregate sibling node of the first aggregate child node.
 *     During the third call to this function, the first parameter node is
 *     returned since there are no more aggregate child nodes.  The fourth
 *     call returns the second, and last, parameter node.  Since there are no
 *     more children, the fifth call returns an undefined structure.  Note
 *     that a sixth call would start the process all over again and return the
 *     first aggregate child node.
 *
 * REVISION HISTORY:
 *     09/18/1998  J.D. Stroup  Original
 *
 ****************************************************************************/

ODLNodePtr odlNextChildNode (ODLNodePtr parent_node, ODLNodePtr child_node)
{
  ODLNodePtr output_node;
  AGGREGATE  agg;
  PARAMETER  parm;
  VALUE      val;
  char       *errmsg;

  /* The local variable, output_node, is what is eventually returned by this
     function.  The structure is initialized to undefined. */

  output_node.node_kind = UNDEFINED_NODE;
  output_node.node_type.undefined = NULL;

  /* Primary control is based on the parent node's kind.  By knowing the
     parent node's kind, we know the possibilities for the child node's
     kinds. */

  switch (parent_node.node_kind)
  {
    /* Parent node is an aggregate. */

    case AGGREGATE_NODE:

      /* Secondary control is based on the child node's kind. */

      switch (child_node.node_kind)
      {
        /* Child node is undefined.  In this case we return the first child of
           the parent, which can be either an aggregate or parameter node.  If
           the parent has no children, the undefined structure is returned. */

        case UNDEFINED_NODE:
          agg = parent_node.node_type.aggregate->first_child;
          if (agg == NULL)
          {
            parm = FirstParameter (parent_node.node_type.aggregate);
            if (parm != NULL)
            {
              output_node.node_kind = PARAMETER_NODE;
              output_node.node_type.parameter = parm;
            }
          }
          else
          {
            output_node.node_kind = AGGREGATE_NODE;
            output_node.node_type.aggregate = agg;
          }
          break;

        /* Child node is an aggregate.  In this case we return the next
           aggregate node (i.e., the child's next sibling), if there is one.
           Otherwise we return the first parameter node of the parent, if
           there is one.  Finally, if the child node has no siblings
           (aggregate or parameter), we return an undefined structure. */

        case AGGREGATE_NODE:
          agg = child_node.node_type.aggregate->right_sibling;
          if (agg == NULL)
          {
            parm = FirstParameter (parent_node.node_type.aggregate);
            if (parm != NULL)
            {
              output_node.node_kind = PARAMETER_NODE;
              output_node.node_type.parameter = parm;
            }
          }
          else
          {
            output_node.node_kind = AGGREGATE_NODE;
            output_node.node_type.aggregate = agg;
          }
          break;

        /* Child node is a parameter.  In this case we return the next
           parameter node, if there is one.  Otherwise, we return an undefined
           structure. */

        case PARAMETER_NODE:
          parm = NextParameter (child_node.node_type.parameter);
          if (parm != NULL)
          {
            output_node.node_kind = PARAMETER_NODE;
            output_node.node_type.parameter = parm;
          }
          break;

        /* Child node is a value (or something else not listed above).  This
           should not occur, thus it's considered a fatal error.  An error
           message is generated and the program is termintated. */

        default:
        {
          char *format = "invalid node kind, '%d', of a child whose parent "
                         "is an aggregate";

          errmsg = (char *) malloc ((strlen (format) + 20) * sizeof (char));
            /* 20 is a fudge factor that takes into account the number of
               digits in the number field - %d. */

          if (errmsg == NULL)
            odlFatalError ("odlNextChildNode", "malloc failure: cannot "
                                  "allocate memory for error message string");

          sprintf (errmsg, format, child_node.node_kind);
          odlFatalError ("odlNextChildNode", errmsg);
          break;
        }
      }
      break; /* case of parent node is an aggregate */

    /* Parent node is a parameter. */

    case PARAMETER_NODE:

      /* Secondary control is based on the child node's kind. */

      /* Child node is undefined.  In this case we return the first child of
         the parent, which can only be a value node.  If the parent has not
         children, the undefined structure is returned. */

      if (child_node.node_kind == UNDEFINED_NODE)
      {
        val = FirstValue (parent_node.node_type.parameter);
        if (val != NULL)
        {
          output_node.node_kind = VALUE_NODE;
          output_node.node_type.value = val;
        }
      }

      /* Child node is a value.  In this case we return the next value node,
         if there is one.  Otherwise, we return an undefined structure.*/

      else if (child_node.node_kind == VALUE_NODE)
      {
        val = NextValue (child_node.node_type.value);
        if (val != NULL)
        {
          output_node.node_kind = VALUE_NODE;
          output_node.node_type.value = val;
        }
      }

      /* Child node is none of the above.  This should not occur, thus it's
         considered a fatal error.  An error message is generated and the
         program is termintated. */

      else
      {
        char *format = "invalid node kind, '%d', of child whose parent is a "
                       "parameter";

        errmsg = (char *) malloc ((strlen (format) + 20) * sizeof (char));
          /* 20 is a fudge factor that takes into account the number of digits
             in the number field - %d. */

        if (errmsg == NULL)
          odlFatalError ("odlNextChildNode",
           "malloc failure: cannot allocate memory for error message string");

        sprintf (errmsg, format, child_node.node_kind);
        odlFatalError ("odlNextChildNode", errmsg);
      }
      break;

    /* Parent node is a value.  Therefore it has no children and we do nothing
       (i.e., we return an undefined structure). */

    case VALUE_NODE:
      break;

    /* The parent node is undefined.  This should not occur, thus it's
       considered a fatal error.  An error message is generated and the
       program is terminated. */

    default:
    {
      char *format = "invalid node kind, '%d', of parent node passed to "
                     "odlNextChildNode";

      errmsg = (char *) malloc ((strlen (format) + 20) * sizeof (char));
        /* 20 is a fudge factor that takes into account the number of digits
           in the number field - %d. */

      if (errmsg == NULL)
        odlFatalError ("odlNextChildNode",
           "malloc failure: cannot allocate memory for error message string");

      sprintf (errmsg, format, parent_node.node_kind);
      odlFatalError ("odlNextChildNode", errmsg);
      break;
    }
  }

  return (output_node);

} /* end odlNextChildNode */

/*****************************************************************************
 * 
 * NAME: odlCompMatch
 *
 * DESCRIPTION: This function returns 1 (TRUE) if the name of the node in the
 *     ODL tree, pointed to by the node structure (ODLNodePtr), matches the
 *     name of the current element, or component.  It returns 0 (FALSE)
 *     otherwise.
 *
 * PARAMETERS:
 *     node (ODLNodePtr)   A structure pointing to a node in the ODL tree.
 *     comp_name (char *)  The name of a component.
 *
 * RETURN VALUE:
 *     odlCompMatch (int)  0 --> the name of the node in the ODL tree does NOT
 *                               match the given component name (comp_name).
 *                         1 --> the name of the node in the ODL tree does 
 *                               match the given component name (comp_name).
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Node is pointing to an aggregate or parameter node in the ODL
 *              tree whose name is the same as the given string (comp_name).
 *        POST: Function returns 1 (true).
 *     2. PRE:  Node is pointing to an aggregate or parameter node in the ODL
 *              tree whose name is NOT the same as the given string
 *              (comp_name).
 *        POST: Function returns 0 (false).
 *     3. PRE:  Node is an undefined/null structure or pointing to a value
 *              node in the ODL tree.
 *        POST: Function returns 0 (false).
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/18/1998  J.D. Stroup  Original
 *     01/21/1999  J.D. Stroup  Moved CompMatch from module ListExpander.c and
 *                              renamed.
 *     01/22/1999  J.D. Stroup  Replaced code that got the given node's name
 *                              with a call to odlComponentName.
 *
 ****************************************************************************/

int odlCompMatch (ODLNodePtr node, char *comp_name)
{
  char *node_name = odlComponentName (&node); /* get the node's name */

  /* If the node is a value node or undefined, then node_name is NULL.
     Therefore return 0. */

  if (node_name == NULL)
    return (0);

  /* Compare the node's name with the given component's name.  If they're the
     same return 1; otherwise, return 0. */

  if (strcmp (node_name, comp_name) == 0)
    return (1);
  else
    return (0);

} /* end odlCompMatch */

/*****************************************************************************
 * 
 * NAME: odlComponentName
 *
 * DESCRIPTION: Given an ODLNodePtr, this function returns a pointer to the
 *		(string) name of the node in the ODL tree.
 *
 * PARAMETERS:  
 *     pODLNode -- (struct ODL_Node) Points at the Node pointer structure for
 *				                     which the name is sought.
 *
 * RETURN VALUE: char pointer to the string name in the ODL tree.
 *     
 * REVISION HISTORY:
 *   09/25/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/
char *odlComponentName(  struct ODL_Node *pODLNode )
{
	if(  pODLNode->node_kind == AGGREGATE_NODE  &&
		           pODLNode->node_type.aggregate != (AGGREGATE) NULL ) 
		return pODLNode->node_type.aggregate->name;
	else if( pODLNode->node_kind == PARAMETER_NODE &&
                   pODLNode->node_type.parameter != (PARAMETER) NULL ) 
		return pODLNode->node_type.parameter ->name;
	else return (char *) NULL;

} /* end odlComponentName */

/*****************************************************************************
 * 
 * NAME: odlIsRootNode
 *
 * DESCRIPTION: Determines whether the given aggregate node is the root node.
 *
 * PARAMETERS:  
 *     base_node (AGGREGATE)  Pointer to an aggregate node.
 *
 * RETURN VALUE:
 *     odlIsRootNode (int)  0 --> given node is NOT the root node.
 *                          1 --> given node is the root node.
 *
 * GLOBAL VARIABLES:
 *	   None.
 *
 * NOTES:
 *     If the pointer to the given node is NULL, then 0 is returned.  The
 *     means for determining whether the given node is root or not is very
 *     simple: if it's parent pointer is NULL, then it's root; otherwise, it's
 *     not root.
 *
 * REVISION HISTORY:
 *     09/17/1998  J.D. Stroup  Original.
 *     01/21/1999  J.D. Stroup  Moved IsRootNode from module ListExpander.c
 *                              and renamed.
 *
 ****************************************************************************/

int odlIsRootNode (AGGREGATE base_node)
{
  if (base_node == NULL) return (0);

  if (base_node->parent == NULL)
    return (1);
  else
    return (0);
} /* end odlIsRootNode */

/*****************************************************************************
 * 
 * NAME: odlOwnerAggregate
 *
 * DESCRIPTION: Returns a pointer to a parameter node's owner.
 *
 * PARAMETERS:  
 *     parm_node (PARAMETER)  Pointer to a parameter node.
 *
 * RETURN VALUE:
 *     odlOwnerAggregate (AGGREGATE)  Pointer to the owner node.  If the value
 *                                    node is NULL, then NULL is returned.
 *     
 * GLOBAL VARIABLES:
 *	   None.
 *
 * REVISION HISTORY:
 *     09/17/1998  J.D. Stroup  Original.
 *     01/21/1999  J.D. Stroup  Moved OwnerAggregate from module
 *                              ListExpander.c and renamed.
 *
 ****************************************************************************/

AGGREGATE odlOwnerAggregate (PARAMETER parm_node)
{
  if (parm_node == NULL)
    return (NULL);
  else
    return (parm_node->owner);
} /* end odlOwnerAggregate */

/*****************************************************************************
 * 
 * NAME: odlValuesParameter
 *
 * DESCRIPTION: Returns a pointer to a value node's parameter.
 *
 * PARAMETERS:  
 *     value_node (VALUE)  Pointer to a value node.
 *
 * RETURN VALUE:
 *     odlValuesParameter (PARAMETER)  Pointer to the value node's parameter
 *                                     node.  If the value node is NULL, then
 *                                     NULL is returned.
 *     
 * GLOBAL VARIABLES:
 *	   None.
 *
 * REVISION HISTORY:
 *     09/17/1998  J.D. Stroup  Original.
 *     01/21/1999  J.D. Stroup  Moved ValuesParameter from module
 *                              ListExpander.c and renamed.
 *
 ****************************************************************************/

PARAMETER odlValuesParameter (VALUE value_node)
{
  if (value_node == NULL)
    return (NULL);
  else
    return (value_node->parameter);
} /* end odlValuesParameter */

/*****************************************************************************
 * 
 * NAME: odlDestructor
 *
 * DESCRIPTION: Frees the given ODL node pointer structure (ODLNodePtr,
 *     defined in ODL_Node.h).
 *
 * PARAMETERS:  
 *     pOdl -- (void *) Pointer to an ODL node pointer structure.  See note 1
 *                      below.
 *
 * RETURN VALUE: (void *) NULL
 *     
 * GLOBAL VARIABLES: None
 *      
 * NOTES:
 *     1. The function's formal argument is a void pointer rather than
 *        a character pointer because this function is frequently used
 *        as an argument to lstDestructor, which must apply odlDestructor
 *        without having knowledge of the data type of odlDestructor's
 *        argument.
 *     2. This function always returns a (void *) NULL.  This is to accomodate
 *        the following usage:
 *                             pOdl = odlDestructor( pOdl );
 *        whereby the string to which pOdl points is freed and pOdl is set
 *        to the NULL pointer in a single statement.
 *
 * REVISION HISTORY:
 *   01/18/99  J.D. Stroup  Original: cloned from strDestructor.
 *
 ****************************************************************************/

void *odlDestructor(void *pOdl)
{
        if( pOdl )free( pOdl );
        return (void *)NULL;
} /* end odlDestructor */

/*****************************************************************************
 * 
 * NAME: odlListParmVals
 *
 * DESCRIPTION: Given a lstList of ODL parameters, this function constructs 
 *              a lstList of lstLists of strings.  Each element of the
 *              return list is a list giving all the values of the corresponding 
 *              parameter. Each value is represented as a string such as would
 *              be produced by sprintf.
 *
 * PARAMETERS:  <<Parameter names should appear in order of argument list.>>
 *     <<parameter name>> -- <<TYPE>> <<Description>>
 *
 * RETURN VALUE:
 *     odlListParmVals -- (lstLIST) A pointer to a list of strings. Each
 *				string in the returned list represents the values (via
 *              sprintf) of the corresponding parameter in the passed list.
 *              If for any reason a value string cannont be determined, the
 *              string "?????" is put in its place.
 *     
 * GLOBAL VARIABLES: None
 *	
 * PRE- & POST-CONDITIONS:
 *     <<description of conditions>>
 *
 * NOTES:
 *     1.  The return value is a lstList of lstLists so that we don't lose
 *         information about which values are associated with which parameter.
 *
 * REVISION HISTORY:
 *   26 January  Ken Jarva   Original
 *
 ****************************************************************************/

lstLIST odlListParmVals( lstLIST pPNodeList )
{
	lstLIST
		pStrValList = NULL,  /* The list of values */
		pParmLists = NULL;   /* List of lists of parmameter values */
	struct ODL_Node *pNode = NULL; /* Parameter Node */

	assert( pPNodeList ); /* Ensure it's not NULL */

	/* Initialize the return list: */
	pParmLists = lstInitList();

	/* For each parameter referenced by the input list, build a list of
	 * values and then add that list to the list to be returned: */
	lstStopIteration( pPNodeList );
	  pNode = lstIterator( pPNodeList );

	  while( pNode ) {

		/* Deal with things that ARE NOT parameters: */
		if( !odlIsParameter(pNode) ) {
			/* pNode doesn't refer to a parameter. Put a marker to indicate
			* that there was a problem: */
			pStrValList = lstInitList();
			lstAddLast( pStrValList, (void *) strdup("?????") );
			lstFreeList( pStrValList );

        /* Or deal with things that are: */
		} else {
			/* Build a list of the parameter's values and put the list
			 * into the return list */
			pStrValList = odlGetValueList( pNode );
			lstAddLast( pParmLists, (void *) pStrValList );
			pStrValList = NULL;
		}

		pNode = lstIterator( pPNodeList );

	  } /* End of loop on pNode */
		  
	return pParmLists;

} /* end odlListParmVals */

