/* FILE: ufm.c
 *
 * PROGRAM: User Friendly Metadata (UFM)
 *
 * DESCRIPTION: This file contains the main function, and some high-level
 *     supporting functions, for the UFM tool.
 *
 * RCS INFO:
 *
 *
 *      Revision 1.3  2000/04/06 23:38:02  ToolsDev
 *      Made changes to display the ODL file header informations:
 *      created new case 'h' to invok the header option in the command
 *      line.
 *      created new variable for the header in ufmControls.
 *      wrote comments to reflect the changes.
 *      (Amer Rahabi)
 *
 * Revision 1.2  1999/04/06  15:51:38  ToolsDev
 * Promotion of branch 1.1.0 (cleared wlint warnings) for release.
 *
 * Revision 1.1.0.2  1999/03/05  18:01:57  ToolsDev
 * *** empty log message ***
 *
 * Revision 1.1.0.1  1999/03/05  16:48:08  ToolsDev
 * Cleared several points that generted lint warnings.
 *
 * Revision 1.1  1999/01/06  20:34:20  ToolsDev
 * Initial revision from prototype baseline.
 *
 *
*/
/* INCLUDED FILES:
*/
#include "ufmMain.h"
#include "macMacro.h"
#include "sltLUT.h"
#include "odldef.h"
#include "odlinter.h"
#include "cmpIFCompiler.h"
#include "ReportWriter.h"
#include "synODL.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
/* ---------------------------------------------------------------------*/

/* GLOBAL VARIABLES: the following variables may be used in other files.
*/
ufmControls ufmCONTROLS; /* UFM controls */
/* ---------------------------------------------------------------------*/
/* STATIC VARIABLES: the following variables are for the exclusive use
 *     of the functions defined within this file.
*/
/* Exit status to return to the operating system. */
const int ExitSuccess =  0;
const int ExitFailure = -1;

/* Return status to return to a calling function. */
const int Success = 1;
const int Failure = 0;
/* ---------------------------------------------------------------------*/

/* STATIC PROTOTYPES: the following prototypes are for functions defined
 *     and used exclusively within this file.
*/
static int CmdLineProc (int argc, char **argv, ufmControls *ufmCONTROLS);
static void CreateDefInstr (AGGREGATE  rootIF);
int getopt (int argc, char *const argv[], const char *opts);
static void ufmUserWarning (char *functionName, char *msg);
static void ufmFatalError (char *functionName, char *msg);
int ListToArray(char *List, char AggregateName[256][256]);
void AddParameter();
/* ---------------------------------------------------------------------*/

/*****************************************************************************
 * 
 * NAME: main
 *
 * DESCRIPTION: This is the main function of the UFM tool.
 *
 * PARAMETERS:
 *     argc (int)              The number of command line arguments.
 *     argv (char **)          An array of character pointers.  Each pointer
 *                             points to one of the command line arguments.
 *
 * RETURN VALUE:
 *     Exit status = ExitSuccess --> The UFM tool completed successfully,
 *                                   although some non-severe errors, or
 *                                   warnings, may have occurred.
 *                 = ExitFailure --> The UFM tool encountered a fatal error
 *                                   and could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS (ufmControls)  A data structure containing various controls
 *                                of the UFM tool.
 *	
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     10/05/1998  J.D. Stroup  Original
 *     03/05/1999  K.V. Jarva   Cleared several lint warnings.
 *
 ****************************************************************************/

int main(int argc, char **argv)
{
  AGGREGATE  rootIF, rootODL;
  struct REP report;

  /* Check for arguments */
  if (argc ==1)
  {
    fprintf (stderr, "\nUsage: ufm [-d|-v|-h|-c <comment>|-f <ODL_file>|"
                     "\n       -g <debug_file>|-i <instructions_file>|"
                     "\n       -o <HTML_file>|-t <title>]\n");
    exit(1);
  }

  /* Process the command line options, and initialize the UFM controls. */

  if (CmdLineProc (argc, argv, &ufmCONTROLS) != Success) exit (ExitFailure);
  
  AddParameter();

    /* Initialize the utility modules: */

  	synInitialization( ufmCONTROLS.debug, ufmCONTROLS.verbose, 0,
		ufmCONTROLS.pDebugFile, ufmCONTROLS.pDebugFile );

  	sltInitialization( ufmCONTROLS.debug, ufmCONTROLS.verbose, 0,
		ufmCONTROLS.pDebugFile, ufmCONTROLS.pDebugFile );

  	lstInitialization( ufmCONTROLS.debug, ufmCONTROLS.verbose, 0,
		ufmCONTROLS.pDebugFile, ufmCONTROLS.pDebugFile );

  /* Create the root node of the instructions file tree. */

  if ((rootIF = NewAggregate (NULL, KA_GROUP, "IF_ROOT", NULL)) == NULL)
    ufmFatalError ("main",
                   "Creation of the instructions file's root node failed");

  /* Read the instructions file, and create its tree under its root node. */

  if (ufmCONTROLS.pInstrFile != NULL)
  {
	  if (ReadLabel (ufmCONTROLS.pInstrFile, rootIF) != Success)
		  ufmFatalError ("main",
                  "ReadLabel failed to create the instructions file's tree.");
  }
  else
		ufmUserWarning ("main",
            "No instructions file.  Default instructions only will be used.");

  /* Add default instructions to ODL instructions tree. */

  CreateDefInstr (rootIF);

  /* Run the first pass compile step. */

	if (cmpPass1 (rootIF) != Success)
		ufmFatalError ("main", "Compiler pass 1 failure");

  /* Create the root node of the input ODL tree. */

	if ((rootODL = NewAggregate (NULL, KA_GROUP, "ROOT", NULL)) == NULL)
    ufmFatalError ("main", "Creation of the input ODL root node failed");

  /* Read the input ODL file, and create its tree under its root node. */

  if (ReadLabel (ufmCONTROLS.pODL_In, rootODL) != Success)
		ufmFatalError ("main",
                   "ReadLabel failed to create the ODL file's tree.");

  /* Run the second pass compile step. */

	if (cmpPass2 (&report, rootODL) != Success)
		ufmFatalError( "main", "Compiler pass 2 failure");

  /* Generate the output, HTML report only if no compiler errors
     encountered. */

  if (! ufmCONTROLS.CompilerFailure) {
    if (ReportWriter (&report) != Success)
		  ufmFatalError ("main", "ReportWriter failure");
    exit (ExitSuccess);
  }
  exit (ExitFailure);
  return -1;
}

/*****************************************************************************
 * 
 * NAME: CmdLineProc
 *
 * DESCRIPTION: This function processes the command line options.  It verifies
 *     that the options are valid for the UFM tool.  It also initializes the
 *     UFM control structure to default values, which may be overridden via
 *     command line options.
 *
 * PARAMETERS:
 *     argc (int)              The number of command line arguments.
 *     argv (char **)          An array of character pointers.  Each pointer
 *                             points to one of the command line arguments.
 *     controls (ufmControls)  The UFM's control structure.  It is initialized
 *                             by this function.
 *
 * RETURN VALUE:
 *     CmdLineProc (int)  Success --> This function completed successfully.
 *                        Failure --> This function encountered an error and
 *                                    could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Insufficient memory on the heap, and a title or comment option
 *              given on the command line.
 *        POST: Error message generated stating that memory for the title or
 *              comment string could not be allocated.  The function returns a
 *              status of "Failed". 
 *     2. PRE:  An invalid option given on the command line.
 *        POST: Error message generated stating such.  The function returns a
 *              status of "Failed".
 *     3. PRE:  An option's argument given that begins with a dash--illegal
 *              syntax.
 *        POST: Error message generated stating such.  The function returns a
 *              status of "Failed".
 *     4. PRE:  An option with an argument that is a file name.  The file does
 *              not exist.
 *        POST: Error message generated stating such.  The function returns a
 *              status of "Failed".
 *
 * NOTES:
 *     The standard rules for specifying Unix command line options apply here.
 *
 *     Most valid options to the UFM tool have arguments.  The name of an
 *     option's argument can not begin with a dash ("-").  If it does, then
 *     this function considers this an error: a message is generated to
 *     STDERR, and this function returns with a "failed" status.
 *
 *     If the same option is given multiple times on the command line, the
 *     last instance of the option is used.
 *
 * REVISION HISTORY:
 *     10/05/1998  J.D. Stroup  Original
 *
 ****************************************************************************/

static int CmdLineProc (int argc, char **argv, ufmControls *controls)
{
  extern char *optarg;
  extern int  optind;
  extern int  opterr;

  const char *DefaultTitle     = "UFM Output";
  const char *ValidOptions     = "dvhc:f:g:i:o:t:";
  const char Dash              = '-';
  const char *DefaultInstrFile = "ufmInstructions.odl";

  enum {None, InvalidOption, InvalidOptionArg, OpenFailureODL, OpenFailureIF,
        OpenFailureHTML, OpenFailureDebug, MallocFailureComment,
        MallocFailureTitle} error;

  int opt;

  opterr = 0;   /* disable error messages in getopt */
  error = None;

/* Initialize the fields of the UFM's controls structure to their default
   values. */

  controls->pODL_In    = stdin;
  controls->pInstrFile = NULL;
  controls->pInstrFile = fopen (DefaultInstrFile, "r");
  controls->pHTML_Out  = stdout;
  controls->pDebugFile = controls->pCompilerErrors = stderr;
  controls->verbose    = controls->debug = controls->CompilerFailure = 0;
  controls->pTitle     = strdup (DefaultTitle);
  controls->pComment   = NULL;
  controls->DHeader    = 0 ;   
				 

  if (controls->pTitle == NULL) error = MallocFailureTitle;

/* Retrieve and process each option given on the command line.  This continues
   until all options are exhauseted or an error is encountered. */

  while ((error == None) &&
         ((opt = getopt (argc, argv, ValidOptions)) != EOF))
  {
    switch ((char) opt)
    {
      /* Comment option (c).  Its argument is a string of text. */

      case 'c':
        if (optarg[0] == Dash)
          error = InvalidOptionArg;
        else
        {
          controls->pComment = strdup (optarg);

          if (controls->pComment == NULL) error = MallocFailureComment;
        }
        break;

      /* Debug option (d).  It has no arguments. */

      case 'd':
        controls->debug = 1;
        break;

      /* Input ODL file option (f).  Its argument is the name of a file
         containing the input ODL. */

      case 'f':
        if (optarg[0] == Dash)
          error = InvalidOptionArg;
        else
          if ((controls->pODL_In = fopen (optarg, "r")) == NULL)
            error = OpenFailureODL;
        break;

      /* Error message file option (g).  Its argument is the name of a file
         to which any error or debug messages are written. */

      case 'g':
        if (optarg[0] == Dash)
          error = InvalidOptionArg;
        else
          if ((controls->pCompilerErrors = fopen (optarg, "w")) == NULL)
            error = OpenFailureDebug;
          else
            controls->pDebugFile = controls->pCompilerErrors;
        break;

	/*  Header Information option (h), copied from the orginal ODL file, It has no argument. */

      case 'h':
	 
	      controls->DHeader    = 1;
          
      
        break;
	

      /* Instruction file option (i).  Its argument is the name of an
         instructions file. */

      case 'i':
        if (optarg[0] == Dash)
          error = InvalidOptionArg;
        else
          if ((controls->pInstrFile = fopen (optarg, "r")) == NULL)
            error = OpenFailureIF;
        break;

      /* Output file option (o).  It argument is the name of a file to which
         the output HTML is written. */

      case 'o':
        if (optarg[0] == Dash)
          error = InvalidOptionArg;
        else
          if ((controls->pHTML_Out = fopen (optarg, "w")) == NULL)
            error = OpenFailureHTML;
        break;

      /* Title option (t).  Its argument is a string of text. */

      case 't':
        if (optarg[0] == Dash)
          error = InvalidOptionArg;
        else
        {
          controls->pTitle = strdup (optarg);

          if (controls->pTitle == NULL) error = MallocFailureTitle;
        }
        break;

      /* Verbose option (v).  It has no arguments. */

      case 'v':
        controls->verbose = 1;
        break;

      /* Invalid option or missing mandatory argument of an option. */

      case '?':
        error = InvalidOption;
        break;
    }
  }

  /* If no error were encountered processing the command line options, then
     this function has completed successfully. */

  if (error == None)
    return (Success);

  /* Otherwise, an error has occurred when processing the command line
     options.  Output the appropriate error message to standard error, and
     display the correct usage for the UFM tool. */

  else
  {
    switch (error)
    {
      case InvalidOptionArg:
        fprintf (stderr, "Option argument's name can not begin with a "
                         "dash ('-')\n");
        break;
      case MallocFailureComment:
        fprintf (stderr, "Malloc failure: cannot allocate memory "
                         "for comment string\n");
        break;
      case MallocFailureTitle:
        fprintf (stderr, "Malloc failure: cannot allocate memory "
                         "for title string\n");
        break;
      case OpenFailureODL:
        fprintf (stderr, "Failed to open the given ODL file\n");
        break;
      case OpenFailureIF:
        fprintf (stderr, "Failed to open the given instructions file\n");
        break;
      case OpenFailureHTML:
        fprintf (stderr, "Failed to open the given HTML output file\n");
        break;
      case OpenFailureDebug:
        fprintf (stderr, "Failed to open the given debug/error file\n");
        break;
      case InvalidOption:
        fprintf (stderr, "Invalid option or missing option's argument\n");
        break;
    }
    fprintf (stderr, "\nUsage: ufm [-d|-v|-h|-c <comment>|-f <ODL_file>|"
                     "-g <debug_file>|\n            -i <instructions_file>|"
                     "-o <HTML_file>|-t <title>]\n");
    return (Failure);
  }
} /* end CmdLineProc */

/*****************************************************************************
 * 
 * NAME: CreateDefInstr
 *
 * DESCRIPTION: This function creates the default instructions and hangs them
 *     onto the end (or the right-most siblings) of the instruction file's ODL
 *     tree.
 *
 * PARAMETERS:
 *     rootIF (AGGREGATE)  Pointer to the root node of the instruction file's
 *                         ODL tree.
 *
 * RETURN VALUE:
 *     None.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     There are three default instructions:
 *       1) an instruction for the root node;
 *       2) an instruction for an AGGREGATE node; and
 *       3) an instruction for a PARAMETER node.
 *
 *      The root node instruction is not absolutely necessary since the root
 *      node is an AGGREGATE node, and would be covered by the general
 *      AGGREGATE node instruction.  But a special instruction is used for
 *      root that is not used of other AGGREGATE nodes: the
 *      "ReplaceWithSubstructure" instruction component.
 *
 * REVISION HISTORY:
 *     10/23/1998  J.D. Stroup  Original
 *     12/16/1998  Ken Jarva -- set value_kind.
 *     12/17/1998  J.D. Stroup -- set value_count.
 *
 ****************************************************************************/

static void CreateDefInstr (AGGREGATE  rootIF)
{
  AGGREGATE  p_agg, p_comp_agg;
  PARAMETER  p_parm;
  VALUE_DATA *parm_value = (VALUE_DATA *) malloc (sizeof (struct Value_Data));

  /* The following parameter value information remains constant, so it's set
     here once. */

  parm_value->type         = TV_STRING;
  parm_value->valid        = 1;
  parm_value->format       = 0;
  parm_value->precision    = 0;

  /**********************************************************/
  /* Create the "ufmROOT" AggregateTranslation instruction. */
  /**********************************************************/

  p_agg = NewAggregate (rootIF, KA_OBJECT, "AGGREGATETRANSLATION", NULL);

  /* target name component */

  p_parm = NewParameter (p_agg, KP_ATTRIBUTE, "TARGET");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("ufmROOT");
  parm_value->value.string = strdup ("ufmROOT");

  NewValue (p_parm, parm_value);

  /* list specification component */

  p_comp_agg = NewAggregate (p_agg, KA_OBJECT, "LISTSPEC", NULL);

  p_parm = NewParameter (p_comp_agg, KP_ATTRIBUTE, "ELEMENTS");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("$(_CHILDREN)");
  parm_value->value.string = strdup ("$(_CHILDREN)");

  NewValue (p_parm, parm_value);

  /* "replace with substructure" component */

  p_parm = NewParameter (p_agg, KP_ATTRIBUTE, "REPLACEWITHSUBSTRUCTURE");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("YES");
  parm_value->value.string = strdup ("YES");

  NewValue (p_parm, parm_value);

  /***************************************************************/
  /* Create the "ufmAGGREGATE" AggregateTranslation instruction. */
  /***************************************************************/

  p_agg = NewAggregate (rootIF, KA_OBJECT, "AGGREGATETRANSLATION", NULL);

  /* target class component */

  p_parm = NewParameter (p_agg, KP_ATTRIBUTE, "TARGETCLASS");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("ufmAGGREGATE");
  parm_value->value.string = strdup ("ufmAGGREGATE");

  NewValue (p_parm, parm_value);

  /* list specification component */

  p_comp_agg = NewAggregate (p_agg, KA_OBJECT, "LISTSPEC", NULL);

  p_parm = NewParameter (p_comp_agg, KP_ATTRIBUTE, "ELEMENTS");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("$(_CHILDREN)");
  parm_value->value.string = strdup ("$(_CHILDREN)");

  NewValue (p_parm, parm_value);

  /***************************************************************/
  /* Create the "ufmPARAMETER" AggregateTranslation instruction. */
  /***************************************************************/

  p_agg = NewAggregate (rootIF, KA_OBJECT, "PARAMETERTRANSLATION", NULL);

  /* target class component */

  p_parm = NewParameter (p_agg, KP_ATTRIBUTE, "TARGETCLASS");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("ufmPARAMETER");
  parm_value->value.string = strdup ("ufmPARAMETER");

  NewValue (p_parm, parm_value);

  /* "replace with substructure" component */

  p_parm = NewParameter (p_agg, KP_ATTRIBUTE, "REPLACEWITHSUBSTRUCTURE");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("YES");
  parm_value->value.string = strdup ("YES");

  NewValue (p_parm, parm_value);

  /* assignment specification component */

  p_comp_agg = NewAggregate (p_agg, KA_OBJECT, "ASSIGNMENTSPEC", NULL);

  /* left side */

  p_parm = NewParameter (p_comp_agg, KP_ATTRIBUTE, "LEFTSIDE");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("$(_CURRCOMP)");
  parm_value->value.string = strdup ("$(_CURRCOMP)");

  NewValue (p_parm, parm_value);

  /* right side */

  p_parm = NewParameter (p_comp_agg, KP_ATTRIBUTE, "RIGHTSIDE");
  p_parm->value_kind = KV_SCALAR;
  p_parm->value_count = 1;

  parm_value->length       = (short) strlen ("$(_VALUE)");
  parm_value->value.string = strdup ("$(_VALUE)");

  NewValue (p_parm, parm_value);

  return;
} /* end CreateDefInstr */

/*****************************************************************************
 * 
 * NAME: ufmFatalError
 *
 * DESCRIPTION: Reports fatal errors and shuts down the program.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

void ufmFatalError( char *functionName, char *msg )
{
  fprintf( ufmCONTROLS.pDebugFile, "%s: FATAL - %s\n", functionName, msg );
	exit(ExitFailure);

} /* end ufmFatalError */

/*****************************************************************************
 * 
 * NAME: ufmUserWarning
 *
 * DESCRIPTION: Reports warnings if the verbose flag in the UFM control
 *     structure is set (= 1).
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.verbose  -- (int)  Flag indicating whether the program is
 *                                    in verbose mode (= 1) or not (= 0).
 *     ufmCONTROLS.pCompilerErrors
 *                          -- (FILE) The debug file to which error messages
 *                                    are to be written.
 *	
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

void ufmUserWarning( char *functionName, char *msg )
{
	if( !ufmCONTROLS.verbose ) return;

  fprintf( ufmCONTROLS.pCompilerErrors, "%s: WARNING - %s\n", functionName,
                                                              msg );
	return;

} /* end ufmUserWarning */

void AddParameter()
{
  int i, j, k, status;
  int ColumnNumber=0, AggregateNumber=0;
  char TargetName[256], column[256];
  char columns[256][256];
  char AggregateName[256][256];
  char tmpname[80];
  FILE *TempFile;
  AGGREGATE root, parent, node, aAggregate, bAggregate;
  PARAMETER aParameter, bParameter;
  VALUE aValue, bValue;
  VALUE_DATA aValueData, *bValueData;
  char *StringValue;

  /* 
  Get columns from instruction file. 
  */
  fseek(ufmCONTROLS.pInstrFile, 0, 0);
  root=NewAggregate(NULL, KA_GROUP, "ODL", NULL);
  status=ReadLabel(ufmCONTROLS.pInstrFile, root);
  if (status!=1)
  {
    fprintf(stderr, "There are syntax errors in instruction file.\n");
    exit(1);
  }

  node=root;
  while(node!=NULL)
  {
    if(strcmp(node->name, "TABLESPECBYCOLUMN")==0)
    {
      /* 
      Get target value. 
      */
      aParameter=FindParameter(node->parent, "TARGET");
      aValue=FirstValue(aParameter);
      if (aValue==NULL)
      {
        fprintf(stderr, "No value is found.\n");
        exit(1);
      }
      switch (aValue->item.type)
      {
        case TV_STRING:
          strcpy(TargetName, aValue->item.value.string);
          break;
        default:
          printf("Not correct data type.\n");
          break;
      }
      strcat(TargetName, "/");

      /* 
      Get column value. 
      */
      aParameter=FindParameter(node, "COLUMNS");
      if (aParameter==NULL)
      {
        fprintf(stderr, "No column parameter is found.\n");
        exit(1);
      }
 
      /*  
      Get value number and value. 
      */
      for (i=0; i<aParameter->value_count; i++)
      {
        if (i==0) 
        {
          aValue=FirstValue(aParameter);
        }
        else
        {
          aValue=NextValue(aValue);
        }

        strcpy(column, TargetName);
        switch (aValue->item.type)
        {
          case TV_STRING:
            strcat(column, aValue->item.value.string);
            break;
          default:
            printf("Not correct data type.\n");
            break;
        }
        strcpy(columns[ColumnNumber], column);
        ColumnNumber++;
      }
    }

    /* 
    Next aggregate. 
    */ 
    node=NextAggregate(node);
  }

  free(root);
  fseek(ufmCONTROLS.pInstrFile, 0, 0);

  /* Add blank value "" to ODL. */
  fseek(ufmCONTROLS.pODL_In, 0, 0);
  root=NewAggregate(NULL, KA_GROUP, "ODL", NULL);
  status=ReadLabel(ufmCONTROLS.pODL_In, root);
  if (status!=1)
  {
    fprintf(stderr, "There are syntax errors in metadata file.\n");
    exit(1);
  }

  /* 
  Make a blank value in "" format. All "" cells in table will be together.
  If use "blank", it is separate. 
  */
  aValueData=ODLConvertString("", 0);
  bValueData=&aValueData;

  /* 
  For all columns.
  */
  for (i=0; i<ColumnNumber; i++)
  {
    AggregateNumber=ListToArray(columns[i], AggregateName);
    node=root;

    /*
    Get target.
    */
    node=FindAggregate(node, AggregateName[0]);
    while(node!=NULL)
    {
      /* If it is target. */
      if(strcmp(node->name, AggregateName[0])==0)
      {
        while(node!=NULL)
        {
          /*
          There are multiple table rows.
          */
          if(strcmp(node->name, AggregateName[1])==0)
          {
            for (j=2; j<AggregateNumber-1; j++)
            {
              parent=node;
              node=FindAggregate(parent, AggregateName[j]);
              if (node==NULL)
              {
                node=NewAggregate(node, KA_GROUP, AggregateName[j], NULL);
                if (node==NULL)
                {
                  fprintf(stderr, "%s cannot be created.\n", AggregateName[j]);
                  exit(1);
                }
              }
            } 
            aParameter=FindParameter(node, AggregateName[AggregateNumber-1]);
            if (aParameter==NULL)
            {
              aParameter=NewParameter(node, KP_ATTRIBUTE, AggregateName[AggregateNumber-1]);
              if (aParameter==NULL)
              {
                fprintf(stderr, "Cannot create %s.\n", AggregateName[AggregateNumber-1]);
                exit(1);
              }
              bValue=NewValue(aParameter, bValueData);
              if (bValue==NULL)
              {
                fprintf(stderr, "Cannot add value to %s.\n", AggregateName[AggregateNumber-1]);
                exit(1);
              }
            }
          }
          node=NextAggregate(node);
        }
      }
      node=NextAggregate(node);
    }
  }

  fclose(ufmCONTROLS.pODL_In);
  ufmCONTROLS.pODL_In=tmpfile();
  WriteLabel(ufmCONTROLS.pODL_In, root); 
  free(root);
  fseek(ufmCONTROLS.pODL_In, 0, 0);
}

/* Get aggregate number and name. */
int ListToArray(char *List, char AggregateName[256][256])
{
  char *aString;
  int AggregateNumber=0;

  aString = strtok(List, "/");
  while(aString != NULL )
  {
    strcpy(AggregateName[AggregateNumber], aString);
    AggregateNumber++;
    aString = strtok(NULL, "/");
  }
  
  return AggregateNumber;
}

